/***********************************************************************
This software module was originally developed by Jean Le Feuvre
(avipix, llc) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. avipix, llc retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2000.
************************************************************************/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  MP4 File Writer
FILE:       main.cpp
AUTHOR:     Jean Le Feuvre (avipix, llc, jean@avipix.com), December 2000
OVERVIEW:   main: main() function for the MP4 File Writer
-------------------------------------------------------------------------*/

#include <direct.h>

#include "defs.h"
#define MAIN
#include "MP4Enc.h"
#include "MP4Dump.h"
#include "Token.h"
#include "SysCoder.h"
#include "VrmlPars.h"


BOOL g_bErrors;
Token nextToken;


void PrintUsage()
{
	printf ("MP4Enc [-w [-t] || -r [-d -f] ] inputfile\n"
			"Writing Options:\n"
			"-w: write an MP4 file: *.script->*.mp4\n"
			"-i: write the samples interleaved\n"
			"-t: interleaving detailed output in M4E_*.trc\n"
			"\nReading Options:\n"
			"-r: read an MP4 file and print info: *.mp4->M4D_*.trc\n"
			"-d: dump each track to files (name at prompt)\n"
			"-f: forces track dump in files:\n"
			"\tM4D_*_[TrackID].trc: SampleDescription and AU timing,\n"
			"\tM4D_*_[TrackID].raw: stream raw data\n"
			"\nMP4Enc v3.1 - Written by Jean Le Feuvre, avipix, llc - (c) 2000\n");

}

int main (int argc, char **argv)
{
	FILE *fList, *fTrace, *fout;
	int nNameLength;
	ZString stMovieName ("");
	ZString stTraceName ("");
	ZString stFileName("");
	ZString stPathName("");
	int i;
	char *arg;
	BOOL bTraceActive = FALSE;
	int ReadMode = -1;
	BOOL dumpRequest = FALSE;
	BOOL forceDump = FALSE;
	BOOL bInter = FALSE;

	fout = NULL;

	//parse our agrs
	for (i = 1; i < argc ; i++) {
		arg = argv[i];
		if (arg[0] != '-') {
			stPathName = arg;
			break;
		}		
		//parse the flags
		switch (arg[1]) {
		case 'w':
			ReadMode = 0;
			break;
		case 's':
			fout = freopen("stdout.txt", "wt", stdout);
			break;
		case 'r':
			ReadMode = 1;
			break;
		case 't':
			bTraceActive = TRUE;
			break;
		case 'd':
			dumpRequest = TRUE;
			break;
		case 'f':
			forceDump = TRUE;
			break;
		case 'i':
			bInter = TRUE;
			break;
		default:
			PrintUsage();
			return -1;
		}
	}

	nNameLength = stPathName.GetLength ();
	stFileName = strrchr (stPathName, '\\');
	if (stFileName.GetLength()) {
		stPathName.Cut(stPathName.GetLength() - stFileName.GetLength() + 1);
		_chdir(stPathName);
		stFileName.DeleteFrom(0);
	} else {
		stFileName = stPathName;
	}
	

	if (!nNameLength) {
		printf("Missing file name\n\n");
		PrintUsage();
		return 0;
	}

	//Quick hack to set to write if -w if not set
	if (ReadMode == -1) {
		if (dumpRequest || forceDump) {
			PrintUsage();
			return 0;
		}
		ReadMode = 0;
	} else if (ReadMode == 1) {
		//Dump the file
		MP4Dump dumper(stFileName, dumpRequest, forceDump);
		dumper.Dump();
		return 0;
	}
	

	/*-----------------------------------------------------------------
	Source : ".script", output ".mp4" and list is ".trc"
	-----------------------------------------------------------------*/

	nNameLength = stFileName.GetLength();
	LPCSTR szExt = strrchr(stFileName, '.');
	if (szExt && !strchr (szExt, '\\'))
		nNameLength -= strlen (szExt);
	else
		stFileName += ".scr";

	VRMLParser parser;
	if (!parser.Init (stFileName))
	{	printf ("Cannot open %s\n", (LPCSTR) stFileName);
		parser.Shutdown ();
		return -1;		
	}

	stFileName.Cut (nNameLength);
	stFileName += ".mp4";
	stMovieName = stFileName;

	stTraceName = "M4E_";
	stFileName.Cut (nNameLength);
	stFileName += ".trc";
	stTraceName  += stFileName;

	stFileName.Cut (nNameLength);
	stFileName += ".trc";
	if ((fList = fopen (stFileName, "w")) == NULL)
	{	printf ("Cannot open %s\n", (LPCSTR) stFileName);
		return -1;
	}
	
	fTrace = NULL;
	if (bTraceActive && bInter) {
		fTrace = fopen((LPCSTR) stTraceName, "wt");
		if (fTrace == NULL) 
			printf("Warning: cannot open trace file...\n");
	}

	printf("Writing file %s...\n", (LPCSTR) stMovieName);
	//Init text parsing
	nextToken.Init (&parser, stFileName, fList);
	//Start encoding
	MP4Enc((LPCSTR) stMovieName, bInter).Start (fList, TRACE_NONE, fTrace);

	fclose(fList);

	printf ("...done\n");

	if (g_bErrors) {
		printf ("Hit any key to continue\n");
		getchar ();
		return -1;
	}

	if (fTrace) fclose(fTrace);
	if (fout) fclose(fout);
	DeleteFile("stdout.txt");
	return 0;
}