/*****************************************************************************
 *
 * This software module was originally developed by
 *
 *   Weiping Li (Optivision, Inc.) and
 *   Fan Ling (General Instrument)
 *
 * and edited by 
 * 
 *   Hong Jiang, Greg Thayer (Intel Corporation)
 * 
 * in the course of development of the MPEG-4 Video (ISO/IEC 14496-2) standard.
 * This software module is an implementation of a part of one or more MPEG-4
 * Video (ISO/IEC 14496-2) tools as specified by the MPEG-4 Video (ISO/IEC
 * 14496-2) standard.
 *
 * ISO/IEC gives users of the MPEG-4 Video (ISO/IEC 14496-2) standard free
 * license to this software module or modifications thereof for use in hardware
 * or software products claiming conformance to the MPEG-4 Video (ISO/IEC
 * 14496-2) standard.
 *
 * Those intending to use this software module in hardware or software products
 * are advised that its use may infringe existing patents. The original
 * developer of this software module and his/her company, the subsequent
 * editors and their companies, and ISO/IEC have no liability for use of this
 * software module or modifications thereof in an implementation. Copyright is
 * not released for non MPEG-4 Video (ISO/IEC 14496-2) Standard conforming
 * products.
 *
 * ACTS-MoMuSys partners retain full right to use the code for his/her own
 * purpose, assign or donate the code to a third party and to inhibit third
 * parties from using the code for non MPEG-4 Video (ISO/IEC 14496-2) Standard
 * conforming products. This copyright notice must be included in all copies or
 * derivative works.
 *
 * Copyright (c) 1997
 *
 * Revision History:
 * 
 *  10/29/99: FGS WD2 VOL header from Intel 
 *  12/15/99: FGS SE (Selective Enhancement) from FT CNET by Bertrand THEBAULT
 *  12/22/99: FGS FW (Frequency Weighting) from Intel by Greg Thayer, Hong Jiang
 *  08/15/00: FPDAM4 Syntax from Intel by Hong Jiang
 *
 *****************************************************************************/

/************************    INCLUDE FILES    ********************************/
#include <stdio.h>
#include <fstream.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "vlcbp.hpp"
#include "typeapi.h"
#include "codehead.h"
#include "global.hpp"
#include "entropy/bitstrm.hpp"
#include "entropy/entropy.hpp"
#include "entropy/huffman.hpp"
#include "dct.hpp"
#include "mode.hpp"
#include "vopses.hpp"
#include "vopsedec.hpp"

// Intel 10/29/99
//FGS : start
//#include "sys/fgs.hpp"
//#include "sys/vlcbp4fw.hpp"
#include "fgs.hpp"
#include "vlcbp4fw.hpp"
//FGS : end

#ifndef SEEK_SET
#define SEEK_SET 0
#endif

const Int zigzag[64] = {
	0, 1, 5, 6,14,15,27,28,
		2, 4, 7,13,16,26,29,42,
		3, 8,12,17,25,30,41,43,
		9,11,18,24,31,40,44,53,
		10,19,23,32,39,45,52,54,
		20,22,33,38,46,51,55,60,
		21,34,37,47,50,56,59,61,
		35,36,48,49,57,58,62,63
};

// added by Wu Feng for FGS error resilience, MSRCN, 2000/8/20
const UInt ifillValue[8] = {
	0x00, 0x01, 0x03, 0x07, 
	0x0f, 0x1f, 0x3f, 0x7f
}; 

// Intel 10/29/99
Void CVideoObjectDecoder::decodeFGSVOHead ()
{
	findStartCode(m_pFGSbitstrmIn);
	//	Added for data partitioning mode By Toshiba(1998-1-16:DP+RVLC)(for Skip Session_Start_Code)
	if( m_pFGSbitstrmIn->peekBits(NUMBITS_START_CODE_SUFFIX)==SESSION_START_CODE){
		m_pFGSbitstrmIn -> getBits (NUMBITS_START_CODE_SUFFIX);
		if (findStartCode (m_pFGSbitstrmIn) == EOF)
			assert(FALSE);
	}
	//	End Toshiba(1998-1-16:DP+RVLC)
	UInt uiVoStartCode = m_pFGSbitstrmIn -> getBits (NUMBITS_VO_START_CODE);
	assert(uiVoStartCode == VO_START_CODE);
	
	m_uiVOId = m_pFGSbitstrmIn -> getBits (NUMBITS_VO_ID);
	m_pFGSbitstrmIn->resetStartCode();
}


// FGS VOL header conform to FGS WD 2.0
// Intel 10/29/99
// Modified to conform to FPDAM4
// Intel 6/22/2000
Void CVideoObjectDecoder::decodeFGSVOLHead ()
{
	findStartCode(m_pFGSbitstrmIn);
	UInt uiVolStartCode = m_pFGSbitstrmIn -> getBits (NUMBITS_VOL_START_CODE);
	assert(uiVolStartCode == VOL_START_CODE);
	UInt uiVOLId = m_pFGSbitstrmIn -> getBits (NUMBITS_VOL_ID);
	
	Bool bRandom = m_pFGSbitstrmIn->getBits (1);				// VOL_Random_Access
	
	// begin FPDAM4 (for fgs_layer_type=FGS), Intel, 6/22/2000
	UInt uiOLType = m_pFGSbitstrmIn -> getBits(8);				// VOL_type_indication
	assert(uiOLType == VIDEO_OBJECT_TYPE_STREAMING_VIDEO);		// Make sure this is an FGS bitstream
	UInt uiFGSLayerType = m_pFGSbitstrmIn -> getBits(2);		// fgs_layer_type 
	assert(uiFGSLayerType == FGS_LAYER_TYPE_FGS);				// Make sure this is an FGS bitstream
	// end   FPDAM4 (for fgs_layer_type=FGS), Intel, 6/22/2000

	m_pFGSbitstrmIn -> getBits (3);								//"VOL_priority", useless flag for now
	
	UInt uiAspect = m_pFGSbitstrmIn -> getBits (4);
	if(uiAspect==15) // extended PAR
	{
		UInt iParWidth = m_pFGSbitstrmIn -> getBits (8);
		UInt iParHeight = m_pFGSbitstrmIn -> getBits (8);
	}
	
	UInt uiMark;
	UInt uiCTP = m_pFGSbitstrmIn -> getBits (1); //VOL_Control_Parameter, useless flag for now
	if(uiCTP)
	{
		UInt uiChromaFormat = m_pFGSbitstrmIn -> getBits (2);
		UInt uiLowDelay = m_pFGSbitstrmIn -> getBits (1);
	}
	//assert(uiCTP == 0);
	
	uiMark = m_pFGSbitstrmIn -> getBits (1);
	assert(uiMark==1);
	Int iClockRate = m_pFGSbitstrmIn -> getBits (NUMBITS_TIME_RESOLUTION);
	assert(iClockRate == m_volmd.iClockRate);
	/*
	m_volmd.iClockRate = m_pFGSbitstrmIn -> getBits (NUMBITS_TIME_RESOLUTION);
	Int iClockRate = m_volmd.iClockRate;
	assert (iClockRate >= 1 && iClockRate < 65536);
	for (m_iNumBitsTimeIncr = 1; m_iNumBitsTimeIncr < NUMBITS_TIME_RESOLUTION; m_iNumBitsTimeIncr++)	{	
	if (iClockRate == 1)			
	break;
	iClockRate = (iClockRate >> 1);
	}
	*/
	
	uiMark = m_pFGSbitstrmIn -> getBits (1);
	assert(uiMark==1);
	Bool bFixFrameRate = m_pFGSbitstrmIn -> getBits (1);
	//assert (bFixFrameRate == FALSE);
	
	if(bFixFrameRate)
	{
		UInt uiFixedVOPTimeIncrement = m_pFGSbitstrmIn -> getBits (m_iNumBitsTimeIncr); //INTEL
	}
	
	UInt uiMarker = m_pFGSbitstrmIn -> getBits (1);
	assert(uiMarker==1);
	uiMarker = m_pFGSbitstrmIn -> getBits (NUMBITS_VOP_WIDTH);
	//		m_ivolWidth = m_pFGSbitstrmIn -> getBits (NUMBITS_VOP_WIDTH);
	uiMarker  = m_pFGSbitstrmIn -> getBits (1);
	assert(uiMarker==1);
	uiMarker = m_pFGSbitstrmIn -> getBits (NUMBITS_VOP_HEIGHT);
	//		m_ivolHeight = m_pFGSbitstrmIn -> getBits (NUMBITS_VOP_HEIGHT);
	uiMarker  = m_pFGSbitstrmIn -> getBits (1);
	assert(uiMarker==1);
	
	uiMark = m_pFGSbitstrmIn -> getBits (1); // interlace (was vop flag)
	//	m_vopmd.bInterlace = m_pFGSbitstrmIn -> getBits (1); // interlace (was vop flag)
	
	// Added for FGS FW by Intel 12/16/99
	// Only real info to be gained from header
	m_volmd.bFGSFreqWeightingEnable = m_pFGSbitstrmIn -> getBits (1);
	if (m_volmd.bFGSFreqWeightingEnable == TRUE) {
		int val, i=0;
		for (i=0; i<BLOCK_SQUARE_SIZE; i++)
			m_volmd.rgiWeightMatrix[i] = 0;

		m_volmd.bLoadWeightMatrix = m_pFGSbitstrmIn -> getBits (1);
		if (m_volmd.bLoadWeightMatrix) {
			// 3-bit fixed length values with zero ending
			i = 0;
			do {
				val = m_pFGSbitstrmIn -> getBits (3);
				m_volmd.rgiWeightMatrix[grgiStandardZigzag[i]] = val;
				i++;
			} while ((i < BLOCK_SQUARE_SIZE) && (val!=0));
		}
	}

//FGS3 : start
	Bool bFGSQuarterSample;
	bFGSQuarterSample = m_pFGSbitstrmIn -> getBits (1);
	assert(bFGSQuarterSample==0);
//FGS3 : end

	// Modified by MSRCN for FGS error resilience, 2000/8/20
	//UInt uiResyncMarkerDisable = m_pFGSbitstrmIn -> getBits (1);
	m_volmd.bFGSErrorResilienceDisable = m_pFGSbitstrmIn -> getBits (1);

	// HONG: Why flush here? It should be taken care of by VOP header handler!
	m_pFGSbitstrmIn->flush();
	m_pFGSbitstrmIn->resetStartCode();
	//	m_volmd.bDeblockFilterDisable = TRUE;							//no deblocking filter
}




Void CVideoObjectDecoder::Init_FGS()
{
	int i, j, k;
	/* init FGS decoding */
	vop_counter = 0;
	FGSstop = 0;
	
//	printf("Initialize FGS data array....\n");
	
	Nblock = m_ivolWidth*m_ivolHeight*6/256;
	
	// added by wu feng for FGS error resilience, MSRCN, 2000/8/20
	assert (Nblock > 0);
	m_numBitsVPMBnum = (int) ceil( log10((double)(Nblock/6)) / log10(2.0));

	for ( i=0; i<Nblock; i++ ) {
		enh_block_data[i].cbp = 1;
		enh_block_data[i].dct_type = 0;
		//MAX_FGS_SE_SHIFT_FACTOR added for FGS SE by FT CNET & PHILIPS Research US
		//MAX_FGS_FW_SHIFT_FACTOR added for FGS FW by INTEL
		for ( j=0; j<12+MAX_FGS_SE_SHIFT_FACTOR+MAX_FGS_FW_SHIFT_FACTOR; j++ ) { 
			//end SE, FW
			for ( k=0; k<64; k++ ) {
				enh_block_data[i].run[j][k]= 63;
				enh_block_data[i].end[j][k]= 0;
			}
		}
		for ( k=0; k<64; k++ ) {
			enh_block_data[i].sign[k]= 0;
			enh_block_data[i].base[k]= 0;
		}
	}

	//added for FGS SE by FT CNET
	Int tmpVal = Nblock/6;
	for ( i=0; i<tmpVal; i++ ) {
		pc_mb_shift_factor_received[i]=0;
		pc_shift_factor[i]=0;
	}//end SE

	Nblock=0;
}


Void CVideoObjectDecoder::updateFGSBuff ()
{
	if (m_vopmd.vopPredType != BVOP) {
		swapVOPU8Pointers (m_pvopcFGS1, m_pvopcFGS2);
	}
}


Int CVideoObjectDecoder::decodeFGSVOP()
{
    Int                 mbnum;
    Int					eof;
    UInt				tmpvar;
    Int					Nbits;
	
	eof=0;
	Nbits=0;
	
	m_pFGSbitstrmIn->turnonFGS();
	tmpvar = 0;

	if (m_bFGSScalability)
	{
		// added by Hong Jiang, INTEL, 8/8/00
		// Note that we assume that FGS VOP Header is NEVER truncated!!!
		// Start with new header

		if (findStartCode (m_pFGSbitstrmIn) == EOF)
		{
			return EOF;
		}

		UInt uiVopStartCode = m_pFGSbitstrmIn -> getBits (NUMBITS_VOP_START_CODE);
		if (uiVopStartCode == GOV_START_CODE) {
			m_bUseGOV = TRUE;
			m_bLinkisBroken = FALSE;
			Int timecode;
			timecode  = m_pFGSbitstrmIn -> getBits (5) * 3600;
			timecode += m_pFGSbitstrmIn -> getBits (6) * 60;
			m_pFGSbitstrmIn -> getBits (1);
			timecode += m_pFGSbitstrmIn -> getBits (6);

			m_tModuloBaseDecd = timecode;
			m_tModuloBaseDisp = timecode;

			cout << "GOV Header (t=" << timecode << ")\n\n";
			Int closed_gov  = m_pFGSbitstrmIn -> getBits (1);
			Int broken_link = m_pFGSbitstrmIn -> getBits (1);
			if ((closed_gov == 0)&&(broken_link == 1))
				m_bLinkisBroken = TRUE;
			
			findStartCode (m_pFGSbitstrmIn);
			uiVopStartCode = m_pbitstrmIn -> getBits (NUMBITS_VOP_START_CODE);
		}

		assert (uiVopStartCode == FGS_VOP_START_CODE);
		// End with the New Header code

/* Removed by Hong Jiang, INTEL, 8/8/00
		m_pFGSbitstrmIn->flush();
		while(1){
			if(m_pFGSbitstrmIn->eof()==EOF){ eof=EOF; goto FGSRecon; }
			tmpvar = (tmpvar<<8) + m_pFGSbitstrmIn->getBits(8);
			Nbits += 8;
	//		if((tmpvar & 0xffffff) == START_CODE_PREFIX) break;
			if((tmpvar ==  (0x100 | FGS_VOP_START_CODE)) || 
				(tmpvar == 0x1B1)) break;

		}
	//	tmpvar = (tmpvar<<8) + m_pFGSbitstrmIn->getBits(8);
	//	Nbits += 8;

		// VISUAL_OBJECT_SEQUENCE_END_CODE == 0x1B1
	//	if((tmpvar == 0x1B1) || (m_pFGSbitstrmIn->eof()==EOF)){ eof=EOF; goto FGSRecon; }
	//	
	//	if(tmpvar !=  (0x100 | FGS_VOP_START_CODE))
	//	{
	//		printf("Empty FGS VOP\n");
	//		goto FGSRecon;
	//	}
*/
		
		// begin FPDAM4, Intel, 6/28/00: read fgs_vop_coding_type
		fgs_vop_coding_type = (Char) m_pFGSbitstrmIn -> getBits (NUMBITS_FGS_VOP_CODING_TYPE);
		Nbits += NUMBITS_FGS_VOP_CODING_TYPE;
		assert(fgs_vop_coding_type==FGS_VOP_CODING_TYPE_I); // testing purpose for now 6/29/00

		// Time reference and VOP_pred_type
		while (m_pFGSbitstrmIn -> getBits (1) != 0);

		m_pFGSbitstrmIn -> getBits (1); //	uiMarker 
		m_pFGSbitstrmIn -> getBits (m_iNumBitsTimeIncr);  // tVopIncr 
		m_pFGSbitstrmIn -> getBits (1); // marker bit
			
		// Intel 11/22/99, changed from 4 bits to 5 bits
		fgs_vop_max_level_y = (Int) m_pFGSbitstrmIn->getBits(5);
		Nbits += 5;
		
		fgs_vop_max_level_u = (Int) m_pFGSbitstrmIn->getBits(5);
		Nbits += 5;
		
		fgs_vop_max_level_v = (Int) m_pFGSbitstrmIn->getBits(5);
		Nbits += 5;
		// end

//FGS3 : start
		m_pFGSbitstrmIn -> getBits (1); //	uiMarker 
		fgs_vop_number_of_vop_bp_coded = (Int) m_pFGSbitstrmIn->getBits(5);

		fgs_vop_mc_bit_plane_not_used = (Int) m_pFGSbitstrmIn->getBits(5);
		Nbits += 5;
//		fgs_vop_mc_bit_plane_not_used = (Int) m_pFGSbitstrmIn->getBits(4);
//		Nbits += 4;
//FGS3 : end
		
		fgs_vop_selective_enhancement_enable = (Int) m_pFGSbitstrmIn->getBits(1);
		Nbits += 1;

	}
	
	Nblock = 6*m_ivolHeight*m_ivolWidth/(16*16);
	
	if(m_pFGSbitstrmIn->fgseof()==EOF){ eof=EOF; goto FGSRecon; }
	
	m_pFGSbitstrmIn->resetStartCode();
	eof = input_enh_bitstream();
	if(FGSstop)m_pFGSbitstrmIn->resetStartCode();
	Nblock = 0;
	
FGSRecon:
	Int iMBX, iMBY;
	Int iMBXstart, iMBXstop, iMBYstart, iMBYstop; // Added by KPN for short headers
	PixelC* ppxlcRefY;
	PixelC* ppxlcRefU;
	PixelC* ppxlcRefV;
	
	if(m_vopmd.vopPredType!=BVOP){
		ppxlcRefY = (PixelC*) m_pvopcFGS1->pixelsY () + m_iStartInRefToCurrRctY;
		ppxlcRefU = (PixelC*) m_pvopcFGS1->pixelsU () + m_iStartInRefToCurrRctUV;
		ppxlcRefV = (PixelC*) m_pvopcFGS1->pixelsV () + m_iStartInRefToCurrRctUV;
	}
	else{
		ppxlcRefY = (PixelC*) m_pvopcFGS2->pixelsY () + m_iStartInRefToCurrRctY;
		ppxlcRefU = (PixelC*) m_pvopcFGS2->pixelsU () + m_iStartInRefToCurrRctUV;
		ppxlcRefV = (PixelC*) m_pvopcFGS2->pixelsV () + m_iStartInRefToCurrRctUV;
	}
	
	iMBXstart=0;
	iMBXstop= m_iNumMBX;
	iMBYstart=0;
	iMBYstop= m_iNumMBY;
	
	PixelC* ppxlcY = ppxlcRefY;
	PixelC* ppxlcU = ppxlcRefU;
	PixelC* ppxlcV = ppxlcRefV;
	
	for (iMBY = iMBYstart; iMBY < iMBYstop; iMBY++) { // [FDS]
		PixelC* ppxlcRefMBY = ppxlcRefY;
		PixelC* ppxlcRefMBU = ppxlcRefU;
		PixelC* ppxlcRefMBV = ppxlcRefV;
		
		for (iMBX = iMBXstart; iMBX < iMBXstop; iMBX++) {
			mbnum = iMBY * iMBXstop + iMBX;
			
			decodeTextureFGSMB(mbnum, MB_SIZE);
			
			//			if(enh_block_data[mbnum*6].mode == 0 ) 

#ifdef FGSPOSTCLIP
			// added the residue output from decodeTextureFGSMB to the final baselayer picture
			// Here ppxlcRefMBY/U/V hold the final base not the prediction portion
			// Changes must be make in decodeVOP()
			addErrorToCurrQ(ppxlcRefMBY,ppxlcRefMBU,ppxlcRefMBV);
#else
			if(m_vopmd.vopPredType == IVOP) 
				copyErrorToCurrQ(ppxlcRefMBY,ppxlcRefMBU,ppxlcRefMBV);
			else
				addErrorToCurrQ(ppxlcRefMBY,ppxlcRefMBU,ppxlcRefMBV);
#endif

			ppxlcRefMBY += MB_SIZE;
			ppxlcRefMBU += BLOCK_SIZE;
			ppxlcRefMBV += BLOCK_SIZE;
		}
		
		ppxlcRefY += m_iFrameWidthYxMBSize;
		ppxlcRefU += m_iFrameWidthUVxBlkSize;
		ppxlcRefV += m_iFrameWidthUVxBlkSize;
	}

    return(eof);
}


Void CVideoObjectDecoder::decodeTextureFGSMB(Int mbnum, Int MB_width)
{
	Int  diff[BLOCK_SIZE * BLOCK_SIZE], sign[BLOCK_SIZE * BLOCK_SIZE];
	Int* rgiBlkCurrQ = NULL;
	Int iWidthCurrQ;
	for (Int iBlk = Y_BLOCK1; iBlk <= V_BLOCK; iBlk++) {
		if (iBlk < (UInt) U_BLOCK) {
			switch (iBlk)
			{
			case (Y_BLOCK1):
				rgiBlkCurrQ = m_ppxliFGSErrorMBY;
				break;
			case (Y_BLOCK2):
				rgiBlkCurrQ = m_ppxliFGSErrorMBY + BLOCK_SIZE;
				break;
			case (Y_BLOCK3):
				rgiBlkCurrQ = m_ppxliFGSErrorMBY + MB_SIZE*BLOCK_SIZE; 
				break;
			case (Y_BLOCK4):
				rgiBlkCurrQ = m_ppxliFGSErrorMBY + MB_SIZE*BLOCK_SIZE + BLOCK_SIZE;
				break;
			}
			iWidthCurrQ = MB_SIZE;
		} else {
			iWidthCurrQ = BLOCK_SIZE;
			rgiBlkCurrQ = (iBlk == U_BLOCK) ? m_ppxliFGSErrorMBU: m_ppxliFGSErrorMBV;
		}
		
		bp2block(mbnum*6+iBlk-1,diff,sign);
		Recoverbp(mbnum*6+iBlk-1,diff,sign,m_rgiDCTcoef);

		m_pidct->apply (m_rgiDCTcoef, BLOCK_SIZE, rgiBlkCurrQ, iWidthCurrQ);
	}
	if (enh_block_data[mbnum*6].dct_type) 
		fieldDCTtoFrameI(m_ppxliFGSErrorMBY);
}

Void CVideoObjectDecoder::setZeroToCurrQ(
										 PixelC* ppxlcQMBY, PixelC* ppxlcQMBU, PixelC* ppxlcQMBV)
{
	CoordI ix, iy, ic = 0;
	for (iy = 0; iy < MB_SIZE; iy++) {
		for (ix = 0; ix < MB_SIZE; ix++, ic++) {
			ppxlcQMBY [ix] = 0;
		}
		ppxlcQMBY += m_iFrameWidthY;
	}
	
	ic = 0;
	for (iy = 0; iy < BLOCK_SIZE; iy++) {
		for (ix = 0; ix < BLOCK_SIZE; ix++, ic++) {
			ppxlcQMBU [ix] = 0;
			ppxlcQMBV [ix] = 0;
		}
		ppxlcQMBU += m_iFrameWidthUV;
		ppxlcQMBV += m_iFrameWidthUV;
	}
}


Void CVideoObjectDecoder::copyErrorToCurrQ(
										   PixelC* ppxlcQMBY, PixelC* ppxlcQMBU, PixelC* ppxlcQMBV)
{
	CoordI ix, iy, ic = 0;
	for (iy = 0; iy < MB_SIZE; iy++) {
		for (ix = 0; ix < MB_SIZE; ix++, ic++) {
			ppxlcQMBY [ix] = m_rgiClipTab[m_ppxliFGSErrorMBY [ic]]; 
		}
		ppxlcQMBY += m_iFrameWidthY;
	}
	
	ic = 0;
	for (iy = 0; iy < BLOCK_SIZE; iy++) {
		for (ix = 0; ix < BLOCK_SIZE; ix++, ic++) {
			ppxlcQMBU [ix] = m_rgiClipTab [m_ppxliFGSErrorMBU [ic]];
			ppxlcQMBV [ix] = m_rgiClipTab [m_ppxliFGSErrorMBV [ic]];
		}
		ppxlcQMBU += m_iFrameWidthUV;
		ppxlcQMBV += m_iFrameWidthUV;
	}
}

Void CVideoObjectDecoder::addErrorToCurrQ(
										  PixelC* ppxlcQMBY, PixelC* ppxlcQMBU, PixelC* ppxlcQMBV)
{
	Int maxVal = (1<<m_volmd.nBits)-1;

	CoordI ix, iy, ic = 0;
	for (iy = 0; iy < MB_SIZE; iy++) {
		for (ix = 0; ix < MB_SIZE; ix++, ic++) {
			ppxlcQMBY [ix] = m_rgiClipTab [ppxlcQMBY [ix] + m_ppxliFGSErrorMBY [ic]]; 
		}
		ppxlcQMBY += m_iFrameWidthY;
	}
	
	ic = 0;
	for (iy = 0; iy < BLOCK_SIZE; iy++) {
		for (ix = 0; ix < BLOCK_SIZE; ix++, ic++) {
			ppxlcQMBU [ix] = m_rgiClipTab [ppxlcQMBU [ix] + m_ppxliFGSErrorMBU [ic]];
			ppxlcQMBV [ix] = m_rgiClipTab [ppxlcQMBV [ix] + m_ppxliFGSErrorMBV [ic]];
		}
		ppxlcQMBU += m_iFrameWidthUV;
		ppxlcQMBV += m_iFrameWidthUV;
	}
}


Void CVideoObjectDecoder::copyFromPredToCurrQ (PixelC* ppxlcCurrQMBY, PixelC* ppxlcCurrQMBU, PixelC* ppxlcCurrQMBV)
{
	Int iUnit = sizeof(PixelC); // NBIT: for memcpy
	
	const PixelC* ppxlcPredMBY = m_ppxlcPredMBY;
	const PixelC* ppxlcPredMBU = m_ppxlcPredMBU;
	const PixelC* ppxlcPredMBV = m_ppxlcPredMBV;
	
	CoordI iY;
	for (iY = 0; iY < BLOCK_SIZE; iY++) {
		memcpy (ppxlcCurrQMBY, ppxlcPredMBY, MB_SIZE*iUnit);
		memcpy (ppxlcCurrQMBU, ppxlcPredMBU, BLOCK_SIZE*iUnit);
		memcpy (ppxlcCurrQMBV, ppxlcPredMBV, BLOCK_SIZE*iUnit);
		
		ppxlcCurrQMBY += m_iFrameWidthY; ppxlcPredMBY += MB_SIZE;
		ppxlcCurrQMBU += m_iFrameWidthUV; ppxlcPredMBU += BLOCK_SIZE;
		ppxlcCurrQMBV += m_iFrameWidthUV; ppxlcPredMBV += BLOCK_SIZE;
		
		memcpy (ppxlcCurrQMBY, ppxlcPredMBY, MB_SIZE*iUnit);
		ppxlcCurrQMBY += m_iFrameWidthY; ppxlcPredMBY += MB_SIZE;
	}
}


Void CVideoObjectDecoder::copyMBInFrame(PixelC* ppxlcYs, PixelC* ppxlcUs, PixelC* ppxlcVs, PixelC* ppxlcYd, PixelC* ppxlcUd, PixelC* ppxlcVd)
{
	CoordI iY;
	for (iY = 0; iY < BLOCK_SIZE; iY++) {
		memcpy (ppxlcYd, ppxlcYs, MB_SIZE*sizeof(PixelC));
		memcpy (ppxlcUd, ppxlcUs, BLOCK_SIZE*sizeof(PixelC));
		memcpy (ppxlcVd, ppxlcVs, BLOCK_SIZE*sizeof(PixelC));
		ppxlcYd += m_iFrameWidthY;
		ppxlcUd += m_iFrameWidthUV;
		ppxlcVd += m_iFrameWidthUV;
		ppxlcYs += m_iFrameWidthY;
		ppxlcUs += m_iFrameWidthUV;
		ppxlcVs += m_iFrameWidthUV;
		
		memcpy (ppxlcYd, ppxlcYs, MB_SIZE*sizeof(PixelC));
		ppxlcYd += m_iFrameWidthY; ppxlcYs += m_iFrameWidthY;
	}
}



Void CVideoObjectDecoder::averagePredToCurrQ (PixelC* ppxlcCurrQMBY, PixelC* ppxlcCurrQMBU, PixelC* ppxlcCurrQMBV)
{
	CoordI ix, iy, ic = 0;
	for (iy = 0; iy < MB_SIZE; iy++) {
		for (ix = 0; ix < MB_SIZE; ix++, ic++) {
			ppxlcCurrQMBY [ix] = m_rgiClipTab [
				((m_ppxlcPredMBY [ic] + m_ppxlcPredMBBackY [ic] + 1) >> 1) ];
		}
		ppxlcCurrQMBY += m_iFrameWidthY;
	}
	
	ic = 0;
	for (iy = 0; iy < BLOCK_SIZE; iy++) {
		for (ix = 0; ix < BLOCK_SIZE; ix++, ic++) {
			ppxlcCurrQMBU [ix] = m_rgiClipTab [
				((m_ppxlcPredMBU [ic] + m_ppxlcPredMBBackU [ic] + 1) >> 1)  ];
			ppxlcCurrQMBV [ix] = m_rgiClipTab [
				((m_ppxlcPredMBV [ic] + m_ppxlcPredMBBackV [ic] + 1) >> 1)  ];
		}
		ppxlcCurrQMBU += m_iFrameWidthUV;
		ppxlcCurrQMBV += m_iFrameWidthUV;
	}
}

Void CVideoObjectDecoder::copyFromPredToCurrQ_BVOP (CMBMode* pmbmd, PixelC* ppxlcCurrQMBY, PixelC* ppxlcCurrQMBU, PixelC* ppxlcCurrQMBV)
{
	switch ( pmbmd->m_mbType){
	case FORWARD:
		copyFromPredToCurrQ(ppxlcCurrQMBY, ppxlcCurrQMBU, ppxlcCurrQMBV);
		break;
	case BACKWARD:
		copyFromPredToCurrQ(ppxlcCurrQMBY, ppxlcCurrQMBU, ppxlcCurrQMBV);
		break;
	case DIRECT:
		averagePredToCurrQ (ppxlcCurrQMBY, ppxlcCurrQMBU, ppxlcCurrQMBV);
		break;
	case INTERPOLATE:
		averagePredToCurrQ (ppxlcCurrQMBY, ppxlcCurrQMBU, ppxlcCurrQMBV);
		break;
	}
	return;
}

Void
CVideoObjectDecoder::Recoverbp(Int m, Int *diff, Int *sign,Int *DCT_block )
{
	Int i;
	Int block[64];
	Int decoded_level, rec_value;
	
	decoded_level = enh_block_data[m].decoded_level;
	rec_value = (decoded_level>2) ? (1<<(decoded_level-3)) : 0;
	for(i=0;i<BLOCK_SQUARE_SIZE;i++){
		if (diff[i]) diff[i] += rec_value;
	}

	// added for FGS SE by FT CNET & PHILIPS Research US
	if (fgs_vop_selective_enhancement_enable)
		shift_bp_down(diff,pc_shift_factor[m/6]);
	// end SE

	// added for FGS FW by INTEL
	// for FPDAM4 by INTEL: 07/16/00
	if (m_volmd.bTemporalFGS)
	{
		if (m_volmd.bFGSTFreqWeightingEnable) {
			for(i=0;i<BLOCK_SQUARE_SIZE;i++)
				diff[i] >>= (m_volmd.rgiFGSTWeightMatrix[grgiStandardZigzag[i]]);
		}
	} 
	else 
	{
		if (m_volmd.bFGSFreqWeightingEnable) {
			for(i=0;i<BLOCK_SQUARE_SIZE;i++)
				diff[i] >>= (m_volmd.rgiWeightMatrix[grgiStandardZigzag[i]]);
		}
	}
	// end FW

#ifdef FGSPOSTCLIP
	for(i=0;i<BLOCK_SQUARE_SIZE;i++){
		block[i]=sign[i]*diff[i];
	}
#else
	for(i=0;i<BLOCK_SQUARE_SIZE;i++){
		block[i]=enh_block_data[m].base[i]+sign[i]*diff[i];
	}
#endif

	for(i=0;i<BLOCK_SQUARE_SIZE;i++)
	{
		DCT_block[i]=block[zigzag[i]];
	}
}

//
// Return value (eof) is not used by the caller
//
Int CVideoObjectDecoder::input_enh_bitstream()
{
//FGS3 : start
//	Int i,max_level,max_level_y,max_level_u,max_level_v;
	Int i,max_level,max_level_y,max_level_u,max_level_v,max_level_decode;
//FGS3 : end
	Int eof;
	char fgs_vop_bp_id;
	
	eof = 0;
	max_level_y = fgs_vop_max_level_y;
	max_level_u = fgs_vop_max_level_u;
	max_level_v = fgs_vop_max_level_v;
	
	for ( i=0; i<Nblock; i++ ) {
		char blocNum=i%6;
		enh_block_data[i].data_level = -1;
		if ( blocNum<4 ) {
			enh_block_data[i].decoded_level = max_level_y - 1;
		} else if ( blocNum==4 ) {
			enh_block_data[i].decoded_level = max_level_u - 1;
		} else {
			enh_block_data[i].decoded_level = max_level_v - 1;
		}
	}
	max_level = max_level_y;
	if ( max_level_u > max_level ) max_level = max_level_u;
	if ( max_level_v > max_level ) max_level = max_level_v;
	
	i=0;
	
//FGS3 :start
  	if (fgs_vop_number_of_vop_bp_coded > max_level)
        	max_level_decode = max_level;  
  	else
        	max_level_decode = fgs_vop_number_of_vop_bp_coded;
//	while ( i < max_level ) 
	while ( i < max_level_decode ) 
//FGS3 :end
	{
		// begin FPDAM4, Intel, 6/28/00
		int misalignement;
		m_pFGSbitstrmIn->peekBitsTillByteAlign(misalignement);
		m_pFGSbitstrmIn->getBits(misalignement); 

		UInt tmpvar = m_pFGSbitstrmIn->peekBits(NUMBITS_START_CODE_PREFIX);
		if (tmpvar == START_CODE_PREFIX) 
		{
			UInt FgsVopBpStartCode = ((START_CODE_PREFIX<<NUMBITS_FGS_VOP_BP_START_CODE_SUFFIX)|FGS_VOP_BP_START_CODE_SUFFIX);
			tmpvar = m_pFGSbitstrmIn->peekBits(NUMBITS_START_CODE_PREFIX+NUMBITS_FGS_VOP_BP_START_CODE_SUFFIX);
			if (tmpvar == FgsVopBpStartCode) 
			{
				tmpvar=m_pFGSbitstrmIn->getBits(NUMBITS_START_CODE_PREFIX+NUMBITS_FGS_VOP_BP_START_CODE_SUFFIX);
				fgs_vop_bp_id=m_pFGSbitstrmIn->getBits(NUMBITS_FGS_VOP_BP_ID);

				// FPDAM4 by INTEL: moved in the StartCode condition
				eof = input_one_bp(i,max_level,max_level_y,max_level_u,max_level_v);
				if (eof == EOF || FGSstop) 
					return EOF;
				i += 1;
			}
			else
			{
				FGSstop = TRUE;
				return EOF;
			}
		}
		else
		{
			FGSstop = TRUE;
			return EOF;
//			assert (FALSE);
		}

		/*
		UInt FgsVopBpStartCode = ((START_CODE_PREFIX<<NUMBITS_FGS_VOP_BP_START_CODE_SUFFIX)|FGS_VOP_BP_START_CODE_SUFFIX);
		UInt tmpvar = m_pFGSbitstrmIn->peekBits(NUMBITS_START_CODE_PREFIX+NUMBITS_FGS_VOP_BP_START_CODE_SUFFIX);
		if(tmpvar == FgsVopBpStartCode) {
			tmpvar=m_pFGSbitstrmIn->getBits(NUMBITS_START_CODE_PREFIX+NUMBITS_FGS_VOP_BP_START_CODE_SUFFIX);
			fgs_vop_bp_id=m_pFGSbitstrmIn->getBits(NUMBITS_FGS_VOP_BP_ID);

			// FPDAM4 by INTEL: moved in the StartCode condition
			eof = input_one_bp(i,max_level,max_level_y,max_level_u,max_level_v);
			if(eof == EOF || FGSstop) return EOF;
			i += 1;
		}
		//end BP_START_CODE
		*/
	}
	return(eof);
}

Int CVideoObjectDecoder::input_one_bp(Int level_index,Int max_level,
									  Int max_level_y,Int max_level_u,Int max_level_v)
{
	Int i, t, k, Nbits, start_decode;
	Int comb_index;
	Int m;
	Int eof;
	Int shift_overhead;
	Byte uc_NoneZeroMacroblock;
	// added for FGS FW by INTEL, 12/28/99
	Int fw_shift;
	fw_shift = m_volmd.rgiWeightMatrix[0];
	// end FW
	
	comb_index = get_comb_index(max_level_y,max_level_u,max_level_v,
		max_level-level_index);
	
	eof = 0;
	Nbits  = 0;
	i=0;
	while (i<Nblock ) {
		int mbNum=i/6;		
		char blocNum=i%6;		
		if ( blocNum<4 ) {
			start_decode = (level_index < (max_level-max_level_y)) ? 0 : 1;
		} else if ( blocNum==4 ) {
			start_decode = (level_index < (max_level-max_level_u)) ? 0 : 1;
		} else {
			start_decode = (level_index < (max_level-max_level_v)) ? 0 : 1;
		}

		if ( blocNum==0 && (level_index==0 || level_index==1) ) {
			Nbits += read_cbp_index(comb_index,level_index,i);
		}

		// Bug fixed by INTEL, 12/30/99: Need to set FGSstop here
		if(m_pFGSbitstrmIn->getStartCode()) {
			FGSstop=1;
			break;
		}
		
		//added for FGS SE by FT CNET & PHILIPS Research US
		if(fgs_vop_selective_enhancement_enable)
		{
			if(blocNum==0){
				if (level_index>1) /* not the first two MSB's */
					uc_NoneZeroMacroblock=1;
				else
				{
					uc_NoneZeroMacroblock=!(enh_block_data[i].cbp&&enh_block_data[i+1].cbp&&
						enh_block_data[i+2].cbp&&enh_block_data[i+3].cbp
						&&enh_block_data[i+4].cbp&&enh_block_data[i+5].cbp);
				}				
				if ((!pc_mb_shift_factor_received[mbNum])&&uc_NoneZeroMacroblock)
				{
					shift_overhead = decode_shift_factor(pc_shift_factor[mbNum]);
					pc_mb_shift_factor_received[mbNum]=1;
#ifdef _DEBUG
					cout<<(int)pc_shift_factor[mbNum]<<":"<<mbNum<<" ";
					cout.flush();
#endif
					i_total_shift_overhead+=shift_overhead;
					Nbits += shift_overhead;
				}
			}
		}
		//end SE 
				
		if ( start_decode==1 ) {
			if ( enh_block_data[i].data_level == -1 ) {
				if ( level_index>1 ) {
					t=(Int) m_pFGSbitstrmIn->getBits(1);
					Nbits += 1;
					enh_block_data[i].cbp=t;
				}
				if(enh_block_data[i].cbp==0){

					// added for FGS FW by Intel 11/23/99
					if ( m_volmd.bFGSFreqWeightingEnable  && fw_shift>=2 ) {
						if ( fw_shift>=4 ) {
							m = dec_one_block_bp(i,0,table_bpc4fw_h[0],table_symbol4fw_h[0]);
						} else {
							m = dec_one_block_bp(i,0,table_bpc4fw_m[0], table_symbol4fw_m[0]);
						}
					} else {
						m = dec_one_block_bp(i,0,table_bpc[0],table_symbol[0]);
					}
					// end FGS FW

					enh_block_data[i].data_level = max_level - level_index;	//Added by FT CNET
					if(m==EOF || FGSstop){
						enh_block_data[i].run[0][0]=63;
						enh_block_data[i].end[0][0]=0;
					}
					if(m == EOF) return(EOF);          //  abort bitstream decoding
					if(FGSstop) break;
					// enh_block_data[i].data_level = max_level - level_index;//Removed by FT CNET
					Nbits += m;
				}
			} else {
				t = enh_block_data[i].data_level - (max_level-level_index);
				k = ( t>3 ) ? 3 : t;

				// added for FGS FW by Intel 11/23/99
				if ( m_volmd.bFGSFreqWeightingEnable  && fw_shift>=2 ) {
					if ( fw_shift>=4 ) {
						m = dec_one_block_bp(i,t,table_bpc4fw_h[k],table_symbol4fw_h[k]);
					} else {
						m = dec_one_block_bp(i,t,table_bpc4fw_m[k], table_symbol4fw_m[k]);
					}
				} else {
					m = dec_one_block_bp(i,t,table_bpc[k],table_symbol[k]);
				}
				// end FGS FW

				if(m==EOF || FGSstop){
					enh_block_data[i].run[t][0]=63;
					enh_block_data[i].end[t][0]=0;
				}
				if(m == EOF) return(EOF);    // abort bitstream decoding
				if(FGSstop) break;
				
				Nbits += m;
			}
		}
		enh_block_data[i].decoded_level = max_level - level_index;

		// added by MSRCN for FGS error resilience, 2000/8/20
		if (!m_volmd.bFGSErrorResilienceDisable && i%6 == 5 && i != Nblock-1 )
		{
			Int misalignement;
			UInt iValue = m_pFGSbitstrmIn->peekBitsTillByteAlign(misalignement);
			UInt tmpvar = m_pFGSbitstrmIn->peekBits(NUMBITS_FGS_VP_MARKER + misalignement);
			tmpvar = (tmpvar & 0x007fffff);
			if (tmpvar == 1 && iValue == ifillValue[misalignement - 1])
			{
//FGS : start
//				m_pFGSbitstrmIn->flush(8);
				m_pFGSbitstrmIn->flush(misalignement);
//FGS : end
				// read FGS VP marker
				m_pFGSbitstrmIn->getBits(NUMBITS_FGS_VP_MARKER);
				// read MB index
				m_pFGSbitstrmIn->getBits(m_numBitsVPMBnum);
			}
		}
		i += 1;
	}

	return(Nbits);
}
	



Int CVideoObjectDecoder::dec_one_block_bp(Int mb_index,Int level,
										  Int *table_one_bp,Int *symbols)
{
	Int j, n, table_size, end, run, Nbits, mod_index, os, s;
	
	table_size=0;
	for ( j=0; j<3; j++ ) table_size += symbols[j];
	table_size += 1;   // add the escape symbol 
	
	n = 0;
	end=0;
	Nbits = 0;
	os = 0;
	
	while ( end==0 ) {
		if(m_pFGSbitstrmIn->getStartCode()){
			FGSstop=1;
			return(0);
		}
		
		mod_index = decode_symbol(table_one_bp,table_size);
		if(mod_index==EOF) return(EOF);
		if(FGSstop==1) return(0);
		if(m_pFGSbitstrmIn->fgseof()==EOF) return(EOF);
		
		Nbits += table_one_bp[2*mod_index+1];
		if ( mod_index<symbols[0] ) { // not end of plane 
			enh_block_data[mb_index].run[level][n] = mod_index;
			enh_block_data[mb_index].end[level][n] = 0;
			os += enh_block_data[mb_index].run[level][n]+1;
			n += 1;
			end = 0;
			
			if ( enh_block_data[mb_index].sign[os-1]==0 ) {
				s=(Int) m_pFGSbitstrmIn->getBits(1);
				if(m_pFGSbitstrmIn->fgseof()==EOF) return(EOF);
				
				enh_block_data[mb_index].sign[os-1] = (s==1)?-1:1;
				Nbits += 1;
			}
		} else if ( mod_index<symbols[0]+symbols[1] ) { // end of plane 
			enh_block_data[mb_index].run[level][n] = mod_index-symbols[0];
			enh_block_data[mb_index].end[level][n] = 1;
			os += enh_block_data[mb_index].run[level][n]+1;
			n += 1;
			end = 1;
			if ( enh_block_data[mb_index].sign[os-1]==0 ) {
				s=(Int) m_pFGSbitstrmIn->getBits(1);
				if(m_pFGSbitstrmIn->fgseof()==EOF) return(EOF);
				
				enh_block_data[mb_index].sign[os-1] = (s==1)?-1:1;
				Nbits += 1;
			}
		} else if ( mod_index<symbols[0]+symbols[1]+symbols[2] ) { // all zero 
			enh_block_data[mb_index].run[level][0] = 63;
			enh_block_data[mb_index].end[level][0] = 0;
			end = 1;
		} else { // escape case 
			run=(Int) m_pFGSbitstrmIn->getBits(6);
			if(m_pFGSbitstrmIn->fgseof()==EOF) return(EOF);
			end=(Int) m_pFGSbitstrmIn->getBits(1);
			if(m_pFGSbitstrmIn->fgseof()==EOF) return(EOF);
			
			enh_block_data[mb_index].run[level][n] = run;
			enh_block_data[mb_index].end[level][n] = end;
			Nbits += 7;
			os += run+1;
			n += 1;
			if ( enh_block_data[mb_index].sign[os-1]==0 ) {
				s=(Int) m_pFGSbitstrmIn->getBits(1);
				if(m_pFGSbitstrmIn->fgseof()==EOF) return(EOF);
				enh_block_data[mb_index].sign[os-1] = (s==1)?-1:1;
				Nbits += 1;
			}
		}
		if ( os>64 ) {
			printf("total length = %d > 64 !!!\n",os);
			assert (FALSE);			// FPDAM4 by INTEL: exit(1);
		}
//FGS2 : start
		if(m_pFGSbitstrmIn->getStartCode()){
			FGSstop=1;
		}
//FGS2 : end
	}
	return(Nbits);
}




Void CVideoObjectDecoder::FGSMBzero(Int mbnum,Int dct_type,Int mode)
{
	Int comp;
	
	for(comp=0;comp<6;comp++){
		enh_block_data[mbnum*6+comp].dct_type=dct_type;
		enh_block_data[mbnum*6+comp].mode=mode;
	}
	Nblock+=6;
}

Void CVideoObjectDecoder::FGSPutzeroBlock(Int dct_type,Int mode)
{
	enh_block_data[Nblock].dct_type=dct_type;
	enh_block_data[Nblock].mode=mode;
	Nblock++;
}

Void CVideoObjectDecoder::FGSPutBlock(Int *DCT_block, Int dct_type, Int mode, Int QuantType)
{
	Int i;
	
	enh_block_data[Nblock].dct_type=dct_type;
	enh_block_data[Nblock].mode=mode;
	for(i=0;i<64;i++)
		enh_block_data[Nblock].base[zigzag[i]]=DCT_block[i];
	Nblock++;
}

void CVideoObjectDecoder::bp2block(Int m, Int *diff, Int *sign)
{
	Int i, j, level, x[64];
	
	level = enh_block_data[m].data_level;
	for ( i=0; i<64; i++ ) diff[i] = 0;
	for ( j=0; j<level; j++ ) {
		one_bp2block(m, x,j);
		for ( i=0; i<64; i++ ) {
			diff[i] = (diff[i]<<1)+x[i];
		}
	}
	for ( i=0; i<64; i++ ) {
		sign[i] = enh_block_data[m].sign[i];
	}
}


Void CVideoObjectDecoder::one_bp2block(Int m, Int *out,Int level)
{
	Int i, j, n, run[64], os;
	
	if ( enh_block_data[m].run[level][0] == 63 &&
		enh_block_data[m].end[level][0] == 0 ) { /* all zero case */
		n = 0;
	} else {
		n = 0;
		while ( enh_block_data[m].end[level][n]==0 ) n += 1;
		n += 1;
	}
	for(i=0;i<n;i++){
		run[i] = enh_block_data[m].run[level][i];
	}
	if ( n==0 ) {
		for ( j=0; j<64; j++ ) out[j] = 0;
	} else {
		os = 0;
		for ( i=0; i<n; i++ ) {
			for ( j=0; j<run[i]; j++ ) out[os+j] = 0;
			out[os+j] = 1;
			os += run[i]+1;
		}
		for ( i=os; i<64; i++ ) out[i] = 0;
	}
}


Int CVideoObjectDecoder::decode_symbol(Int *vlc_table,Int size)
{
	int i, j, k;
	int n;
	int min, max;
	
	min=1;max=16;
	
	n=m_pFGSbitstrmIn->peekBits(max); 
	
	for(i=min;i<=max;i++){
		k=n>>(max-i);
		for(j=0;j<size;j++)
			if(k==vlc_table[2*j] && i==vlc_table[2*j+1]) break;
			if(j!=size) break;
	}
	if(i==max+1){
		FGSstop=1;
		return(0);
	}
	
	if(i<=8)
		m_pFGSbitstrmIn->getBits(i);
	else{
		m_pFGSbitstrmIn->getBits(8);
		if(m_pFGSbitstrmIn->getStartCode()){
			FGSstop=1;
			return(0);
		}
		m_pFGSbitstrmIn->getBits(i-8);
	}
	
	if(m_pFGSbitstrmIn->getStartCode()){
		FGSstop=1;
		return(0);
	}
	
	return(j);
}


Int CVideoObjectDecoder::get_comb_index(Int max_level_y,Int max_level_u,Int max_level_v,Int remain_level)
{
	Int comb_index;
	
	if ( max_level_y>=remain_level &&
		max_level_u>=remain_level &&
		max_level_v>=remain_level ) {
		comb_index = 0;
	} else if ( max_level_y>=remain_level &&
		max_level_u>=remain_level &&
		max_level_v<remain_level ) {
		comb_index = 1;
	} else if ( max_level_y>=remain_level &&
		max_level_u<remain_level &&
		max_level_v>=remain_level ) {
		comb_index = 2;
	} else if ( max_level_y>=remain_level &&
		max_level_u<remain_level &&
		max_level_v<remain_level ) {
		comb_index = 3;
	} else if ( max_level_y<remain_level &&
		max_level_u>=remain_level &&
		max_level_v>=remain_level ) {
		comb_index = 4;
	} else if ( max_level_y<remain_level &&
		max_level_u>=remain_level &&
		max_level_v<remain_level ) {
		comb_index = 5;
	} else if ( max_level_y<remain_level &&
		max_level_u<remain_level &&
		max_level_v>=remain_level ) {
		comb_index = 6;
	} else {
		comb_index = 7;
		printf("This cannot happen. Something is wrong!!!\n");
		exit(1);
	}
	return(comb_index);
}

Int CVideoObjectDecoder::read_cbp_index(Int comb_index,Int level_index,Int block_index)
{
	Int Nbits, b;
	Int cbp_table_3[4] = {7, 11, 13, 14};
	Int cbp_table_2[6] = {3, 5, 6, 9, 10, 12};
	Int cbp_table_1[4] = {8, 4, 2, 1};
	
	Nbits = 0;
	if ( level_index==0 ) {
		switch(comb_index) {
		case 0: b = (Int) m_pFGSbitstrmIn->getBits(1);
			Nbits += 1;
			if ( b==1 ) {
				use_code[block_index/6] = 1;
				enh_block_data[block_index].cbp = 1;
				enh_block_data[block_index+1].cbp = 1;
				enh_block_data[block_index+2].cbp = 1;
				enh_block_data[block_index+3].cbp = 1;
				enh_block_data[block_index+4].cbp = 1;
				enh_block_data[block_index+5].cbp = 1;
			} else {
				use_code[block_index/6] = 0;
				b = (Int) m_pFGSbitstrmIn->getBits(1);
				Nbits += 1;
				enh_block_data[block_index].cbp =
					(Int) m_pFGSbitstrmIn->getBits(1);
				enh_block_data[block_index+1].cbp =
					(Int) m_pFGSbitstrmIn->getBits(1);
				enh_block_data[block_index+2].cbp =
					(Int) m_pFGSbitstrmIn->getBits(1);
				enh_block_data[block_index+3].cbp =
					(Int) m_pFGSbitstrmIn->getBits(1);
				Nbits += 4;
				if ( b==1 ) {
					enh_block_data[block_index+4].cbp = 1;
					enh_block_data[block_index+5].cbp = 1;
				} else {
					b = (Int) m_pFGSbitstrmIn->getBits(2);
					Nbits += 2;
					enh_block_data[block_index+4].cbp = 0x1 & ((b-1)>>1);
					enh_block_data[block_index+5].cbp = 0x1 & (b-1);
				}
			}
			break;
		case 1: b = (Int) m_pFGSbitstrmIn->getBits(1);
			Nbits += 1;
			if ( b==1 ) {
				use_code[block_index/6] = 1;
				enh_block_data[block_index].cbp = 1;
				enh_block_data[block_index+1].cbp = 1;
				enh_block_data[block_index+2].cbp = 1;
				enh_block_data[block_index+3].cbp = 1;
				enh_block_data[block_index+4].cbp = 1;
			} else {
				use_code[block_index/6] = 0;
				b = (Int) m_pFGSbitstrmIn->getBits(5);
				Nbits += 5;
				enh_block_data[block_index].cbp = 0x1 & ((b-1)>>4);
				enh_block_data[block_index+1].cbp = 0x1 & ((b-1)>>3);
				enh_block_data[block_index+2].cbp = 0x1 & ((b-1)>>2);
				enh_block_data[block_index+3].cbp = 0x1 & ((b-1)>>1);
				enh_block_data[block_index+4].cbp = 0x1 & (b-1);
			}
			break;
		case 2: b = (Int) m_pFGSbitstrmIn->getBits(1);
			Nbits += 1;
			if ( b==1 ) {
				use_code[block_index/6] = 1;
				enh_block_data[block_index].cbp = 1;
				enh_block_data[block_index+1].cbp = 1;
				enh_block_data[block_index+2].cbp = 1;
				enh_block_data[block_index+3].cbp = 1;
				enh_block_data[block_index+5].cbp = 1;
			} else {
				use_code[block_index/6] = 0;
				b = (Int) m_pFGSbitstrmIn->getBits(5);
				Nbits += 5;
				enh_block_data[block_index].cbp = 0x1 & ((b-1)>>4);
				enh_block_data[block_index+1].cbp = 0x1 & ((b-1)>>3);
				enh_block_data[block_index+2].cbp = 0x1 & ((b-1)>>2);
				enh_block_data[block_index+3].cbp = 0x1 & ((b-1)>>1);
				enh_block_data[block_index+5].cbp = 0x1 & (b-1);
			}
			break;
		case 3: b = (Int) m_pFGSbitstrmIn->getBits(1);
			Nbits += 1;
			if ( b==1 ) {
				use_code[block_index/6] = 1;
				enh_block_data[block_index].cbp = 1;
				enh_block_data[block_index+1].cbp = 1;
				enh_block_data[block_index+2].cbp = 1;
				enh_block_data[block_index+3].cbp = 1;
			} else {
				use_code[block_index/6] = 0;
				b = (Int) m_pFGSbitstrmIn->getBits(1);
				Nbits += 1;
				if ( b==1 ) {
					b = (Int) m_pFGSbitstrmIn->getBits(2);
					Nbits += 2;
					enh_block_data[block_index].cbp = 0x1&(cbp_table_3[b]>>3);
					enh_block_data[block_index+1].cbp = 0x1&(cbp_table_3[b]>>2);
					enh_block_data[block_index+2].cbp = 0x1&(cbp_table_3[b]>>1);
					enh_block_data[block_index+3].cbp = 0x1&cbp_table_3[b];
				} else {
					b = (Int) m_pFGSbitstrmIn->getBits(1);
					Nbits += 1;
					if ( b==1 ) {
						b = (Int) m_pFGSbitstrmIn->getBits(3);
						Nbits += 3;
						enh_block_data[block_index].cbp = 0x1&(cbp_table_2[b]>>3);
						enh_block_data[block_index+1].cbp = 0x1&(cbp_table_2[b]>>2);
						enh_block_data[block_index+2].cbp = 0x1&(cbp_table_2[b]>>1);
						enh_block_data[block_index+3].cbp = 0x1&cbp_table_2[b];
					} else {
						b = (Int) m_pFGSbitstrmIn->getBits(1);
						Nbits += 1;
						if ( b==1 ) {
							b=(Int)m_pFGSbitstrmIn->getBits(2);
							Nbits += 2;
							enh_block_data[block_index].cbp = 0x1&(cbp_table_1[b]>>3);
							enh_block_data[block_index+1].cbp=0x1&(cbp_table_1[b]>>2);
							enh_block_data[block_index+2].cbp=0x1&(cbp_table_1[b]>>1);
							enh_block_data[block_index+3].cbp=0x1&cbp_table_1[b];
						} else {
							b=(Int)m_pFGSbitstrmIn->getBits(1);
							Nbits += 1;
							enh_block_data[block_index].cbp = 0;
							enh_block_data[block_index+1].cbp = 0;
							enh_block_data[block_index+2].cbp = 0;
							enh_block_data[block_index+3].cbp = 0;
						}
					}
				}
			}
			break;
		case 4: b = (Int) m_pFGSbitstrmIn->getBits(2);
			Nbits += 2;
			enh_block_data[block_index+4].cbp = 0x1 & (b>>1);
			enh_block_data[block_index+5].cbp = 0x1 & b;
			if ( b==3 ) {
				use_code[block_index/6] = 1;
			} else {
				use_code[block_index/6] = 0;
			}
			break;
		case 5: b = (Int) m_pFGSbitstrmIn->getBits(1);
			Nbits += 1;
			enh_block_data[block_index+4].cbp = b;
			use_code[block_index/6] = b;
			break;
		case 6: b = (Int) m_pFGSbitstrmIn->getBits(1);
			Nbits += 1;
			enh_block_data[block_index+5].cbp = b;
			use_code[block_index/6] = b;
			break;
		default: printf("Invalid combination!\n");
			exit(1);
    }
  } else if ( use_code[block_index/6]==1 ) {
	  switch(comb_index) {
      case 0: b = (Int) m_pFGSbitstrmIn->getBits(3);
		  Nbits += 3;
		  if ( b==0 ) {
			  b = (Int) m_pFGSbitstrmIn->getBits(6);
			  Nbits += 6;
			  enh_block_data[block_index].cbp = 0x1 & (b>>5);
			  enh_block_data[block_index+1].cbp = 0x1 & (b>>4);
			  enh_block_data[block_index+2].cbp = 0x1 & (b>>3);
			  enh_block_data[block_index+3].cbp = 0x1 & (b>>2);
			  enh_block_data[block_index+4].cbp = 0x1 & (b>>1);
			  enh_block_data[block_index+5].cbp = 0x1 & b;
		  } else if ( b==1 ) {
			  b = (Int) m_pFGSbitstrmIn->getBits(2);
			  Nbits += 2;
			  if ( b<2 ) {
                  enh_block_data[block_index].cbp = 0;
                  enh_block_data[block_index+1].cbp = 0;
                  enh_block_data[block_index+2].cbp = 0;
                  enh_block_data[block_index+3].cbp = 0;
			  } else {
                  enh_block_data[block_index].cbp = 1;
                  enh_block_data[block_index+1].cbp = 1;
                  enh_block_data[block_index+2].cbp = 1;
                  enh_block_data[block_index+3].cbp = 1;
			  }
			  enh_block_data[block_index+4].cbp = 0x1 & b;
			  enh_block_data[block_index+5].cbp = 1 - (0x1 & b);
		  } else if ( b==2 ) {
			  enh_block_data[block_index].cbp = 0;
			  enh_block_data[block_index+1].cbp = 0;
			  enh_block_data[block_index+2].cbp = 0;
			  enh_block_data[block_index+3].cbp = 0;
			  enh_block_data[block_index+4].cbp = 1;
			  enh_block_data[block_index+5].cbp = 1;
		  } else if ( b==3 ) {
			  enh_block_data[block_index].cbp = 1;
			  enh_block_data[block_index+1].cbp = 1;
			  enh_block_data[block_index+2].cbp = 1;
			  enh_block_data[block_index+3].cbp = 1;
			  enh_block_data[block_index+4].cbp = 1;
			  enh_block_data[block_index+5].cbp = 1;
		  } else {
			  b <<= 2;
			  b |= (Int) m_pFGSbitstrmIn->getBits(2);
			  Nbits += 2;
			  enh_block_data[block_index].cbp = 0x1 & (b>>3);
			  enh_block_data[block_index+1].cbp = 0x1 & (b>>2);
			  enh_block_data[block_index+2].cbp = 0x1 & (b>>1);
			  enh_block_data[block_index+3].cbp = 0x1 & b;
			  if ( b==16 || b==31 ) {
                  enh_block_data[block_index+4].cbp = 0;
                  enh_block_data[block_index+5].cbp = 0;
			  } else {
                  enh_block_data[block_index+4].cbp = 1;
                  enh_block_data[block_index+5].cbp = 1;
			  }
		  }
		  break;
      case 1: b = (Int) m_pFGSbitstrmIn->getBits(5);
		  Nbits += 5;
		  enh_block_data[block_index].cbp = 0x1 & (b>>4);
		  enh_block_data[block_index+1].cbp = 0x1 & (b>>3);
		  enh_block_data[block_index+2].cbp = 0x1 & (b>>2);
		  enh_block_data[block_index+3].cbp = 0x1 & (b>>1);
		  enh_block_data[block_index+4].cbp = 0x1 & b;
		  break;
      case 2: b = (Int) m_pFGSbitstrmIn->getBits(5);
		  Nbits += 5;
		  enh_block_data[block_index].cbp = 0x1 & (b>>4);
		  enh_block_data[block_index+1].cbp = 0x1 & (b>>3);
		  enh_block_data[block_index+2].cbp = 0x1 & (b>>2);
		  enh_block_data[block_index+3].cbp = 0x1 & (b>>1);
		  enh_block_data[block_index+5].cbp = 0x1 & b;
		  break;
      case 3: b = (Int) m_pFGSbitstrmIn->getBits(4);
		  Nbits += 4;
		  enh_block_data[block_index].cbp = 0x1 & (b>>3);
		  enh_block_data[block_index+1].cbp = 0x1 & (b>>2);
		  enh_block_data[block_index+2].cbp = 0x1 & (b>>1);
		  enh_block_data[block_index+3].cbp = 0x1 & b;
		  break;
      case 4: b = (Int) m_pFGSbitstrmIn->getBits(2);
		  Nbits += 2;
		  enh_block_data[block_index+4].cbp = 0x1 & (b>>1);
		  enh_block_data[block_index+5].cbp = 0x1 & b;
		  break;
      case 5: b = (Int) m_pFGSbitstrmIn->getBits(1);
		  Nbits += 1;
		  enh_block_data[block_index+4].cbp = b;
		  break;
      case 6: b = (Int) m_pFGSbitstrmIn->getBits(1);
		  Nbits += 1;
		  enh_block_data[block_index+5].cbp = b;
		  break;
      default: printf("Invalid combination!\n");
		  exit(1);
	  }
  } else {
	  switch(comb_index) {
      case 0: if ( enh_block_data[block_index].data_level == -1 ) {
		  b = (Int) m_pFGSbitstrmIn->getBits(1);
		  Nbits += 1;
		  enh_block_data[block_index].cbp = b;
              }
		  if ( enh_block_data[block_index+1].data_level == -1 ) {
			  b = (Int) m_pFGSbitstrmIn->getBits(1);
			  Nbits += 1;
			  enh_block_data[block_index+1].cbp = b;
		  }
		  if ( enh_block_data[block_index+2].data_level == -1 ) {
			  b = (Int) m_pFGSbitstrmIn->getBits(1);
			  Nbits += 1;
			  enh_block_data[block_index+2].cbp = b;
		  }
		  if ( enh_block_data[block_index+3].data_level == -1 ) {
			  b = (Int) m_pFGSbitstrmIn->getBits(1);
			  Nbits += 1;
			  enh_block_data[block_index+3].cbp = b;
		  }
		  if ( enh_block_data[block_index+4].data_level == -1 ) {
			  b = (Int) m_pFGSbitstrmIn->getBits(1);
			  Nbits += 1;
			  enh_block_data[block_index+4].cbp = b;
		  }
		  if ( enh_block_data[block_index+5].data_level == -1 ) {
			  b = (Int) m_pFGSbitstrmIn->getBits(1);
			  Nbits += 1;
			  enh_block_data[block_index+5].cbp = b;
		  }
		  break;
      case 1: if ( enh_block_data[block_index].data_level == -1 ) {
		  b = (Int) m_pFGSbitstrmIn->getBits(1);
		  Nbits += 1;
		  enh_block_data[block_index].cbp = b;
              }
		  if ( enh_block_data[block_index+1].data_level == -1 ) {
			  b = (Int) m_pFGSbitstrmIn->getBits(1);
			  Nbits += 1;
			  enh_block_data[block_index+1].cbp = b;
		  }
		  if ( enh_block_data[block_index+2].data_level == -1 ) {
			  b = (Int) m_pFGSbitstrmIn->getBits(1);
			  Nbits += 1;
			  enh_block_data[block_index+2].cbp = b;
		  }
		  if ( enh_block_data[block_index+3].data_level == -1 ) {
			  b = (Int) m_pFGSbitstrmIn->getBits(1);
			  Nbits += 1;
			  enh_block_data[block_index+3].cbp = b;
		  }
		  if ( enh_block_data[block_index+4].data_level == -1 ) {
			  b = (Int) m_pFGSbitstrmIn->getBits(1);
			  Nbits += 1;
			  enh_block_data[block_index+4].cbp = b;
		  }
		  break;
      case 2: if ( enh_block_data[block_index].data_level == -1 ) {
		  b = (Int) m_pFGSbitstrmIn->getBits(1);
		  Nbits += 1;
		  enh_block_data[block_index].cbp = b;
              }
		  if ( enh_block_data[block_index+1].data_level == -1 ) {
			  b = (Int) m_pFGSbitstrmIn->getBits(1);
			  Nbits += 1;
			  enh_block_data[block_index+1].cbp = b;
		  }
		  if ( enh_block_data[block_index+2].data_level == -1 ) {
			  b = (Int) m_pFGSbitstrmIn->getBits(1);
			  Nbits += 1;
			  enh_block_data[block_index+2].cbp = b;
		  }
		  if ( enh_block_data[block_index+3].data_level == -1 ) {
			  b = (Int) m_pFGSbitstrmIn->getBits(1);
			  Nbits += 1;
			  enh_block_data[block_index+3].cbp = b;
		  }
		  if ( enh_block_data[block_index+5].data_level == -1 ) {
			  b = (Int) m_pFGSbitstrmIn->getBits(1);
			  Nbits += 1;
			  enh_block_data[block_index+5].cbp = b;
		  }
		  break;
      case 3: if ( enh_block_data[block_index].data_level == -1 ) {
		  b = (Int) m_pFGSbitstrmIn->getBits(1);
		  Nbits += 1;
		  enh_block_data[block_index].cbp = b;
              }
		  if ( enh_block_data[block_index+1].data_level == -1 ) {
			  b = (Int) m_pFGSbitstrmIn->getBits(1);
			  Nbits += 1;
			  enh_block_data[block_index+1].cbp = b;
		  }
		  if ( enh_block_data[block_index+2].data_level == -1 ) {
			  b = (Int) m_pFGSbitstrmIn->getBits(1);
			  Nbits += 1;
			  enh_block_data[block_index+2].cbp = b;
		  }
		  if ( enh_block_data[block_index+3].data_level == -1 ) {
			  b = (Int) m_pFGSbitstrmIn->getBits(1);
			  Nbits += 1;
			  enh_block_data[block_index+3].cbp = b;
		  }
		  break;
      case 4: if ( enh_block_data[block_index+4].data_level == -1 ) {
		  b = (Int) m_pFGSbitstrmIn->getBits(1);
		  Nbits += 1;
		  enh_block_data[block_index+4].cbp = b;
              }
		  if ( enh_block_data[block_index+5].data_level == -1 ) {
			  b = (Int) m_pFGSbitstrmIn->getBits(1);
			  Nbits += 1;
			  enh_block_data[block_index+5].cbp = b;
		  }
		  break;
      case 5: if ( enh_block_data[block_index+4].data_level == -1 ) {
		  b = (Int) m_pFGSbitstrmIn->getBits(1);
		  Nbits += 1;
		  enh_block_data[block_index+4].cbp = b;
              }
		  break;
      case 6: if ( enh_block_data[block_index+5].data_level == -1 ) {
		  b = (Int) m_pFGSbitstrmIn->getBits(1);
		  Nbits += 1;
		  enh_block_data[block_index+5].cbp = b;
              }
		  break;
      default: printf("Invalid combination!\n");
		  exit(1);
    }
  }
  return(Nbits);
}

/* end of FGS (Optivision) */

//added for FGS SE by FT CNET & PHILIPS US Research
void CVideoObjectDecoder:: shift_bp_down(Int diff[], char sf)
{
	Int k;
	
	if (sf <=0 ) return;
#ifdef _DEBUG	
	if(sf>MAX_FGS_SE_SHIFT_FACTOR)
	{
		sf=MAX_FGS_SE_SHIFT_FACTOR;
		cerr<<"Max shift reached"<<endl;
	}
#endif
	for (k=0;k<64;k++)
		diff[k] >>= sf;
}

Int CVideoObjectDecoder::decode_shift_factor(char &rc_sf)
{
	Int length;
	Int next_bit;
	
	length=0;
	do
	{
		next_bit=(Int) m_pFGSbitstrmIn->getBits(1);
		length ++;
	}
	while (!next_bit);
	rc_sf=length-1;
	return length;
}
//end SE
