/******************************************************************************
 *                                                                          
 * This software module was originally developed by 
 *
 *    Weiping Li (Optivision, Inc.) and
 *    Fan Ling (General Instrument)
 *
 * and edited by
 *    Hong Jiang and Greg Thayer (Intel Corporation)
 *
 * in the course of development of the MPEG-4 Video (ISO/IEC 14496-2) standard.
 * This software module is an implementation of a part of one or more MPEG-4 
 * Video (ISO/IEC 14496-2) tools as specified by the MPEG-4 Video (ISO/IEC 
 * 14496-2) standard. 
 *
 * ISO/IEC gives users of the MPEG-4 Video (ISO/IEC 14496-2) standard free 
 * license to this software module or modifications thereof for use in hardware
 * or software products claiming conformance to the MPEG-4 Video (ISO/IEC 
 * 14496-2) standard. 
 *
 * Those intending to use this software module in hardware or software products
 * are advised that its use may infringe existing patents. The original 
 * developer of this software module and his/her company, the subsequent 
 * editors and their companies, and ISO/IEC have no liability for use of this 
 * software module or modifications thereof in an implementation. Copyright is 
 * not released for non MPEG-4 Video (ISO/IEC 14496-2) standard conforming 
 * products. 
 *
 * ACTS-MoMuSys partners retain full right to use the code for his/her own 
 * purpose, assign or donate the code to a third party and to inhibit third 
 * parties from using the code for non MPEG-4 Video (ISO/IEC 14496-2) standard
 * conforming products. This copyright notice must be included in all copies or
 * derivative works. 
 *
 * Copyright (c) 1997
 *
 * Revision History:
 * 
 *  10/29/99: FGS WD2 VOL header from Intel by Hong Jiang, Greg Thayer
 *  12/15/99: FGS SE (Selective Enhancement) from FT CNET by Bertrand THEBAULT
 *  12/21/99: FGS FW (Frequency Weighting) from INTEL by Greg Thayer, Hong Jiang
 *  08/15/00: FGS FPDAM4 from INTEL by Hong Jiang
 *
 * *****************************************************************************/


/************************    INCLUDE FILES    ********************************/
#include <stdio.h>
#include <fstream.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "vlcbp.hpp"
#include "typeapi.h"
#include "codehead.h"
#include "global.hpp"
#include "entropy/bitstrm.hpp"
#include "entropy/entropy.hpp"
#include "entropy/huffman.hpp"
#include "mode.hpp"
#include "dct.hpp"
#include "vopses.hpp"
#include "vopseenc.hpp"

//FGS : start
//#include "sys/vlcbp4fw.hpp" // added by INTEL for FGS FW
#include "vlcbp4fw.hpp" // added by INTEL for FGS FW
//FGS : end

#ifdef __MFC_
#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

#define new DEBUG_NEW
#endif // __MFC_

#   define  ABS(x)                (((x) < 0) ? -(x) : (x))
#   define  SIGN(x)               (((x) < 0) ? -1 : 1)

const Int zigzag[64] = {
	0, 1, 5, 6,14,15,27,28,
		2, 4, 7,13,16,26,29,42,
		3, 8,12,17,25,30,41,43,
		9,11,18,24,31,40,44,53,
		10,19,23,32,39,45,52,54,
		20,22,33,38,46,51,55,60,
		21,34,37,47,50,56,59,61,
		35,36,48,49,57,58,62,63
};

/***********************************************************
*
* -- CodeFGSVop -- Coding enhancement VOP of FGS
*
* Author :
*	Weiping Li, Optivision
* Modified by: 
*	Fan Ling, General Instrument
*
* Created :
* 	12-18-98
*
* Purpose :
* 
* Arguments in : 
*
* Arguments in/out :
*
* Arguments out :
*
* Return values :
* none
*
* Side effects :
* -
*
* Description :
* -
*
* See also :
*
* Modified :
*
***********************************************************/

Void CVideoObjectEncoder::codeFGSVOHead ()
{
	m_pbitstrmOutFGS -> putBits (START_CODE_PREFIX, NUMBITS_START_CODE_PREFIX, "VO_Start_Code");
	m_pbitstrmOutFGS -> putBits (VO_START_CODE, NUMBITS_VO_START_CODE, "VO_Start_Code");		//plus 3 bits
	m_pbitstrmOutFGS -> putBits (m_uiVOId, NUMBITS_VO_ID, "VO_Id");
	UInt uiNumBits = NUMBITS_START_CODE_PREFIX + NUMBITS_VO_START_CODE + NUMBITS_VO_ID;;
	m_statsFGSVOL.nBitsHead += uiNumBits;
	cout << "FGS VO Overhead" << "\t\t" << uiNumBits << "\n\n";
	cout.flush ();
}


// FGS VOL header conform to FGS WD 2.0
// Intel 10/29/99
// modified to conform to FPDAM4
// Intel 08/12/00
Void CVideoObjectEncoder::codeFGSVOLHead (Int iSessionWidth, Int iSessionHeight)
{
	Int bits;

	m_pbitstrmOutFGS -> putBits (START_CODE_PREFIX, NUMBITS_START_CODE_PREFIX, "VOL_Start_Code_Prefix");
	m_pbitstrmOutFGS -> putBits (VOL_START_CODE, NUMBITS_VOL_START_CODE, "VOL_Start_Code");		//plus 4 bits
	m_pbitstrmOutFGS -> putBits ((Int) 1, NUMBITS_VOL_ID, "VOL_Id"); // by katata
	m_statsFGSVOL.nBitsHead += NUMBITS_START_CODE_PREFIX + NUMBITS_VOL_START_CODE+NUMBITS_VOL_ID;

	m_pbitstrmOutFGS -> putBits ((Int) 1, 1, "VOL_Random_Access");		//isn't this a system level flg?
	m_statsFGSVOL.nBitsHead++;

	m_pbitstrmOutFGS -> putBits ((Int) VIDEO_OBJECT_TYPE_STREAMING_VIDEO, 8, "VOL_Type_Indicator"); // Set to indicate FGS profile.
	m_statsFGSVOL.nBitsHead+=8;
	m_pbitstrmOutFGS -> putBits ((Int) FGS_LAYER_TYPE_FGS, NUMBITS_FGS_LAYER_TYPE, "fgs_layer_type"); // Set to indicate FGS profile.
	m_statsFGSVOL.nBitsHead+=NUMBITS_FGS_LAYER_TYPE;

	m_pbitstrmOutFGS -> putBits ((Int) 2, 3, "VOL_Priority"); // VOL_priority is set to 2 for FGS and 1 for base (supposely)
	m_statsFGSVOL.nBitsHead+=3;

	m_pbitstrmOutFGS -> putBits ((Int) 1, 4, "aspect_ratio_info"); // square pix
	m_statsFGSVOL.nBitsHead+=4;
	// assuming no extended aspect ration info here.
	
	m_pbitstrmOutFGS -> putBits (m_volmd.uiVolControlParameters, 1, "VOL_Control_Parameters"); //useless flag for now
	m_statsFGSVOL.nBitsHead++;
	
	if(m_volmd.uiVolControlParameters)
	{
		m_pbitstrmOutFGS -> putBits (m_volmd.uiChromaFormat, 2, "Chroma_Format");
		m_pbitstrmOutFGS -> putBits (m_volmd.uiLowDelay, 1, "Low_Delay");
		m_statsFGSVOL.nBitsHead += 2 + 1;
	}

	m_pbitstrmOutFGS -> putBits (1, 1, "Marker");
	m_statsFGSVOL.nBitsStuffing ++;
	m_pbitstrmOutFGS -> putBits (m_volmd.iClockRate, NUMBITS_TIME_RESOLUTION, "VOL_Time_Increment_Resolution"); 
	m_pbitstrmOutFGS -> putBits (1, 1, "Marker");
	m_statsFGSVOL.nBitsStuffing ++;
	m_pbitstrmOutFGS -> putBits (1, 1, "VOL_Fixed_Vop_Rate");
	m_statsFGSVOL.nBitsHead += NUMBITS_TIME_RESOLUTION + 1;
	if (1) { //fixed VOP rate
		bits = (Int) ceil(log((double)m_volmd.iClockRate)/log(2.0));
		m_pbitstrmOutFGS -> putBits (m_volmd.iTemporalRate, bits, "VOL_Fixed_Vop_Time_Increment");	//INTEL
		m_statsFGSVOL.nBitsHead += bits;
	}

	m_pbitstrmOutFGS -> putBits (MARKER_BIT, 1, "Marker_Bit");
	m_statsFGSVOL.nBitsStuffing ++;
	m_pbitstrmOutFGS -> putBits (iSessionWidth, NUMBITS_VOP_WIDTH, "VOL_Width"); 
	m_pbitstrmOutFGS -> putBits (MARKER_BIT, 1, "Marker_Bit");
	m_statsFGSVOL.nBitsStuffing ++;
	m_pbitstrmOutFGS -> putBits (iSessionHeight, NUMBITS_VOP_HEIGHT, "VOL_Height"); 
	m_pbitstrmOutFGS -> putBits (MARKER_BIT, 1, "Marker_Bit");
	m_statsFGSVOL.nBitsStuffing ++;
	m_statsFGSVOL.nBitsHead += NUMBITS_VOP_WIDTH + NUMBITS_VOP_HEIGHT;

	m_pbitstrmOutFGS -> putBits (m_vopmd.bInterlace, (UInt) 1, "VOL_interlace");
	m_statsFGSVOL.nBitsHead++;

	// Loading FGS Frequency Weighting Matrix for FGS
	if (1) 
	{
		m_pbitstrmOutFGS -> putBits (m_volmd.bFGSFreqWeightingEnable, 1, "VOL_FGSFreqWeightingEnable");
		m_statsFGSVOL.nBitsHead++;
		if (m_volmd.bFGSFreqWeightingEnable) {
			m_pbitstrmOutFGS -> putBits (m_volmd.bLoadWeightMatrix, 1, "VOL_Load_FGS_Freq_Weight_Matrix");
			m_statsFGSVOL.nBitsHead++;

			if (m_volmd.bLoadWeightMatrix) {
				int val, i=0;
				do {
					val = m_volmd.rgiWeightMatrix[grgiStandardZigzag[i]];
					m_pbitstrmOutFGS -> putBits (val, 3);
					m_statsFGSVOL.nBitsHead += 3;
					i++;
				} while ((i < BLOCK_SQUARE_SIZE) && (val!=0));
			}
		}
	}

//FGS3 : start
	m_pbitstrmOutFGS -> putBits (0, 1, "VOL_Quarter_sample");
	m_statsFGSVOL.nBitsHead ++;
//FGS3 : end

	// modified by Wu Feng for FGS error resilience, MSRCN, 2000/8/18
	//m_pbitstrmOutFGS -> putBits (0, 1, "VOL_resync_marker_disable");
	m_pbitstrmOutFGS -> putBits (m_volmd.bFGSErrorResilienceDisable, 1, "VOL_resync_marker_disable");
	m_statsFGSVOL.nBitsHead ++;

	// added by Hong Jiang, INTEL, 8/12/00
	m_pbitstrmOutFGS -> flush();

	// added by Hong Jiang, INTEL, 8/8/00
	cout << "FGS VOL Total" << "\t\t" << m_pbitstrmOutFGS -> getCounter() << "\n\n";
	cout.flush ();
}


// added for FGS SE by FT CNET
Void CVideoObjectEncoder::initFGSVOP() 
{
	process_vop_shift_factors(); 
}
//end SE

// added for FPDAM4 by INTEL
Int CVideoObjectEncoder::encodeFGSVOPHead() 
{	
	Int Nbits = 0;

	// changed by Hong Jiang, INTEL, 8/8/00: from reset() to flush()
	m_pbitstrmOutFGS -> reset(); 	

	m_pbitstrmOutFGS -> putBits (START_CODE_PREFIX, NUMBITS_START_CODE_PREFIX, "FGS_VOP_Start_Code");
	m_pbitstrmOutFGS -> putBits (FGS_VOP_START_CODE, NUMBITS_VOP_START_CODE, "FGS_VOP_Start_Code");
	m_pbitstrmOutFGS -> putBits (fgs_vop_coding_type, NUMBITS_FGS_VOP_CODING_TYPE, "FGS_VOP_Coding_Type");
	Nbits += NUMBITS_START_CODE_PREFIX+NUMBITS_VOP_START_CODE+NUMBITS_FGS_VOP_CODING_TYPE;

	// Time reference
	m_pbitstrmOutFGS -> putBits ((Int) 0xFFFFFFFE, (UInt) m_nBitsModuloBase + 1, "VOP_Modulo_Time_Base");
	m_pbitstrmOutFGS->putBits ((Int) 1, 1, "Marker");		// marker bit    Added for error resilient mode by Toshiba(1997-11-14)
	m_pbitstrmOutFGS -> putBits (m_iVopTimeIncr, m_iNumBitsTimeIncr, "VOP_Time_Incr");
	m_pbitstrmOutFGS->putBits ((Int) 1, 1, "Marker");		// marker bit
	Nbits += m_nBitsModuloBase + 1 + 1 + m_iNumBitsTimeIncr + 1;
	
    m_pbitstrmOutFGS->putBits(fgs_vop_max_level_y,5,"FGS_VOP_Max_level_Y");
    Nbits += 5;
    m_pbitstrmOutFGS->putBits(fgs_vop_max_level_u,5,"FGS_VOP_Max_level_u");
    Nbits += 5;
    m_pbitstrmOutFGS->putBits(fgs_vop_max_level_v,5,"FGS_VOP_Max_level_V");
    Nbits += 5;

//FGS3 : start
    m_pbitstrmOutFGS->putBits ((Int) 1, 1, "Marker");// marker bit
    m_pbitstrmOutFGS->putBits(m_fgs_vop_number_of_vop_bp_coded,5,"FGS_VOP_Number_Of_Vop_BP_Coded");
    Nbits += 1+5;

//FGS3 : start
    m_pbitstrmOutFGS->putBits(fgs_vop_mc_bit_plane_not_used,5,"FGS_VOP_MC_BIT_PLANE_NOT_USED");
    Nbits += 5;
//    m_pbitstrmOutFGS->putBits(fgs_vop_mc_bit_plane_not_used,4,"FGS_VOP_MC_BIT_PLANE_NOT_USED");
//    Nbits += 4;
//FGS3 : end
    m_pbitstrmOutFGS->putBits(m_fgs_selective_enhancement_enable,1,"FGS_Selective_Enhancement_Enable");
    Nbits += 1;

	return (Nbits);
}
	
// rewrote for FPDAM4, Intel, 6/28/00
Void CVideoObjectEncoder::encodeFGSVOP()
{
	Int Nbits = 0;
	Int i;
//FGS3 : start
	Int max_level;
//FGS3 : end

	if(m_fgs_selective_enhancement_enable)
	{
		process_vop_shift_factors();
		i_total_shift_overhead=0;
		for ( i=0; i<Nblock; i++ )
		{
			pc_mb_shift_factor_reached[i/6]=0;
			enh_block_data[i].cbp=1;    
		}
	}
	
	find_max_level(&fgs_vop_max_level_y,
		&fgs_vop_max_level_u,
		&fgs_vop_max_level_v);
	fgs_vop_max_level = fgs_vop_max_level_y;
	if ( fgs_vop_max_level_u > fgs_vop_max_level )
		fgs_vop_max_level = fgs_vop_max_level_u;
	if ( fgs_vop_max_level_v > fgs_vop_max_level )
		fgs_vop_max_level = fgs_vop_max_level_v;

    fgs_vop_mc_bit_plane_not_used=fgs_vop_max_level;

	if (m_bFGSScalability) 
	{
		fgs_vop_coding_type = FGS_VOP_CODING_TYPE_I;
	} 
	else if (m_volmd.bTemporalFGS) 
	{
		if (m_vopmd.vopPredType == PVOP)
			fgs_vop_coding_type = FGS_VOP_CODING_TYPE_P;
		else if (m_vopmd.vopPredType == BVOP)
			fgs_vop_coding_type = FGS_VOP_CODING_TYPE_B;
		else
			assert(0);		// FGS_FGST is not implemented yet, INTEL
	}

	if (m_bFGSScalability)
	{
		Nbits += encodeFGSVOPHead();
	}							// For FGST, the header including MV are code already
	else						// But has to update the info
	{
		m_pbitstrmOutFGS->putBitsFGS(fgs_vop_max_level_y,5,"FGS_VOP_Max_level_Y");
		m_pbitstrmOutFGS->putBitsFGS(fgs_vop_max_level_u,5,"FGS_VOP_Max_level_u");
		m_pbitstrmOutFGS->putBitsFGS(fgs_vop_max_level_v,5,"FGS_VOP_Max_level_V");

//FGS3 : start
    		m_pbitstrmOutFGS->putBitsFGS((Int) 1, 1, "Marker");// marker bit
    		m_pbitstrmOutFGS->putBitsFGS(m_fgs_vop_number_of_vop_bp_coded,5,"FGS_VOP_Number_Of_Vop_BP_Coded");

		m_pbitstrmOutFGS->putBitsFGS(fgs_vop_mc_bit_plane_not_used,5,"FGS_VOP_MC_BIT_PLANE_NOT_USED");
//		m_pbitstrmOutFGS->putBitsFGS(fgs_vop_mc_bit_plane_not_used,4,"FGS_VOP_MC_BIT_PLANE_NOT_USED");
//FGS3 : end
		m_pbitstrmOutFGS->putBitsFGS(m_fgs_selective_enhancement_enable,1,"FGS_Selective_Enhancement_Enable");
		m_pbitstrmOutFGS->flushFGS();
	}

    for ( i=0; i<Nblock; i++ ) enh_block_data[i].cbp=1;    
	
	// added by MSRCN for FGS error resilience, 2000/8/20
	assert(m_iNumMBX>0 && m_iNumMBY>0);
	m_numBitsVPMBnum = (int) ceil( log10((double)(m_iNumMBX * m_iNumMBY)) / log10(2.0));

//FGS3 : start
  	if (m_fgs_vop_number_of_vop_bp_coded > fgs_vop_max_level)
        	max_level = fgs_vop_max_level;  
  	else
        	max_level = m_fgs_vop_number_of_vop_bp_coded;
    for ( i=0; i<max_level-2; i++ ) {
//    for ( i=0; i<fgs_vop_max_level-2; i++ ) {
//FGS3 : end
		// begin FPDAM4, Intel, 6/28/00
		// FGS_VOP_BP_START_CODE at the beginning of all bitplanes 
		Nbits += m_pbitstrmOutFGS->flush();

		m_pbitstrmOutFGS->putBits((START_CODE_PREFIX<<NUMBITS_FGS_VOP_BP_START_CODE_SUFFIX)|FGS_VOP_BP_START_CODE_SUFFIX,
			NUMBITS_START_CODE_PREFIX+NUMBITS_FGS_VOP_BP_START_CODE_SUFFIX,
			"FGS_VOP_BP_START_CODE_SUFFIX");
		m_pbitstrmOutFGS->putBits(i,NUMBITS_FGS_VOP_BP_ID,"FGS_VOP_BP_ID");
		Nbits+=32;
		// end  FPDAM4, Intel, 6/28/00

		Nbits += output_one_bp(i,fgs_vop_max_level,
			fgs_vop_max_level_y,
			fgs_vop_max_level_u,
			fgs_vop_max_level_v);
    }
	
	fgs_vop_bytes = Nbits/8 + 1;
	
	printf("\nfgs_vop_bytes = %d\n",fgs_vop_bytes);
	if(m_fgs_selective_enhancement_enable)
	{
		cout<<"SE bytes overhead ="<<(i_total_shift_overhead+7)/8<<endl;
		cout.flush();
	}
	m_pbitstrmOutFGS->flush();
	Nblock=0; 

	return;
}

Void CVideoObjectEncoder::FGSencodeBVOPMB_Skipped (
												   CMBMode* pmbmd, const CMotionVector* pmv, const CMotionVector* pmvBackward,
												   const CMBMode* pmbmdRef, const CMotionVector* pmvRef,
												   Int iMBX, Int iMBY,
												   CoordI x, CoordI y,
												   Int iMBNum
												   )
{	
	motionCompAndDiff_BVOP_MB (pmv, pmvBackward, pmbmd, x, y, &m_rctRefVOPY0 , &m_rctRefVOPY1);


	// added for FPDAM4 by INTEL
	if (m_volmd.bTemporalFGS)	// code the skip code
	{
		assert (pmbmd -> m_bSkip);
		codeMBTextureHeadOfBVOP (pmbmd);
	}
	//	quantizeTextureInterMB (pmbmd, pmv, NULL, bSkip); 
	
	Int diff[64], sign[64];
	Int* rgiBlkCurrQ;
	Int iWidthCurrQ;
	Int iBlkEnd = V_BLOCK;
	
	for (UInt iBlk = (UInt) Y_BLOCK1; iBlk <= (UInt)iBlkEnd; iBlk++) {
		if (iBlk < (UInt) U_BLOCK || iBlk> (UInt) V_BLOCK) {
			switch (iBlk)
			{
			case (Y_BLOCK1):
				rgiBlkCurrQ = m_ppxliErrorMBY;
				break;
			case (Y_BLOCK2):
				rgiBlkCurrQ = m_ppxliErrorMBY + BLOCK_SIZE;
				break;
			case (Y_BLOCK3):
				rgiBlkCurrQ = m_ppxliErrorMBY + MB_SIZE * BLOCK_SIZE;
				break;
			case (Y_BLOCK4):
				rgiBlkCurrQ = m_ppxliErrorMBY + MB_SIZE * BLOCK_SIZE + BLOCK_SIZE;
				break;
			}
			iWidthCurrQ = MB_SIZE;
		}
		else {
			iWidthCurrQ = BLOCK_SIZE;
			rgiBlkCurrQ = (iBlk == U_BLOCK) ? m_ppxliErrorMBU: m_ppxliErrorMBV;
		}
		m_pfdct->apply  ((PixelI* )rgiBlkCurrQ, iWidthCurrQ, m_rgiDCTcoef, BLOCK_SIZE);
		for (Int i = 0; i < BLOCK_SQUARE_SIZE; i++) {
			diff[i] = ABS(m_rgiDCTcoef [i]);
			sign[i] = (diff[i]==0)?0:SIGN(m_rgiDCTcoef[i]);
		}
		//added for FGS SE by FT CNET 
		
		if (m_fgs_selective_enhancement_enable)
		{
			shift_bp_up(diff,pc_shift_factor[iMBNum]);
		}
		//end SE 

		// Residue Weighting, Intel, 10/25/99
		if (m_volmd.bFGSFreqWeightingEnable) {
			for (Int i = 0; i < BLOCK_SQUARE_SIZE; i++)
				diff[i] = diff[i] << (m_volmd.rgiWeightMatrix[i]);
		}

		block2bp(diff,sign);
	}
}



Void CVideoObjectEncoder::find_max_level(Int *mx_y, Int *mx_u, Int *mx_v)
{
	Int i;
	
	*mx_y = *mx_u = *mx_v = 0;
	for ( i=0; i<Nblock; i++ ) {
		char blocNum=i%6;
		if ( blocNum<4 ) {
			if ( enh_block_data[i].data_level > *mx_y )
				*mx_y = enh_block_data[i].data_level;
		} else if ( blocNum==4 ) {
			if ( enh_block_data[i].data_level > *mx_u )
				*mx_u = enh_block_data[i].data_level;
		} else {
			if ( enh_block_data[i].data_level > *mx_v )
				*mx_v = enh_block_data[i].data_level;
		}
	}
}



Int CVideoObjectEncoder::output_one_bp(Int level_index,Int max_level,
									   Int max_level_y,Int max_level_u,Int max_level_v)
{
	Int i, j, t, k, Nbits, start_encode;
	Int comb_index, cbp_index;
	int shift_factor_vlc_code,shift_factor_vlc_length;
	Byte uc_NoneZeroMacroblock;
	Int fw_shift; // added by Intel for FGS FW
	Int iVPCounter;  // added by MSRCN for FGS error resilience, 2000/8/18

	fw_shift = m_volmd.rgiWeightMatrix[0];

	comb_index = get_comb_index(max_level_y,max_level_u,max_level_v,
		max_level-level_index);
	
	Nbits  = 0;
	iVPCounter = 0;  // added by MSRCN for FGS error resilience, 2000/8/18
	for ( i=0; i<Nblock; i++ ) {
		int mbNum=i/6;		
		char blocNum=i%6;		
		if((mbNum>=96)&&(level_index>=1))
				cout<<"";
		if ( blocNum<4 ) {
			start_encode = (level_index < (max_level-max_level_y)) ? 0 : 1;
		} else if ( blocNum==4 ) {
			start_encode = (level_index < (max_level-max_level_u)) ? 0 : 1;
		} else {
			start_encode = (level_index < (max_level-max_level_v)) ? 0 : 1;
		}
		if ( blocNum==0 && (level_index==0 || level_index==1) ) {
			cbp_index = get_cbp_index(comb_index,i,max_level-level_index);
			Nbits += write_cbp_index(comb_index,level_index,i,cbp_index,
				max_level-level_index);
		}
		//added for FGS SE by FT CNET & Philips Research US
		if(m_fgs_selective_enhancement_enable)
		{
			if(blocNum==0){
				if (level_index>1) /* not the first two MSB's */
					uc_NoneZeroMacroblock=1;
				else
				{
					uc_NoneZeroMacroblock=!(enh_block_data[i].cbp&&enh_block_data[i+1].cbp&&
						enh_block_data[i+2].cbp&&enh_block_data[i+3].cbp
						&&enh_block_data[i+4].cbp&&enh_block_data[i+5].cbp);
				}
				if ((!pc_mb_shift_factor_reached[mbNum])&&uc_NoneZeroMacroblock)
				{
					get_shift_factor_vlc(pc_shift_factor[mbNum],shift_factor_vlc_code,shift_factor_vlc_length);
					m_pbitstrmOutFGS->putBits(shift_factor_vlc_code,shift_factor_vlc_length,"Shift factor VLC");
					pc_mb_shift_factor_reached[mbNum]=1;
#ifdef _DEBUG
					cout<<(int)pc_shift_factor[mbNum]<<":"<<mbNum<<" ";
					cout.flush();
#endif
					i_total_shift_overhead+=shift_factor_vlc_length;
					Nbits += shift_factor_vlc_length;
				}
			}
		}
		//end SE 
		if ( start_encode==1 ) {
			t = enh_block_data[i].data_level - (max_level-level_index);
			j=(t>=0?1:0);
			if(enh_block_data[i].cbp!=0 && level_index>1){
				m_pbitstrmOutFGS->putBits(1-j,1,"FLAG");
				enh_block_data[i].cbp=1-j;
				Nbits += 1;
			}
			if(enh_block_data[i].cbp==0){
				k = ( t>3 ) ? 3 : t;
				// added for FGS FW by INTEL
				if ( m_volmd.bFGSFreqWeightingEnable && fw_shift>=2 ) {
					if ( fw_shift>=4 ) {
						Nbits += enc_one_block_bp(i,t,table_bpc4fw_h[k],table_symbol4fw_h[k]);
					} else {
						Nbits += enc_one_block_bp(i,t,table_bpc4fw_m[k],table_symbol4fw_m[k]);
					}
				//end FW
				} else {
					Nbits += enc_one_block_bp(i,t,table_bpc[k],table_symbol[k]);
				}
			}
		}
		// added by MSRCN for FGS error resilience, 2000/8/20
		if (!m_volmd.bFGSErrorResilienceDisable && i%6 == 5 && i != Nblock-1)
		{
			if (Nbits - iVPCounter > m_volmd.iFGSVideoPacketLength)
			{
				iVPCounter = Nbits;
				m_pbitstrmOutFGS -> flush();
				// VP marker
				m_pbitstrmOutFGS -> putBits(FGS_VP_MARKER, NUMBITS_FGS_VP_MARKER, "VP header");
				// MB index
				m_pbitstrmOutFGS -> putBits((i+1)/6, m_numBitsVPMBnum, "VP header");
			}
		}
	}
	return(Nbits);
}


Int CVideoObjectEncoder::enc_one_block_bp(Int mb_index,Int level,Int *table_one_bp,Int *symbols)
{
	Int mod_index, Nbits, i, os, ind, n;
	
	Nbits = 0;
	n = 0;
	if ( enh_block_data[mb_index].run[level][0]==63 &&
		enh_block_data[mb_index].end[level][0]==0) {
		/* all zero case */
		mod_index = symbols[0]+symbols[1];
		m_pbitstrmOutFGS->putBits(table_one_bp[2*mod_index],
			table_one_bp[2*mod_index+1]);
		Nbits += table_one_bp[2*mod_index+1];
	} else {
		while ( enh_block_data[mb_index].end[level][n]==0 ) n += 1;
		n += 1;
	}
	os = 0;
	for(i=0;i<n;i++){
		ind = enh_block_data[mb_index].run[level][i];
		if ( i!=n-1 ) {
			mod_index=(ind<symbols[0]?ind:-1);
		} else {
			mod_index=(ind<symbols[1]?ind+symbols[0]:-1);
		}
		
		if(mod_index!=-1) {
			m_pbitstrmOutFGS->putBits(table_one_bp[2*mod_index],
				table_one_bp[2*mod_index+1]);
			Nbits += table_one_bp[2*mod_index+1];
		} else { /* escape case */
			mod_index=symbols[0]+symbols[1]+symbols[2];
			m_pbitstrmOutFGS->putBits(table_one_bp[2*mod_index],
				table_one_bp[2*mod_index+1],"Escape");
			m_pbitstrmOutFGS->putBits(enh_block_data[mb_index].run[level][i],
				6,"Escape_Code");
			m_pbitstrmOutFGS->putBits((i==n-1?1:0),1,"Escape,EOB");
			Nbits += table_one_bp[2*mod_index+1]+7;
		}
		if ( enh_block_data[mb_index].sign[os+ind]!=2 &&
			enh_block_data[mb_index].sign[os+ind]!=0 ) {
			m_pbitstrmOutFGS->putBits((enh_block_data[mb_index].sign[os+ind]==-1?1:0),1,"Escape-Sign");
			Nbits += 1;
			enh_block_data[mb_index].sign[os+ind] = 2;
		}
		os += enh_block_data[mb_index].run[level][i]+1;
	}
	return(Nbits);
}


Void CVideoObjectEncoder::FGSquantizeDCTcoef(Int* m_rgiFGScoef,Int mbnum)
{
	Int diff[64], sign[64];
//getting wrong values in arrays, must init correctly (copy errors?)	
	for (Int i = 0; i < BLOCK_SQUARE_SIZE; i++) {
		diff[i]=m_rgiFGScoef[i] - m_rgiDCTcoef[i]; //intel, 10/27/99
		sign[i]=SIGN(diff[i]);
		diff[i]=sign[i]*diff[i];
	}
	//added for FGS SE by FT CNET & Philips Research US
	if (m_fgs_selective_enhancement_enable)
		shift_bp_up(diff,pc_shift_factor[mbnum]);
	//end SE 
		
	// Residue Weighting, Intel, 10/25/99
	if (m_volmd.bFGSFreqWeightingEnable) 
	{
		for (Int i = 0; i < BLOCK_SQUARE_SIZE; i++)
			diff[i] = diff[i] << (m_volmd.rgiWeightMatrix[i]);
	}
	else if (m_volmd.bFGSTFreqWeightingEnable) 
	{
		for (Int i = 0; i < BLOCK_SQUARE_SIZE; i++)
			diff[i] = diff[i] << (m_volmd.rgiFGSTWeightMatrix[i]);
	}

	block2bp(diff,sign);
}


Void CVideoObjectEncoder::block2bp(Int *diff, Int *sign)
{
	Int i, j, level, x[64];
	
	level = find_level(diff);
	enh_block_data[Nblock].data_level = level;
	for ( j=0; j<level; j++ ) {
		for ( i=0; i<64; i++ ){
			x[zigzag[i]] = 0x1 & (diff[i]>>(level-1-j));
		}
		one_bp(x,j);
	}
	for ( i=0; i<64; i++ ) {
		enh_block_data[Nblock].sign[zigzag[i]] = sign[i];
	}
	Nblock += 1;
}

Void CVideoObjectEncoder::one_bp(Int *in, Int level)
{
	Int i;
	Int n;
	Int curr_run;
	Int runs[64];
	
	curr_run = 0;
	n = 0;
	for ( i=0; i<64; i++ ) {
		if ( in[i]==0 ) {
			curr_run += 1;
		} else {
			runs[n] = curr_run;
			curr_run = 0;
			n++;
		}
	}
	
	if(n==0) { /* all zero case */
		enh_block_data[Nblock].run[level][0] = 63;
		enh_block_data[Nblock].end[level][0] = 0;
	}
	for(i=0;i<n;i++){
		enh_block_data[Nblock].run[level][i] = runs[i];
		if ( i!=n-1 ) {
			enh_block_data[Nblock].end[level][i] = 0;
		} else {
			enh_block_data[Nblock].end[level][i] = 1;
		}
	}
}

Int CVideoObjectEncoder::find_level(Int *in)
{
	Int i, mx, bp_level;
	Int Lmax=12;
	//added for FGS SE by FT CNET
	if(m_fgs_selective_enhancement_enable)
		Lmax+=MAX_FGS_SE_SHIFT_FACTOR;
	//end SE

	// Residue Weighting, Intel, 12/28/99
	if (m_volmd.bFGSFreqWeightingEnable || m_volmd.bFGSTFreqWeightingEnable)
		Lmax+=MAX_FGS_FW_SHIFT_FACTOR;
	
	mx = 0;
	for ( i=0; i<64; i++ ) {
		if ( in[i]>mx ) {
			mx = in[i];
		}
	}
	i = Lmax-1;
	while ( (mx & (1<<i))==0 && i>=0 ) {
		i -= 1;
	}
	bp_level = i+1;
	return(bp_level);
}

Int CVideoObjectEncoder::get_comb_index(Int max_level_y,Int max_level_u,Int max_level_v,Int remain_level)
{
	Int comb_index;
	
	if ( max_level_y>=remain_level &&
		max_level_u>=remain_level &&
		max_level_v>=remain_level ) {
		comb_index = 0;
	} else if ( max_level_y>=remain_level &&
		max_level_u>=remain_level &&
		max_level_v<remain_level ) {
		comb_index = 1;
	} else if ( max_level_y>=remain_level &&
		max_level_u<remain_level &&
		max_level_v>=remain_level ) {
		comb_index = 2;
	} else if ( max_level_y>=remain_level &&
		max_level_u<remain_level &&
		max_level_v<remain_level ) {
		comb_index = 3;
	} else if ( max_level_y<remain_level &&
		max_level_u>=remain_level &&
		max_level_v>=remain_level ) {
		comb_index = 4;
	} else if ( max_level_y<remain_level &&
		max_level_u>=remain_level &&
		max_level_v<remain_level ) {
		comb_index = 5;
	} else if ( max_level_y<remain_level &&
		max_level_u<remain_level &&
		max_level_v>=remain_level ) {
		comb_index = 6;
	} else {
		comb_index = 7;
		printf("This cannot happen. Something is wrong!!!\n");
		exit(1);
	}
	return(comb_index);
}

Int CVideoObjectEncoder::get_cbp_index(Int comb_index,Int block_index,Int remain_level)
{
	Int cbp, i, t[6], b[6];
	
	for ( i=0; i<6; i++ ) {
		t[i] = enh_block_data[block_index+i].data_level - remain_level;
		b[i] = (t[i]>=0) ? 0 : 1;
	}
	switch(comb_index) {
    case 0: cbp = (b[0] << 5) + (b[1] << 4) + (b[2] << 3) + (b[3] << 2)
                + (b[4] << 1) + b[5];
		break;
    case 1: cbp = (b[0] << 4) + (b[1] << 3) + (b[2] << 2) + (b[3] << 1) + b[4];
		break;
    case 2: cbp = (b[0] << 4) + (b[1] << 3) + (b[2] << 2) + (b[3] << 1) + b[5];
		break;
    case 3: cbp = (b[0] << 3) + (b[1] << 2) + (b[2] << 1) + b[3];
		break;
    case 4: cbp = (b[4] << 1) + b[5];
		break;
    case 5: cbp = b[4];
		break;
    case 6: cbp = b[5];
		break;
    default: printf("Invalid combination!\n");
		exit(1);
	}
	return(cbp);
}

Int CVideoObjectEncoder::write_cbp_index(Int comb_index,Int level_index,Int block_index,Int cbp_index, Int remain_level)
{
	Int i, Nbits, b;
	Int cbp_table_3[4] = {7, 11, 13, 14};
	Int cbp_table_2[6] = {3, 5, 6, 9, 10, 12};
	Int cbp_table_1[4] = {8, 4, 2, 1};
	
	Nbits = 0;
	if ( level_index==0 ) {
		switch(comb_index) {
		case 0: if ( cbp_index==63 ) {
			use_code[block_index/6] = 1;
			m_pbitstrmOutFGS->putBits(1,1);
			Nbits += 1;
				} else {
					use_code[block_index/6] = 0;
					m_pbitstrmOutFGS->putBits(0,1,"cbp");
					Nbits += 1;
					if ( (cbp_index & 0x3)==3 ) {
						m_pbitstrmOutFGS->putBits(1,1,"cbp");
						m_pbitstrmOutFGS->putBits(cbp_index>>2,4,"cbp");
						Nbits += 5;
					} else {
						m_pbitstrmOutFGS->putBits(0,1,"cbp");
						m_pbitstrmOutFGS->putBits(cbp_index>>2,4,"cbp");
						m_pbitstrmOutFGS->putBits((cbp_index & 0x3)+1,2,"cbp");
						Nbits += 7;
					}
				}
				enh_block_data[block_index].cbp = 0x1 & (cbp_index >> 5);
				enh_block_data[block_index+1].cbp = 0x1 & (cbp_index >> 4);
				enh_block_data[block_index+2].cbp = 0x1 & (cbp_index >> 3);
				enh_block_data[block_index+3].cbp = 0x1 & (cbp_index >> 2);
				enh_block_data[block_index+4].cbp = 0x1 & (cbp_index >> 1);
				enh_block_data[block_index+5].cbp = 0x1 & cbp_index;
				break;
		case 1: if ( cbp_index==31 ) {
			use_code[block_index/6] = 1;
			m_pbitstrmOutFGS->putBits(1,1,"cbp");
			Nbits += 1;
				} else {
					use_code[block_index/6] = 0;
					m_pbitstrmOutFGS->putBits(0,1,"cbp");
					m_pbitstrmOutFGS->putBits(cbp_index+1,5,"cbp");
					Nbits += 6;
				}
				enh_block_data[block_index].cbp = 0x1 & (cbp_index >> 4);
				enh_block_data[block_index+1].cbp = 0x1 & (cbp_index >> 3);
				enh_block_data[block_index+2].cbp = 0x1 & (cbp_index >> 2);
				enh_block_data[block_index+3].cbp = 0x1 & (cbp_index >> 1);
				enh_block_data[block_index+4].cbp = 0x1 & cbp_index;
				break;
		case 2: if ( cbp_index==31 ) {
			use_code[block_index/6] = 1;
			m_pbitstrmOutFGS->putBits(1,1,"cbp");
			Nbits += 1;
				} else {
					use_code[block_index/6] = 0;
					m_pbitstrmOutFGS->putBits(0,1,"cbp");
					m_pbitstrmOutFGS->putBits(cbp_index+1,5,"cbp");
					Nbits += 6;
				}
				enh_block_data[block_index].cbp = 0x1 & (cbp_index >> 4);
				enh_block_data[block_index+1].cbp = 0x1 & (cbp_index >> 3);
				enh_block_data[block_index+2].cbp = 0x1 & (cbp_index >> 2);
				enh_block_data[block_index+3].cbp = 0x1 & (cbp_index >> 1);
				enh_block_data[block_index+5].cbp = 0x1 & cbp_index;
				break;
		case 3: if ( cbp_index==15 ) {
			use_code[block_index/6] = 1;
			m_pbitstrmOutFGS->putBits(1,1,"cbp");
			Nbits += 1;
				} else {
					use_code[block_index/6] = 0;
					m_pbitstrmOutFGS->putBits(0,1,"cbp");
					Nbits += 1;
					if ( ones_of(cbp_index)==3 ) {
						m_pbitstrmOutFGS->putBits(1,1,"cbp");
						Nbits += 1;
						for ( i=0; i<4; i++ ) {
							if ( cbp_index==cbp_table_3[i] ) {
								m_pbitstrmOutFGS->putBits(i,2,"cbp");
								Nbits += 2;
							}
						}
					} else {
						m_pbitstrmOutFGS->putBits(0,1,"cbp");
						Nbits += 1;
						if ( ones_of(cbp_index)==2 ) {
							m_pbitstrmOutFGS->putBits(1,1,"cbp");
							Nbits += 1;
							for ( i=0; i<6; i++ ) {
								if ( cbp_index==cbp_table_2[i] ) {
									m_pbitstrmOutFGS->putBits(i,3,"cbp");
									Nbits += 3;
								}
							}
						} else {
							m_pbitstrmOutFGS->putBits(0,1,"cbp");
							Nbits += 1;
							if ( ones_of(cbp_index)==1 ) {
								m_pbitstrmOutFGS->putBits(1,1,"cbp");
								Nbits += 1;
								for ( i=0; i<4; i++ ) {
									if ( cbp_index==cbp_table_1[i] ) {
										m_pbitstrmOutFGS->putBits(i,2,"cbp");
										Nbits += 2;
									}
								}
							} else {
								m_pbitstrmOutFGS->putBits(0,1,"cbp");
								m_pbitstrmOutFGS->putBits(1,1,"cbp");
								Nbits += 2;
							}
						}
					}
				}
				enh_block_data[block_index].cbp = 0x1 & (cbp_index >> 3);
				enh_block_data[block_index+1].cbp = 0x1 & (cbp_index >> 2);
				enh_block_data[block_index+2].cbp = 0x1 & (cbp_index >> 1);
				enh_block_data[block_index+3].cbp = 0x1 & cbp_index;
				break;
		case 4: m_pbitstrmOutFGS->putBits(cbp_index,2,"cbp");
			Nbits += 2;
			enh_block_data[block_index+4].cbp = 0x1 & (cbp_index >> 1);
			enh_block_data[block_index+5].cbp = 0x1 & cbp_index;
			if ( cbp_index==3 ) {
                use_code[block_index/6] = 1;
			} else {
                use_code[block_index/6] = 0;
			}
			break;
		case 5: m_pbitstrmOutFGS->putBits(cbp_index,1,"cbp");
			Nbits += 1;
			enh_block_data[block_index+4].cbp = cbp_index;
			use_code[block_index/6] = cbp_index;
			break;
		case 6: m_pbitstrmOutFGS->putBits(cbp_index,1,"cbp");
			Nbits += 1;
			enh_block_data[block_index+5].cbp = cbp_index;
			use_code[block_index/6] = cbp_index;
			break;
		default: printf("Invalid combination!\n");
			exit(1);
    }
  } else if ( use_code[block_index/6]==1 ) {
	  switch(comb_index) {
      case 0: if ( cbp_index==63 ) {
		  m_pbitstrmOutFGS->putBits(3,3,"cbp");
		  Nbits += 3;
              } else if ( cbp_index==3 ) {
				  m_pbitstrmOutFGS->putBits(2,3,"cbp");
				  Nbits += 3;
              } else if ( (cbp_index & 0x3)==3 ) {
				  m_pbitstrmOutFGS->putBits(1,1,"cbp");
				  m_pbitstrmOutFGS->putBits(cbp_index>>2,4,"cbp");
				  Nbits += 5;
              } else if ( cbp_index==0 ) {
				  m_pbitstrmOutFGS->putBits(16,5,"cbp");
				  Nbits += 5;
              } else if ( cbp_index==1 ) {
				  m_pbitstrmOutFGS->putBits(4,5,"cbp");
				  Nbits += 5;
              } else if ( cbp_index==2 ) {
				  m_pbitstrmOutFGS->putBits(5,5,"cbp");
				  Nbits += 5;
              } else if ( cbp_index==60 ) {
				  m_pbitstrmOutFGS->putBits(31,5,"cbp");
				  Nbits += 5;
              } else if ( cbp_index==61 ) {
				  m_pbitstrmOutFGS->putBits(6,5,"cbp");
				  Nbits += 5;
              } else if ( cbp_index==62 ) {
				  m_pbitstrmOutFGS->putBits(7,5,"cbp");
				  Nbits += 5;
              } else {
				  m_pbitstrmOutFGS->putBits(0,3,"cbp");
				  m_pbitstrmOutFGS->putBits(cbp_index,6,"cbp");
				  Nbits += 9;
              }
              enh_block_data[block_index].cbp = 0x1 & (cbp_index >> 5);
              enh_block_data[block_index+1].cbp = 0x1 & (cbp_index >> 4);
              enh_block_data[block_index+2].cbp = 0x1 & (cbp_index >> 3);
              enh_block_data[block_index+3].cbp = 0x1 & (cbp_index >> 2);
              enh_block_data[block_index+4].cbp = 0x1 & (cbp_index >> 1);
              enh_block_data[block_index+5].cbp = 0x1 & cbp_index;
              break;
      case 1: m_pbitstrmOutFGS->putBits(cbp_index,5,"cbp");
		  Nbits += 5;
		  enh_block_data[block_index].cbp = 0x1 & (cbp_index >> 4);
		  enh_block_data[block_index+1].cbp = 0x1 & (cbp_index >> 3);
		  enh_block_data[block_index+2].cbp = 0x1 & (cbp_index >> 2);
		  enh_block_data[block_index+3].cbp = 0x1 & (cbp_index >> 1);
		  enh_block_data[block_index+4].cbp = 0x1 & cbp_index;
		  break;
      case 2: m_pbitstrmOutFGS->putBits(cbp_index,5,"cbp");
		  Nbits += 5;
		  enh_block_data[block_index].cbp = 0x1 & (cbp_index >> 4);
		  enh_block_data[block_index+1].cbp = 0x1 & (cbp_index >> 3);
		  enh_block_data[block_index+2].cbp = 0x1 & (cbp_index >> 2);
		  enh_block_data[block_index+3].cbp = 0x1 & (cbp_index >> 1);
		  enh_block_data[block_index+5].cbp = 0x1 & cbp_index;
		  break;
      case 3: m_pbitstrmOutFGS->putBits(cbp_index,4,"cbp");
		  Nbits += 4;
		  enh_block_data[block_index].cbp = 0x1 & (cbp_index >> 3);
		  enh_block_data[block_index+1].cbp = 0x1 & (cbp_index >> 2);
		  enh_block_data[block_index+2].cbp = 0x1 & (cbp_index >> 1);
		  enh_block_data[block_index+3].cbp = 0x1 & cbp_index;
		  break;
      case 4: m_pbitstrmOutFGS->putBits(cbp_index,2,"cbp");
		  Nbits += 2;
		  enh_block_data[block_index+4].cbp = 0x1 & (cbp_index >> 1);
		  enh_block_data[block_index+5].cbp = 0x1 & cbp_index;
		  break;
      case 5: m_pbitstrmOutFGS->putBits(cbp_index,1,"cbp");
		  Nbits += 1;
		  enh_block_data[block_index+4].cbp = cbp_index;
		  break;
      case 6: m_pbitstrmOutFGS->putBits(cbp_index,1,"cbp");
		  Nbits += 1;
		  enh_block_data[block_index+5].cbp = cbp_index;
		  break;
      default: printf("Invalid combination!\n");
		  exit(1);
	  }
  } else {
	  switch(comb_index) {
      case 0: if ( enh_block_data[block_index].cbp==1 ) {
		  b = 0x1 & (cbp_index >> 5);
		  m_pbitstrmOutFGS->putBits(b,1,"cbp");
		  Nbits += 1;
		  enh_block_data[block_index].cbp = b;
              }
		  if ( enh_block_data[block_index+1].cbp==1 ) {
			  b = 0x1 & (cbp_index >> 4);
			  m_pbitstrmOutFGS->putBits(b,1,"cbp");
			  Nbits += 1;
			  enh_block_data[block_index+1].cbp = b;
		  }
		  if ( enh_block_data[block_index+2].cbp==1 ) {
			  b = 0x1 & (cbp_index >> 3);
			  m_pbitstrmOutFGS->putBits(b,1,"cbp");
			  Nbits += 1;
			  enh_block_data[block_index+2].cbp = b;
		  }
		  if ( enh_block_data[block_index+3].cbp==1 ) {
			  b = 0x1 & (cbp_index >> 2);
			  m_pbitstrmOutFGS->putBits(b,1,"cbp");
			  Nbits += 1;
			  enh_block_data[block_index+3].cbp = b;
		  }
		  if ( enh_block_data[block_index+4].cbp==1 ) {
			  b = 0x1 & (cbp_index >> 1);
			  m_pbitstrmOutFGS->putBits(b,1,"cbp");
			  Nbits += 1;
			  enh_block_data[block_index+4].cbp = b;
		  }
		  if ( enh_block_data[block_index+5].cbp==1 ) {
			  b = 0x1 & cbp_index;
			  m_pbitstrmOutFGS->putBits(b,1,"cbp");
			  Nbits += 1;
			  enh_block_data[block_index+5].cbp = b;
		  }
		  break;
      case 1: if ( enh_block_data[block_index].cbp==1 ) {
		  b = 0x1 & (cbp_index >> 4);
		  m_pbitstrmOutFGS->putBits(b,1,"cbp");
		  Nbits += 1;
		  enh_block_data[block_index].cbp = b;
              }
		  if ( enh_block_data[block_index+1].cbp==1 ) {
			  b = 0x1 & (cbp_index >> 3);
			  m_pbitstrmOutFGS->putBits(b,1,"cbp");
			  Nbits += 1;
			  enh_block_data[block_index+1].cbp = b;
		  }
		  if ( enh_block_data[block_index+2].cbp==1 ) {
			  b = 0x1 & (cbp_index >> 2);
			  m_pbitstrmOutFGS->putBits(b,1,"cbp");
			  Nbits += 1;
			  enh_block_data[block_index+2].cbp = b;
		  }
		  if ( enh_block_data[block_index+3].cbp==1 ) {
			  b = 0x1 & (cbp_index >> 1);
			  m_pbitstrmOutFGS->putBits(b,1,"cbp");
			  Nbits += 1;
			  enh_block_data[block_index+3].cbp = b;
		  }
		  if ( enh_block_data[block_index+4].cbp==1 ) {
			  b = 0x1 & cbp_index;
			  m_pbitstrmOutFGS->putBits(b,1,"cbp");
			  Nbits += 1;
			  enh_block_data[block_index+4].cbp = b;
		  }
		  break;
      case 2: if ( enh_block_data[block_index].cbp==1 ) {
		  b = 0x1 & (cbp_index >> 4);
		  m_pbitstrmOutFGS->putBits(b,1,"cbp");
		  Nbits += 1;
		  enh_block_data[block_index].cbp = b;
              }
		  if ( enh_block_data[block_index+1].cbp==1 ) {
			  b = 0x1 & (cbp_index >> 3);
			  m_pbitstrmOutFGS->putBits(b,1,"cbp");
			  Nbits += 1;
			  enh_block_data[block_index+1].cbp = b;
		  }
		  if ( enh_block_data[block_index+2].cbp==1 ) {
			  b = 0x1 & (cbp_index >> 2);
			  m_pbitstrmOutFGS->putBits(b,1,"cbp");
			  Nbits += 1;
			  enh_block_data[block_index+2].cbp = b;
		  }
		  if ( enh_block_data[block_index+3].cbp==1 ) {
			  b = 0x1 & (cbp_index >> 1);
			  m_pbitstrmOutFGS->putBits(b,1,"cbp");
			  Nbits += 1;
			  enh_block_data[block_index+3].cbp = b;
		  }
		  if ( enh_block_data[block_index+5].cbp==1 ) {
			  b = 0x1 & cbp_index;
			  m_pbitstrmOutFGS->putBits(b,1,"cbp");
			  Nbits += 1;
			  enh_block_data[block_index+5].cbp = b;
		  }
		  break;
      case 3: if ( enh_block_data[block_index].cbp==1 ) {
		  b = 0x1 & (cbp_index >> 3);
		  m_pbitstrmOutFGS->putBits(b,1,"cbp");
		  Nbits += 1;
		  enh_block_data[block_index].cbp = b;
              }
		  if ( enh_block_data[block_index+1].cbp==1 ) {
			  b = 0x1 & (cbp_index >> 2);
			  m_pbitstrmOutFGS->putBits(b,1,"cbp");
			  Nbits += 1;
			  enh_block_data[block_index+1].cbp = b;
		  }
		  if ( enh_block_data[block_index+2].cbp==1 ) {
			  b = 0x1 & (cbp_index >> 1);
			  m_pbitstrmOutFGS->putBits(b,1,"cbp");
			  Nbits += 1;
			  enh_block_data[block_index+2].cbp = b;
		  }
		  if ( enh_block_data[block_index+3].cbp==1 ) {
			  b = 0x1 & cbp_index;
			  m_pbitstrmOutFGS->putBits(b,1,"cbp");
			  Nbits += 1;
			  enh_block_data[block_index+3].cbp = b;
		  }
		  break;
      case 4: if ( enh_block_data[block_index+4].cbp==1 ) {
		  b = 0x1 & (cbp_index >> 1);
		  m_pbitstrmOutFGS->putBits(b,1,"cbp");
		  Nbits += 1;
		  enh_block_data[block_index+4].cbp = b;
              }
		  if ( enh_block_data[block_index+5].cbp==1 ) {
			  b = 0x1 & cbp_index;
			  m_pbitstrmOutFGS->putBits(b,1,"cbp");
			  Nbits += 1;
			  enh_block_data[block_index+5].cbp = b;
		  }
		  break;
      case 5: if ( enh_block_data[block_index+4].cbp==1 ) {
		  b = 0x1 & cbp_index;
		  m_pbitstrmOutFGS->putBits(b,1,"cbp");
		  Nbits += 1;
		  enh_block_data[block_index+4].cbp = b;
              }
		  break;
      case 6: if ( enh_block_data[block_index+5].cbp==1 ) {
		  b = 0x1 & cbp_index;
		  m_pbitstrmOutFGS->putBits(b,1,"cbp");
		  Nbits += 1;
		  enh_block_data[block_index+5].cbp = b;
              }
		  break;
      default: printf("Invalid combination!\n");
		  exit(1);
    }
  }
  return(Nbits);
}

Int CVideoObjectEncoder::ones_of(Int number)
{
	Int i, ones;
	
	ones = 0;
	for ( i=0; i<31; i++ ) {
		ones += 0x1 & (number>>i);
	}
	return(ones);
}

/* end of FGS (Optivision) */




void CVideoObjectEncoder::process_vop_shift_factors()
{
	int  iNumMBX;
	int  iNumMBY;
	int i,j;
	int MbNum;
	int MbInc;

	if (!m_fgs_selective_enhancement_enable)
		return;

	if(m_fgsSelEnhMode->initNotDone)
	{
		switch(m_fgsSelEnhMode->fgsSelEnhAlgo)
		{
		case e_constantRoi:
			iNumMBX= (m_ivolWidth+15)/16;
			iNumMBY= (m_ivolHeight+15)/16;
			MbNum=m_fgsSelEnhMode->fgsSelEnhMbNumStartX+(iNumMBX*m_fgsSelEnhMode->fgsSelEnhMbNumStartY);
			assert(m_fgsSelEnhMode->fgsSelEnhMbNumEndX>=m_fgsSelEnhMode->fgsSelEnhMbNumStartX);
			assert(m_fgsSelEnhMode->fgsSelEnhMbNumEndX<=iNumMBX);
			assert(m_fgsSelEnhMode->fgsSelEnhMbNumEndY>=m_fgsSelEnhMode->fgsSelEnhMbNumStartY);
			assert(m_fgsSelEnhMode->fgsSelEnhMbNumEndY<=iNumMBY);
			MbInc=iNumMBX-m_fgsSelEnhMode->fgsSelEnhMbNumEndX+m_fgsSelEnhMode->fgsSelEnhMbNumStartX;
			memset(pc_shift_factor,0,iNumMBX*iNumMBY);
			for(i=m_fgsSelEnhMode->fgsSelEnhMbNumStartY; i<m_fgsSelEnhMode->fgsSelEnhMbNumEndY; i++)
			{
				for (j=m_fgsSelEnhMode->fgsSelEnhMbNumStartX; j<m_fgsSelEnhMode->fgsSelEnhMbNumEndX; j++)
				{
					pc_shift_factor[MbNum++]=m_fgsSelEnhMode->fgsSelEnhShiftValue;
				}
				MbNum+=MbInc;
			}
			break;
		case e_fileGivenRoi:
			iNumMBX= (m_ivolWidth+15)/16;
			iNumMBY= (m_ivolHeight+15)/16;
			int itemp;
			*m_fgsSelEnhMode->fgsSelEnhInStream>>itemp;
			if(itemp!=iNumMBX)
			{
				cerr<<"Bad Mb num width for file given ROI"<<endl;
				abort();
			}
			*m_fgsSelEnhMode->fgsSelEnhInStream>>itemp;
			if(itemp!=iNumMBY)
			{
				cerr<<"Bad Mb num height for file given ROI"<<endl;
				abort();
			}
			for(i=0; i<iNumMBY*iNumMBX; i++)
			{
				*m_fgsSelEnhMode->fgsSelEnhInStream>>itemp;
				pc_shift_factor[i]=itemp;
			}
			break;
		case e_selProcessed:
			cerr<<"Self processing selective enhancement not implemented yet"<<endl;
			abort();
			break;
		}
		m_fgsSelEnhMode->initNotDone=0;
	}
}


//added for FGS SE by FT CNET & Philips US Research

void CVideoObjectEncoder::shift_bp_up(Int diff[],Int sf)
{
  Int k;

  if (sf <=0 ) return;
  for (k=0;k<64;k++)
    diff[k] <<= sf;
}

void CVideoObjectEncoder::get_shift_factor_vlc(Int sf, Int &vlc_code, Int &vlc_length)
{
  /* 0: 1
     1: 01
     2: 001
     3: 0001
     4: 00001
     5: 000001
     6: 0000001
  */

  vlc_code=1;
  vlc_length=sf+1;
}
/* end SE */
