/******************************************************************************
 *                                                                          
 * This software module was originally developed by 
 *
 *   Fujitsu Laboratories Ltd. (contact: Eishi Morimatsu)
 *
 * in the course of development of the MPEG-4 Video (ISO/IEC 14496-2) standard.
 * This software module is an implementation of a part of one or more MPEG-4 
 * Video (ISO/IEC 14496-2) tools as specified by the MPEG-4 Video (ISO/IEC 
 * 14496-2) standard. 
 *
 * ISO/IEC gives users of the MPEG-4 Video (ISO/IEC 14496-2) standard free 
 * license to this software module or modifications thereof for use in hardware
 * or software products claiming conformance to the MPEG-4 Video (ISO/IEC 
 * 14496-2) standard. 
 *
 * Those intending to use this software module in hardware or software products
 * are advised that its use may infringe existing patents. THE ORIGINAL 
 * DEVELOPER OF THIS SOFTWARE MODULE AND HIS/HER COMPANY, THE SUBSEQUENT 
 * EDITORS AND THEIR COMPANIES, AND ISO/IEC HAVE NO LIABILITY FOR USE OF THIS 
 * SOFTWARE MODULE OR MODIFICATIONS THEREOF. No license to this software
 * module is granted for non MPEG-4 Video (ISO/IEC 14496-2) standard 
 * conforming products.
 *
 * Fujitsu Laboratories Ltd. retains full right to use the software module for 
 * his/her own purpose, assign or donate the software module to a third party 
 * and to inhibit third parties from using the code for non MPEG-4 Video 
 * (ISO/IEC 14496-2) standard conforming products. This copyright notice must 
 * be included in all copies or derivative works of the software module. 
 *
 * Copyright (c) 1999.  
 *

Module Name:

	rrv.cpp

Abstract:

	Utility functions for Reduced Resolution Vop (RRV) mode

Revision History:

 *****************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <fstream.h>
#include <iostream.h>
#include "typeapi.h"
#include "codehead.h"
#include "global.hpp"
#include "mode.hpp"
#include "sesenc.hpp"
#include "vopses.hpp"
#include "vopseenc.hpp"


#define RRV_MAX(x,y) (((x)>(y))?(x):(y))
#define RRV_MIN(x,y) (((x)<(y))?(x):(y))
#define RRV_BLOCK_SIZE 16

#define	RRV_C1	2.5
#define	RRV_C2	2.5
#define	RRV_FR1	6.0
#define	RRV_FR2	8.0
#define	RRV_QP1	14
#define	RRV_QP2	6

Void	CVideoObjectEncoder::cutoffDCTcoef()
{
	if((m_vopmd.RRVmode.iCutoffThr == 8)||(m_vopmd.RRVmode.iOnOff != 1))
	{
		return;
	}

	assert((m_vopmd.RRVmode.iCutoffThr >= 4)&&(m_vopmd.RRVmode.iCutoffThr <= 7));
	Int		x, y, pos;

	for(y = 0 ; y < BLOCK_SIZE; y ++)
	{
		for(x = 0 ; x < BLOCK_SIZE; x ++)
		{
			if((x >= m_vopmd.RRVmode.iCutoffThr)||(y >= m_vopmd.RRVmode.iCutoffThr))
			{
				pos	= x +y *BLOCK_SIZE;
				m_rgiDCTcoef[pos]	= 0;
			}
		}
	}
}

Void	CVideoObjectEncoder::redefineVOLMembersRRV()
{
	m_iNumMBX	= m_iVOPWidthY / MB_SIZE;
	m_iNumMBY	= m_rctCurrVOPY.height () / MB_SIZE;
	m_iNumMB	= m_iNumMBX * m_iNumMBY;
	m_iNumOfTotalMVPerRow	= PVOP_MV_PER_REF_PER_MB * m_iNumMBX;
	if(m_vopmd.RRVmode.iRRVOnOff == 1)
	{
		m_iRRVScale	= 2;
		m_iNumMBX	= m_iNumMBX /2;
		m_iNumMBY	= m_iNumMBY /2;
		m_iNumMB	= m_iNumMB /4;
		m_iNumOfTotalMVPerRow	= m_iNumOfTotalMVPerRow /2;
	}
	else
	{
		m_iRRVScale	= 1;
	}
	m_iFrameWidthYxMBSize = (MB_SIZE *m_iRRVScale) 
		*m_pvopcRefQ0->whereY ().width; 
	m_iFrameWidthYxBlkSize = (BLOCK_SIZE *m_iRRVScale) 
		*m_pvopcRefQ0->whereY ().width; 
	m_iFrameWidthUVxBlkSize = (BLOCK_SIZE *m_iRRVScale) *
		m_pvopcRefQ0->whereUV ().width;
}


Void	CVideoObjectEncoder::resetAndCalcRRV()
{
	assert(m_vopmd.RRVmode.iOnOff == 1);
	assert(m_vopmd.RRVmode.iCycle != 0);
	
	static Int iCount	= 0;
	static Int iFirst	= 0;
	
    /* fixed interval mode */
	if(m_vopmd.RRVmode.iCycle == 0)
	{
		m_vopmd.RRVmode.iRRVOnOff	= 0;
	}
	else if(m_vopmd.RRVmode.iCycle < 0)
	{
//		m_vopmd.RRVmode.iCutoffThr	= 8;
		switch(iFirst)
		{
		  case 0:
			m_vopmd.RRVmode.iRRVOnOff	= 0;
			m_vopmd.RRVmode.iCutoffThr	= 8;
			iCount++;
			iFirst	= 1;
			break;
		  case 1:
			iCount	= iCount % m_vopmd.RRVmode.iCycle;
			if(iCount == 0)
			{
				m_vopmd.RRVmode.iRRVOnOff	^= 0x01;
				if(m_vopmd.RRVmode.iRRVOnOff == 0)
				{
				    m_vopmd.RRVmode.iCutoffThr	= 4;
				}
				else
				{
				    m_vopmd.RRVmode.iCutoffThr	= 8;
				}
			}
			else
			{
			    if(m_vopmd.RRVmode.iCutoffThr < 8)
			    {
				m_vopmd.RRVmode.iCutoffThr ++;
			    }
			}
			iCount++;	      
			break;
		  default:
			fprintf(stderr, "Error in resetAndCalcRRV\n");
			exit(1);
			break;
		}
	}
	/* adaptive mode */
	else
	{
		Float	f_alpha, f_TH1, f_TH2;

		if(m_vopmd.RRVmode.iQave == 0)
		{
			/* perhaps first I-VOP */
			m_vopmd.RRVmode.iRRVOnOff	= 0;
			m_vopmd.RRVmode.iCutoffThr	= 8;
			return;
		}
		f_alpha	= (Float)(m_vopmd.RRVmode.iQave) 
			*(Float)(m_vopmd.RRVmode.iNumBits);

		m_vopmd.RRVmode.iQP1	= RRV_QP1;
		m_vopmd.RRVmode.iQP2	= RRV_QP2;
		m_vopmd.RRVmode.fFR1	= RRV_FR1;
		m_vopmd.RRVmode.fFR2	= RRV_FR2;
		
//		printf("target bitrate %d\n", m_iBufferSize);
		
		f_TH1	= m_vopmd.RRVmode.iQP1 *(Float)(m_iBufferSize) 
			/m_vopmd.RRVmode.fFR1;
		
		f_TH2	= m_vopmd.RRVmode.iQP2 *(Float)(m_iBufferSize) 
			/m_vopmd.RRVmode.fFR2;

		if((m_vopmd.RRVmode.iRRVOnOff == 0)&&(f_alpha > f_TH1))
		{
			m_vopmd.RRVmode.iRRVOnOff	= 1;
			m_vopmd.RRVmode.iCutoffThr	= 8;
		}
		else if((m_vopmd.RRVmode.iRRVOnOff == 1)&&(f_alpha < f_TH2))
		{
			m_vopmd.RRVmode.iRRVOnOff	= 0;
			m_vopmd.RRVmode.iCutoffThr	= 4;
		}
		else if((m_vopmd.RRVmode.iRRVOnOff == 0)&&(m_vopmd.RRVmode.iCutoffThr < 8))
		{
			m_vopmd.RRVmode.iCutoffThr ++;
		}

//		printf("RRV_debug: QxB = %f, TH1 = %f, TH2 = %f, Mode = %d, Thr = %d\n",
//			   f_alpha, f_TH1, f_TH2, m_vopmd.RRVmode.iRRVOnOff, m_vopmd.RRVmode.iCutoffThr);
	}
}

