/*
 * FGST_server.c
 *
 * M. van der Schaar, Philips Research USA
 *
 * Truncate a bitstream of FGS+FGST for the case P-frames are employed in the temporal FGST-layer
 * This truncation program is also reordering the frames, since at encoding time the frames are
 * outputed correctly (i.e. each BL -frame is followed by an FGST-P-frame), but the current decoder 
 * is processing them in a different order (i.e the next BL_frame is processed prior to the
 * FGST-P-frame. This reordering is now performed by this reordering program, but idealy the decoder 
 * needs to be modified. Moreover, this reordering is ONLY necessary if the FGST-frames are P-frames (not B)
 * and an FGS+FGST-stream is streamed (i.e. the reordering is not necessary if FGS and FGST are in independent
 * streams. 
 *
 * This program is based on the FGS_server program provided by Optivision.
 */

#include <stdio.h>
#include <math.h>

#define FGS_VOP_START_CODE 0x1B9
#define VISUAL_OBJECT_SEQUENCE_END_CODE 0x1B1

main(ac,av)
int ac;
char *av[];
{
    FILE *infp, *outfp, *fopen(), 
         *intermfp /* File used for the intermediate storage of the FGST-P frames */;
    int i,start_code, N_vop, bytes_per_vop, N_bytes, byte, byte1, byte2, byte3, byte4, diff_bytes;
    int total_in_bytes, total_out_bytes;
    double br, fr;

    if ( ac != 5 ) {
        printf("Usage: %s finame foname bit-rate frame-rate\n",av[0]);
        exit(1);
    }
    infp = fopen(av[1],"r");
    outfp = fopen(av[2],"w");
    sscanf(av[3],"%lf",&br);
    sscanf(av[4],"%lf",&fr);
    printf("Input bitstream file: %s\n",av[1]);
    printf("Output bitstream file: %s\n",av[2]);
    bytes_per_vop = ceil((br/fr)/8);

    N_vop = 0;
    N_bytes = 0;
    total_in_bytes = 0;
    total_out_bytes = 0;
    start_code = 0;
    do {
        byte = getc(infp);
        putc(byte,outfp);
        N_bytes += 1;
        start_code = (start_code<<8) | byte;
    } while ( byte!=EOF && start_code!=FGS_VOP_START_CODE);
    
    total_in_bytes += N_bytes - (byte==EOF);
    total_out_bytes += N_bytes - (byte==EOF);
    printf("Number of VOL overhead bytes = %d %d\n",N_bytes-4,total_out_bytes);
    
    
    while ( byte != EOF ) {
        N_bytes = 0;
        start_code = 0;
	    
        if ( N_vop % 2 == 1 )                    
	    intermfp = fopen("interm","w");
        do {
                byte = getc(infp);
		if ( byte!=EOF && N_bytes<bytes_per_vop ) {
		  if ( N_vop % 2 == 1 )    	
                  putc(byte,intermfp);
		  else
		  {
		      putc(byte,outfp);
		      total_out_bytes += 1;
		  }
		  
		}
                N_bytes += 1;
                start_code = (start_code<<8) | byte;
        } while ( byte!=EOF && start_code!=FGS_VOP_START_CODE);
            
        total_in_bytes += N_bytes - (byte==EOF);
        diff_bytes = N_bytes - bytes_per_vop;
		
        if ( diff_bytes>4 ) diff_bytes = 4;
        if ( diff_bytes<0 ) diff_bytes = 0;
	
        for ( i=0; i<diff_bytes; i++ ) {
               byte1 = 0xff & (FGS_VOP_START_CODE>>(8*(diff_bytes-1-i)));
	       if (N_vop % 2 == 1 ) putc(byte1,intermfp);
               else
	       {
		       putc(byte1,outfp);
                       total_out_bytes += 1; 
	       }   
        }

	if (N_vop >= 1)
	{
	       if (N_vop % 2 == 1 )   fclose(intermfp);
	       else 
	       {	    
	              intermfp = fopen("interm","r");      
	              if (byte != EOF)
	              {
	                  byte1 = getc(intermfp);
                          while ( byte1!=EOF) {
                                putc(byte1,outfp);
	                        total_out_bytes += 1;
                                N_bytes += 1;
	                        byte1 = getc(intermfp);
                          } 
	              }
	              else
	              {
				byte1 = getc(intermfp);
	    		        
				for (i=0;i<bytes_per_vop;i++)
	     			{
            			    putc(byte1,outfp);
	 		            total_out_bytes ++;
              			    N_bytes ++;
				    
	         	            byte1 = getc(intermfp);
			       }	    
                       }	      
	               fclose(intermfp);
               }
	}    
   
        N_vop += 1;
    }
    
     for ( i=0; i<4; i++ ) {
        byte = 0xff & (VISUAL_OBJECT_SEQUENCE_END_CODE>>(8*(3-i)));
        putc(byte,outfp);
        total_out_bytes += 1;
    }
    fclose(outfp);
    printf("Total input bytes = %d\n",total_in_bytes);
    printf("Total output bytes = %d\n",total_out_bytes);
    printf("Number of VOPs = %d\n",N_vop);
}
