/**********************************************************************
 * Copyright 1998, 1999 by Optivision, Inc., Palo Alto, CA.
 * Permission to use this program royalty-free for MPEG work is hereby
 * granted, provided that this copyright and permission notice appear
 * on all SOURCE copies and that the name Optivision, Inc. not be used
 * in advertising or publicity pertaining to distribution of the
 * program without specific prior written permission.  Optivision,
 * Inc. makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express
 * or implied warranty.
 **********************************************************************/

/*
 * FGS_server.c
 *
 */

#include <stdio.h>
#include <math.h>

#define VOP_START_CODE 0x1B6
#define VOP_START_CODE_LENGTH   32

#define FGS_VOP_START_CODE 0x1B9
#define FGS_VOP_START_CODE_LENGTH   32

#define VISUAL_OBJECT_SEQUENCE_END_CODE 0x1B1

main(ac,av)
int ac;
char *av[];
{
    FILE *infp, *outfp, *fopen();
    int i, start_code, N_vop, bytes_per_vop, N_bytes, byte, diff_bytes;
    int total_in_bytes, total_out_bytes;
    double br, fr;

    if ( ac != 5 ) {
        printf("Usage: %s finame foname bit-rate frame-rate\n",av[0]);
        exit(1);
    }
    infp = fopen(av[1],"r");
    outfp = fopen(av[2],"w");
    sscanf(av[3],"%lf",&br);
    sscanf(av[4],"%lf",&fr);
    printf("Input bitstream file: %s\n",av[1]);
    printf("Output bitstream file: %s\n",av[2]);
    bytes_per_vop = ceil((br/fr)/8);

    N_vop = 0;
    N_bytes = 0;
    total_in_bytes = 0;
    total_out_bytes = 0;
    start_code = 0;
    do {
        byte = getc(infp);
        putc(byte,outfp);
        N_bytes += 1;
        start_code = (start_code<<8) | byte;
    } while ( byte!=EOF && start_code!=FGS_VOP_START_CODE );
    total_in_bytes += N_bytes - (byte==EOF);
    total_out_bytes += N_bytes - (byte==EOF);
    printf("Number of overhead bytes = %d\n",N_bytes-4);
    while ( byte != EOF ) {
        N_bytes = 0;
        start_code = 0;
        do {
            byte = getc(infp);
            if ( byte!=EOF && N_bytes<bytes_per_vop ) {
                putc(byte,outfp);
                total_out_bytes += 1;
            }
            N_bytes += 1;
            start_code = (start_code<<8) | byte;
        } while ( byte!=EOF && start_code!=FGS_VOP_START_CODE );
        total_in_bytes += N_bytes - (byte==EOF);
        if ( byte!=EOF ) {
            diff_bytes = N_bytes - bytes_per_vop;
            if ( diff_bytes>4 ) diff_bytes = 4;
            if ( diff_bytes<0 ) diff_bytes = 0;
            for ( i=0; i<diff_bytes; i++ ) {
                byte = 0xff & (start_code>>(8*(diff_bytes-1-i)));
                putc(byte,outfp);
                total_out_bytes += 1;
            }
        }
        N_vop += 1;
    }
    start_code = VISUAL_OBJECT_SEQUENCE_END_CODE;
    for ( i=0; i<4; i++ ) {
        byte = 0xff & (start_code>>(8*(3-i)));
        putc(byte,outfp);
        total_out_bytes += 1;
    }
    printf("Total input bytes = %d\n",total_in_bytes);
    printf("Total output bytes = %d\n",total_out_bytes);
    printf("Number of VOPs = %d\n",N_vop);
}
