/*****************************************************************************
 *                                                                          
 * This software module was developed by 
 *
 *   Weiping Li (Optivision, Inc.) and
 *   Fan Ling (General Instrument)
 *
 * in the course of development of the MPEG-4 Video (ISO/IEC 14496-2) standard.
 * This software module is an implementation of a part of one or more MPEG-4 
 * Video (ISO/IEC 14496-2) tools as specified by the MPEG-4 Video (ISO/IEC 
 * 14496-2) standard. 
 *
 * ISO/IEC gives users of the MPEG-4 Video (ISO/IEC 14496-2) standard free 
 * license to this software module or modifications thereof for use in hardware
 * or software products claiming conformance to the MPEG-4 Video (ISO/IEC 
 * 14496-2) standard. 
 *
 * Those intending to use this software module in hardware or software products
 * are advised that its use may infringe existing patents. The original 
 * developer of this software module and his/her company, the subsequent 
 * editors and their companies, and ISO/IEC have no liability for use of this 
 * software module or modifications thereof in an implementation. Copyright is 
 * not released for non MPEG-4 Video (ISO/IEC 14496-2) Standard conforming 
 * products. 
 *
 * ACTS-MoMuSys partners retain full right to use the code for his/her own 
 * purpose, assign or donate the code to a third party and to inhibit third 
 * parties from using the code for non MPEG-4 Video (ISO/IEC 14496-2) Standard
 * conforming products. This copyright notice must be included in all copies or
 * derivative works. 
 *
 * Copyright (c) 1997
 *
 *****************************************************************************/

/*
 * bp_fgs_enh.h
 *
 */
#ifndef BP_ENH_H
#define BP_ENH_H
#include <stdio.h>
#include <stdlib.h>

#define FGS_VOP_START_CODE 0x1B9 /* use this one for now */
/* Philips Research; FGS-TS */
#define SVP_VO_TYPE 18 /* VO_type for FGS & FGST */
#define FGS_LAYER_TYPE 		1
#define FGST_LAYER_TYPE 	2
#define FGS_FGST_LAYER_TYPE 	3
/* end FGS-TS */
#define FGS_VOP_START_CODE_LENGTH 32
/* fgs_vop_bp_marker, Optivision */
#define FGS_VOP_BP_MARKER_HEADER 0xA /* first 27 bits of fgs_vop_bp_marker */
#define FGS_VOP_BP_MARKER_HEADER_LENGTH 27
#define FGS_VOP_BP_MARKER_ID_LENGTH 5
/* fgs_vop_bp_marker, Optivision */

/* FGS error resilience, MSRCN, 2000/8/20*/
#define FGS_VP_MARKER                1
#define NUMBITS_FGS_VP_MARKER		23

#define VISUAL_OBJECT_SEQUENCE_END_CODE 0x1B1

/* to cover both selective enhancement and frequency weighting */
#define MAX_BP_LEVEL 20

/*#define FGSPOSTCLIP*/ /* M. van der Schaar, Aug. 15 */
typedef struct FGS_GLOBAL
{
  Int fgs_on;
  Int write_fgs_on;/* Philips Research; FGS-TS */
  Int fgs_intra;
  Int fgs_resync_marker_disable;

  /* added by Wu Feng for FGS error resilience, MSRCN, 2000/8/20*/
  Int fgs_video_packet_length;
  Int fgs_number_bits_mb;

  Int Nblock;
  Int fgs_co_located_not_coded;
  Int fgs_vop_max_level_y;
  Int fgs_vop_max_level_u;
  Int fgs_vop_max_level_v;
  Int fgs_vop_number_of_vop_bp_coded;
  Int fgs_vop_mc_bit_plane_not_used;
  Int fgs_vop_total_bits;
  Int fgs_decode;
  Int fgs_vop_end;
  Int index;
  Int time_increment_resolution;
  Float time;

  /* Philips Research; SE */
  Int fgs_selective_enhancement_enable;
  Int fgs_mask_type;
  Int fgs_mask_start_x;
  Int fgs_mask_start_y;
  Int fgs_mask_end_x;
  Int fgs_mask_end_y;
  Int fgs_rectangular_shift_factor;
  char fgs_mask_file[100];
  /* end SE */
  /* Frequency Weighting, Optivision */
  Int fgs_frequency_weighting_enable;
  Int fgs_load_frequency_weighting_matrix;
  Int fgs_shift_matrix[64];
  /* Frequency Weighting, Optivision */
  /* Philips Research; FGS-TS */
  Int fgst_frequency_weighting_enable;
  Int fgst_load_frequency_weighting_matrix;
  Int fgst_shift_matrix[64];
  Int temporal_scalability;
  Int FGST_flag; /* indicates on a frame-basis whether the frame is FGS or FGST*/
  Int fgs_and_fgst_in_one_stream;  /* FGS & FGST in one stream or not */
  /* end  FGS-TS */ 
} FGS_GLOBAL;

typedef struct ENH_BLOCK
{
  Int data_level;
  Int decoded_level;
  Int dct_type;
  Int cbp;
  Int mode;
  Int run[MAX_BP_LEVEL][64];
  Int end[MAX_BP_LEVEL][64];
  Int sign[64];
  Int residue_sign[64];
} ENH_BLOCK;

#endif /* BP_ENH_H */
