/*****************************************************************************
 *
 * This software module was originally developed by
 *
 *   Weiping Li (Optivision, Inc.) and
 *   Fan Ling (General Instrument)
 *
 * and edited by
 *    Yingwei Chen - SE (Philips Research)
 *    Mihaela van der Schaar - FGS-TS (Philips Research, Mihaela.vanderschaar@philips.com)
 *    Feng Wu - FGS error resilience (Microsoft Research China, fengwu@microsoft.com)
 *
 * in the course of development of the MPEG-4 Video (ISO/IEC 14496-2) standard.
 * This software module is an implementation of a part of one or more MPEG-4
 * Video (ISO/IEC 14496-2) tools as specified by the MPEG-4 Video (ISO/IEC
 * 14496-2) standard.
 *
 * ISO/IEC gives users of the MPEG-4 Video (ISO/IEC 14496-2) standard free
 * license to this software module or modifications thereof for use in hardware
 * or software products claiming conformance to the MPEG-4 Video (ISO/IEC
 * 14496-2) standard.
 *
 * Those intending to use this software module in hardware or software products
 * are advised that its use may infringe existing patents. The original
 * developer of this software module and his/her company, the subsequent
 * editors and their companies, and ISO/IEC have no liability for use of this
 * software module or modifications thereof in an implementation. Copyright is
 * not released for non MPEG-4 Video (ISO/IEC 14496-2) Standard conforming
 * products.
 *
 * ACTS-MoMuSys partners retain full right to use the code for his/her own
 * purpose, assign or donate the code to a third party and to inhibit third
 * parties from using the code for non MPEG-4 Video (ISO/IEC 14496-2) Standard
 * conforming products. This copyright notice must be included in all copies or
 * derivative works.
 *
 * Copyright (c) 1997
 *
 *****************************************************************************/

/***********************************************************HeaderBegin*******
 *                                                                         
 * File:        bp_fgs_dec.c
 *
 * Author:      Weiping Li, Optivision
                Fan Ling, General Instrument
 * Created:     
 *                                                                         
 * Description:
 *
 * Notes:
 *
 * Modified:
 *     August-06-99 Yingwei Chen: Changed FGS enhancement layer residual signal
 *                                decoding and FGS enhanced video composition
 *
 ***********************************************************HeaderEnd*********/

/************************    INCLUDE FILES    ********************************/

#include <string.h>
#include "momusys.h"
#include "mom_bitstream_d.h"
#include "vm_compos.h"
#include "text_decode.h"
#include "mot_padding.h"
#include "newpred_d.h"  /** added for NEWPRED (Oki) 16-AUG-1999 **/
#include "io_generic.h"
#include "vm_dec_main.h"
#include "mom_vol.h"
#include "mom_access.h"
#include "mot_comp.h"
#include "zigzag.h"
#include "post_filter.h"
#include "mom_vo.h"
#include "combined_decode.h"
#include "vm_common_defs.h"
#include "sprite_util.h"
#include "sprite_dec_util.h"
#include "sprite_dec_piece.h"
#include "alp_dec_mom.h"
#include "alp_dec_header.h"
#include "text_quant_mat_def.h"
#include "interface.h"
#include "do_bgc.h" 
#include "scal_dec.h"
#include "startcode.h"

#ifndef SEEK_SET
#define SEEK_SET 0
#endif

#include "bp_fgs_enh.h"
#include "vlcbp.h"
#include "vlcbp4fw.h"
#include "bp_fgs_dec.h"

ENH_BLOCK *enh_block_data;
static Int use_code[1620];
FGS_GLOBAL *fgs_global;

/* Philips Research; SE */
Int shift_factor[720*576/256];
Int mb_shift_factor_received[720*576/256];
/* end SE */

/* Philips Research; FGS-TS */
Int vop_time_inc;
extern Int time_base;
/* end FGS-TS */

Void
fgs_dec_init(Vop *curr_vop)
{
  int i, j, k, nblock;

  nblock = 6*GetVopWidth(curr_vop)*GetVopHeight(curr_vop)/(16*16);
  printf("Initialize FGS data array....\n");

  if(enh_block_data==NULL)
    enh_block_data = (ENH_BLOCK *)malloc(nblock*sizeof(ENH_BLOCK));
  for ( i=0; i<nblock; i++ ) {
    enh_block_data[i].cbp = 1;
    enh_block_data[i].dct_type = 0;
    enh_block_data[i].mode = INTER; /* added by M. van der Schaar */
    
    for ( j=0; j<MAX_BP_LEVEL; j++ ) {
      for ( k=0; k<64; k++ ) {
        enh_block_data[i].run[j][k]= 63;
        enh_block_data[i].end[j][k]= 0;
      }
    }
    for ( k=0; k<64; k++ ) {
      enh_block_data[i].sign[k]= 0;
      enh_block_data[i].residue_sign[k]=0;
    }
  }
  fgs_global->fgs_decode = 0;
  fgs_global->fgs_vop_end = 0;

  /* Philips Research; SE */
  for (i=0;i<nblock/6;i++)
    mb_shift_factor_received[i]=0;
  /* end SE */
}

Void
DecodeFGSVop(Bitstream **stream, Int vo_id, Trace *trace, Vop *curr_vop,
	     Int short_video_header)
{
  Vop *fgs_vop;
  Vop *tmp_vop;

  Int num_pixels;
  Int num_lines;
  Int vol_id, nvol_id;
  Int Nbits;
  Int tmpvar;
Image *Y_tmp, *Y_curr;

  nvol_id = GetVopId( curr_vop );
  vol_id = bs_vol[vo_id][nvol_id];

/* We assume volID "vol_id+10" is not used by other vols, and
   vol_id+10<MAX_NUM_VOLS
*/
 
   /* Philips Research; FGS-TS */
   if (fgs_global->FGST_flag == 0)
   /* end FGS-TS */
   vol_id = vol_id + 10;  
   /* Philips Research; FGS-TS */
   else if (fgs_global->fgs_and_fgst_in_one_stream ==0) vol_id = 1;
   else vol_id = 10;
   /* end FGS-TS */

  /*****
   *
   *	Free the forelast VOP and copy the current (which is now the previous
   *	VOP) into the right place of the structure
   *
   *****/

  fgs_vop = GetVopFGS(curr_vop);

  tmp_vop = AllocVop(GetVopWidth(fgs_vop),GetVopHeight(fgs_vop),GetVopAuxCompCount(fgs_vop));

  CopyVopNonImageField(fgs_vop,tmp_vop);

  num_pixels = GetVopWidth(curr_vop);
  num_lines = GetVopHeight(curr_vop);
  fgs_global->Nblock = 6*num_pixels*num_lines/(16*16);
  fgs_global->fgs_vop_total_bits = 0;
  fgs_global->fgs_decode = 1;
  
  /* Philips Research FGST */
  PutVopTimeInc(vop_time_inc, fgs_vop );
  Nbits = 0;    
  /* end FGST */
  
  /* added by wu feng for FGS error resilience, MSRCN, 2000/8/20*/
  fgs_global->fgs_number_bits_mb = (int)ceil(log10((double)(fgs_global->Nblock/6))/log10(2.0));

  Nbits += input_enh_bitstream(stream[vol_id],trace);

  fgs_global->fgs_decode = 0;
  fgs_global->Nblock = 0;

  DecodeFGSVopCombinedMotionShapeTexture(vo_id, 
  				      trace, tmp_vop, short_video_header);

#ifdef FGSPOSTCLIP
  VopTextureUpdate(tmp_vop,fgs_vop,trace);
#else
  if(GetVopPredictionType(fgs_vop) == I_VOP){
    CopyImage(GetVopY(tmp_vop),GetVopY(fgs_vop));
    CopyImage(GetVopU(tmp_vop),GetVopU(fgs_vop));
    CopyImage(GetVopV(tmp_vop),GetVopV(fgs_vop));
  }else
    VopTextureUpdate(tmp_vop,fgs_vop,trace);
#endif
  
  FreeVop(tmp_vop);

/* do byte alignment,and forward to next VOP_START_CODE,
   0x0001 indicates end_of_file
 */

  BitstreamFlushBits(stream[ vol_id ], stream[vol_id]->incnt % 8);
  tmpvar = BitstreamShowBits(stream[vol_id],32);

  while(tmpvar!=FGS_VOP_START_CODE && tmpvar!=VISUAL_OBJECT_SEQUENCE_END_CODE) {
    BitstreamFlushBits(stream[vol_id], 8);
    tmpvar = BitstreamShowBits(stream[vol_id],32);
  }
  return;
}
 
Int
DecodeFGSVopHeader(Bitstream *stream, Vop *curr_vop, Trace *trace, 
     Int *vop_coding_type, Int VopTimeIncrementResolution
     /* M. van der Schaar - The last 3 parameters are added for FGST */)
{
  Int tmpvar,  bits;
  Int Nbits=0;

  time_base=0;
  tmpvar = (Int) BitstreamReadBitsSC(stream, FGS_VOP_START_CODE_LENGTH,
                                     "fgs_vop_start_code",trace,CODE);
   				     
  /* Philips Research; FGS-TS */
  if (tmpvar != FGS_VOP_START_CODE) {
    printf("FGS/FGST VOP does not start with FGS_VOP_START_CODE but with %d\n",tmpvar);
  } else {	     
    Nbits = FGS_VOP_START_CODE_LENGTH;
    fgs_global->fgs_vop_end=0;

    tmpvar = (Int) BitstreamReadBits(stream,2,"vop_coding_type",trace,FLAG);
    Nbits += 2;
    *vop_coding_type = tmpvar;
    PutVopPredictionType(tmpvar,curr_vop); 
   /* end FGST */
 
    tmpvar = (Int) BitstreamReadBits(stream,1,"modulo_time_base",trace,FLAG);
    Nbits += 1;
    while(tmpvar == 1) {
      tmpvar = (Int) BitstreamReadBits(stream,1,"modulo_time_base",trace,FLAG);
      Nbits += 1;
      time_base++;
    }

    /* marker_bit (1 bit) */
    tmpvar = (Int) BitstreamReadBits(stream,1,"marker_bit",trace,FLAG);
    Nbits += 1;

    /* Philips Research; FGS-TS */
    /* vop_time_increment (1-16 bits) */
    bits = ceil(log((double)VopTimeIncrementResolution)/log(2.0));
    /* end FGST */
  
    if (bits<1) bits=1;
    tmpvar=(Int)BitstreamReadBits(stream,bits,"vop_time_increment",trace,NUM);

    Nbits += bits;
    /* Philips Research; FGS-TS */
    vop_time_inc=tmpvar;
    /* end FGST */
    /* PutVopTimeInc(tmpvar, curr_vop ); comented out by M. van der Schaar */

    /* marker_bit (1 bit) */
    tmpvar = (Int) BitstreamReadBits(stream,1,"marker_bit",trace,FLAG);
    Nbits += 1;
  }

     /* Commented out by M. van der Schaar
    tmpvar = (Int) BitstreamReadBits(stream,6,"fgs_vop_quant",trace,NUM);
    Nbits += 6;
    PutVopQuantizer(tmpvar,curr_vop);*/

    tmpvar = (Int) BitstreamReadBits(stream,5,"fgs_vop_max_level_y",trace,NUM);
    Nbits += 5;
    fgs_global->fgs_vop_max_level_y = tmpvar;

    tmpvar = (Int) BitstreamReadBits(stream,5,"fgs_vop_max_level_u",trace,NUM);
    Nbits += 5;
    fgs_global->fgs_vop_max_level_u = tmpvar;

    tmpvar = (Int) BitstreamReadBits(stream,5,"fgs_vop_max_level_v",trace,NUM);
    Nbits += 5;
    fgs_global->fgs_vop_max_level_v = tmpvar;

    /* mofidied by M. van der Schaar, Nov. 27, 2000 */
    /* marker bit */
    /* marker_bit (1 bit) */
    tmpvar = (Int) BitstreamReadBits(stream,1,"marker_bit",trace,FLAG);
    Nbits += 1;
    tmpvar = (Int) BitstreamReadBits(stream,5,"fgs_vop_number_of_vop_bp_coded",
                                     trace,NUM);
    /* added M. van der Schaar*/
    fgs_global->fgs_vop_number_of_vop_bp_coded = tmpvar;
    /* end added M. van der Schaar*/				     
    Nbits += 5;

    tmpvar = (Int) BitstreamReadBits(stream,5,"fgs_vop_mc_bit_plane_not_used",
                                     trace,NUM);
    Nbits += 5;
    fgs_global->fgs_vop_mc_bit_plane_not_used = tmpvar;

    /* end modification M. van der Schaar, Nov. 27, 2000 */
    tmpvar = (Int) BitstreamReadBits(stream,1,
                                     "fgs_selective_enhancement_enable",
                                     trace,FLAG);
    Nbits += 1;
    fgs_global->fgs_selective_enhancement_enable = tmpvar;

    if (GetVopPredictionType(curr_vop)!=I_VOP) 
      if (GetVopInterlaced(curr_vop)) 
	{
	  tmpvar = BitstreamReadBits(stream,1,"TopFieldFirst",trace,FLAG);
	  PutVopTopFieldFirst(tmpvar, curr_vop); /* U. Benzler 001221 added interlaced FGS support */
	}

    
    if (fgs_global->FGST_flag == 0)
	BitstreamFlushBits(stream, stream->incnt % 8);
    return(Nbits);
}

Void
DecodeFGSVopCombinedMotionShapeTexture(Int vo_id, 
				       Trace *trace, 
				       Vop * rec_vop, Int short_video_header)
{
    Int 	      	ef=0, *error_flag = &ef; 
    Int                 mbnum, x_pos, y_pos;
    Int			MB_in_height, MB_in_width;
    Int                 rec_height, rec_width;
    Int                 B_in_width, B_in_height;
    Macroblock          *mblock;
    Image               *Y_rec, *U_rec, *V_rec;
    Int                 skipped_flag=0;
    UInt		new_vop;
    Int 		y_start, y_stop;
    Int interlace_info;
    Int num_macroblocks_in_gob, num_gobs_in_vop, gob_number, gob_num;
    Int newgob;

    /*****
     *
     *	Set some useful parameters
     *
     *****/

    rec_height = GetVopHeight (rec_vop);
    rec_width = GetVopWidth (rec_vop);
    MB_in_height = rec_height / MB_SIZE;
    MB_in_width = rec_width / MB_SIZE;
    B_in_width = 2 * MB_in_width;
    B_in_height = 2 * MB_in_height;

  
    /*****
     *
     *	Get the image fields of the VOP
     *
     *****/
    Y_rec = GetVopY (rec_vop);
    U_rec = GetVopU (rec_vop);
    V_rec = GetVopV (rec_vop);
  
    /*****
     *
     *	Get VOP quantizer and allocate MB structure
     *
     *****/

    mblock = (Macroblock *) malloc (sizeof (Macroblock));

    if (mblock == NULL)
    {
	printf (" Error allocating macroblock\n");
	exit(1);
    }

    /*****
     *
     *	Loop over all MBs, decode texture of each MB
     *
     *****/
    y_start = 0;
    y_stop = MB_in_height;
    newgob=0;
    new_vop = 1;
  
    if(short_video_header){
	gob_number = GetVopGobNumber(rec_vop);
	num_gobs_in_vop = GetVopNumGobsInVop(rec_vop);
	num_macroblocks_in_gob = GetVopNumMacroblocksInGob(rec_vop);
    }else{
	num_gobs_in_vop = 1;
	num_macroblocks_in_gob = MB_in_height * MB_in_width;
    }

    interlace_info = GetVopInterlaced(rec_vop);

    for(gob_num = 0;gob_num<num_gobs_in_vop;gob_num++){
	 
	 y_start = gob_num*num_macroblocks_in_gob/MB_in_width;
	 y_stop = (gob_num+1)*num_macroblocks_in_gob/MB_in_width;
	 
      for (y_pos = y_start; y_pos < y_stop; y_pos++) {
      for (x_pos = 0; x_pos < MB_in_width; x_pos++)
	{
	  mbnum = y_pos * MB_in_width + x_pos;
        
	  GetFGSMBblockdata(mbnum,trace,
			    mblock,&skipped_flag, error_flag, 
			    MB_in_width, interlace_info,
			    rec_vop,short_video_header);
	  
	  FillVop (mblock, mbnum, Y_rec, U_rec, V_rec);
	}    /* x */
      newgob=0;
	 }    /* y */
		gob_number++;
		PutVopGobNumber(gob_number,rec_vop);
	}/*gob_num*/	 
    
    free ((Char *)mblock);
    return;
}

Void
GetFGSMBblockdata(Int mbnum,
		  Trace *trace,
		  Macroblock * mblock,
		  Int * skipped_flag,
		  Int * error_flag,
		  Int MB_width,
		  Int field_dct,
		  Vop *rec_vop,
		  Int short_video_header)
{
  Int  comp;
  Int  DCT_block[B_SIZE * B_SIZE];
  Int  r_block[B_SIZE * B_SIZE];
  Int  r_block_2D[B_SIZE][B_SIZE];
  Int  diff[B_SIZE * B_SIZE], sign[B_SIZE * B_SIZE];
  Int  m, n;
  Int mode;
    
  MBzero (mblock);
 
  for (comp=0; comp<4; comp++)
  {
    bp2block(mbnum*6+comp,diff,sign);    
    Recoverbp(mbnum*6+comp,diff,sign,DCT_block);

    BlockIDCT(DCT_block, r_block_2D,GetVopBrightWhite(rec_vop));
	
    for (m = 0; m < B_SIZE; m++)
      for (n = 0; n < B_SIZE; n++)
      {
        r_block[m*B_SIZE+n] = r_block_2D[m][n];
	
      }
    
    
    PutBlock (comp, r_block, mblock);
  }
   
  if (field_dct && enh_block_data[mbnum*6].dct_type) {
    static Char inv_shuffle[] = {
      1,  0,  8,  1,  4,  8,  2,  4,  0,  2,  3,  0,  9,  3, 12,  9,  6, 12,
      0,  6,  5,  0, 10,  5,  0, 10,  7,  0, 11,  7, 13, 11, 14, 13,  0, 14
    };
    Int tmp[MB_SIZE];
    
    for (n = 0; n < sizeof inv_shuffle; n += 2)
      memcpy(inv_shuffle[n+1] ? &mblock->lum[inv_shuffle[n+1]][0] : &tmp[0],
	     inv_shuffle[n+0] ? &mblock->lum[inv_shuffle[n+0]][0] : &tmp[0],
	     MB_SIZE * sizeof(Int));
  }
  
  for (comp=4; comp<6; comp++)
  {
    bp2block(mbnum*6+comp,diff,sign);
    Recoverbp(mbnum*6+comp,diff,sign,DCT_block);
    BlockIDCT(DCT_block, r_block_2D,GetVopBrightWhite(rec_vop));
  
    for (m = 0; m < B_SIZE; m++)
      for (n = 0; n < B_SIZE; n++)
        r_block[m*B_SIZE+n] = r_block_2D[m][n];
    PutBlock (comp, r_block, mblock);
  }
  mode=enh_block_data[mbnum*6].mode;
#ifndef FGSPOSTCLIP  /* M. van der Schaar, Aug. 15, 2000 */
  if(mode==MODE_INTRA || mode==MODE_INTRA_Q)
    MB_clip (mblock, GetVopBrightWhite(rec_vop));
#endif
}

Void
Recoverbp(Int m, Int *diff, Int *sign,Int *DCT_block )
{
  Int i;
  Int block[64];
  Int decoded_level, rec_value;

  decoded_level = enh_block_data[m].decoded_level;
  rec_value = (decoded_level>2) ? (1<<(decoded_level-3)) : 0;

  for (i=0;i<64;i++)
  {
    if (diff[i]) diff[i] += rec_value;
    /* Frequency Weighting, Optivision */
    if ( fgs_global->fgs_frequency_weighting_enable )
      diff[i] >>= fgs_global->fgs_shift_matrix[i];
    /* Frequency Weighting, Optivision */
    /* Philips Research; SE */
    if (fgs_global->fgs_selective_enhancement_enable)
      diff[i] >>= shift_factor[m/6];
    /* end SE */
    #ifdef FGSPOSTCLIP /* M. van der Schaar, Aug. 15, 2000 */
      block[i] = sign[i]*diff[i];
    #else 
      block[i] = enh_block_data[m].sign[i] + sign[i]*diff[i]; 
    #endif  
  }

  for(i=0;i<64;i++){
    DCT_block[i]=block[zigzag[i]];
  }
}

Int input_enh_bitstream(Bitstream *stream,Trace *trace)
{
  Int i,max_level,max_level_y,max_level_u,max_level_v,Nbits, max_level_decode;
  Int tmpvar;

  max_level_y = fgs_global->fgs_vop_max_level_y;
  max_level_u = fgs_global->fgs_vop_max_level_u;
  max_level_v = fgs_global->fgs_vop_max_level_v;

  max_level = max_level_y;
  if ( max_level_u > max_level ) max_level = max_level_u;
  if ( max_level_v > max_level ) max_level = max_level_v;
  for ( i=0; i<fgs_global->Nblock; i++ ) {
    enh_block_data[i].data_level = -1;
    if ( (i%6)<4 ) {
      enh_block_data[i].decoded_level = max_level_y - 1;
    } else if ( (i%6)==4 ) {
      enh_block_data[i].decoded_level = max_level_u - 1;
    } else {
      enh_block_data[i].decoded_level = max_level_v - 1;
    }
  }

  Nbits = 0;
  i=0;
 
   /* added M. van der Schaar*/
  if (fgs_global->fgs_vop_number_of_vop_bp_coded > max_level)
        max_level_decode = max_level;  
  else
        max_level_decode = fgs_global->fgs_vop_number_of_vop_bp_coded;
  /* end added M. van der Schaar*/
  printf("!!max_level_decode =%d\n",max_level_decode); 
  while ( fgs_global->fgs_vop_end==0 && i < max_level_decode ) {
    /*if (i) - commented out by M. van der Schaar since now we have a resync marker before
    each bp */ { /* beginning of bp and not MSB */
      tmpvar = (Int) BitstreamReadBits(stream,FGS_VOP_BP_MARKER_HEADER_LENGTH,
                                       "fgs_vop_bp_marker_header",trace,CODE);
      Nbits += FGS_VOP_BP_MARKER_HEADER_LENGTH;
      tmpvar = (Int) BitstreamReadBits(stream,FGS_VOP_BP_MARKER_ID_LENGTH,
                                       "fgs_vop_bp_marker_id",trace,CODE);
      Nbits += FGS_VOP_BP_MARKER_ID_LENGTH;
    }
    Nbits += input_one_bp(stream,i,max_level,
			  max_level_y,max_level_u,max_level_v,
			  trace);
    if ( fgs_global->fgs_vop_end==0 ) Nbits += BitstreamByteAlign(stream);
      
    i += 1;
  }

  return(Nbits);
}


Int
input_one_bp(Bitstream *stream,Int level_index,Int max_level,
             Int max_level_y,Int max_level_u,Int max_level_v,Trace *trace)
{
  Int i, t, k, Nbits, start_decode;
  Int comb_index;
  Int tmpvar;
  /* Philips Research; SE */
  Int non_zero_macroblock;
  /* end SE */

  /* added by wu feng for FGS error resilience, MSRCN, 2000/8/20 */
  Int misalignement;
  UInt iValue;
  Int ifillValue[8] = { 0x00, 0x01, 0x03, 0x07,
						0x0f, 0x1f, 0x3f, 0x7f};

  comb_index = get_comb_index(max_level_y,max_level_u,max_level_v,
                              max_level-level_index);

  Nbits  = 0;
  i=0;
  while ( fgs_global->fgs_vop_end==0 && i<fgs_global->Nblock ) {
    if ( (i%6)<4 ) {
      start_decode = (level_index < (max_level-max_level_y)) ? 0 : 1;
    } else if ( (i%6)==4 ) {
      start_decode = (level_index < (max_level-max_level_u)) ? 0 : 1;
    } else {
      start_decode = (level_index < (max_level-max_level_v)) ? 0 : 1;
    }
    if ( (i%6)==0 && (level_index==0 || level_index==1) ) {
      Nbits += read_cbp_index(comb_index,level_index,i,stream,trace);
    }

    /* Philips Research; SE */
    if (fgs_global->fgs_selective_enhancement_enable && (i%6)==0)
    {
      if (level_index<2 &&
          enh_block_data[i].cbp&&enh_block_data[i+1].cbp&&
          enh_block_data[i+2].cbp&&enh_block_data[i+3].cbp&&
          enh_block_data[i+4].cbp&&enh_block_data[i+5].cbp)
        non_zero_macroblock=0;
      else
        non_zero_macroblock=1;

      if (!mb_shift_factor_received[i/6] && non_zero_macroblock)
      { /* shift_fact or not received yet and mb is not empty */
        shift_factor[i/6] = -1;
        do {
          tmpvar = (Int) BitstreamReadBits(stream,1,
                                           "fgs_shift_factor",trace,CODE);
          Nbits += 1;
          shift_factor[i/6] += 1;
        } while ( tmpvar==0 );
        mb_shift_factor_received[i/6]=1;
      }
    }
    /* end SE */

    if ( start_decode==1 ) {
      if ( enh_block_data[i].data_level == -1 ) {
        if ( level_index>1 ) {
	  t=(Int) BitstreamReadBits(stream,1,"cbp",trace,NUM);
          Nbits += 1;
          enh_block_data[i].cbp=t;
        }
        if(enh_block_data[i].cbp==0){
          if ( fgs_global->fgs_frequency_weighting_enable &&
               fgs_global->fgs_shift_matrix[0]>=2 ) {
            if ( fgs_global->fgs_shift_matrix[0]>=4 ) {
              Nbits += dec_one_block_bp(stream,i,0,table_bpc4fw_h[0],
                                        table_symbol4fw_h[0],trace);
            } else {
              Nbits += dec_one_block_bp(stream,i,0,table_bpc4fw_m[0],
                                        table_symbol4fw_m[0],trace);
            }
          } else {
            Nbits += dec_one_block_bp(stream,i,0,table_bpc[0],table_symbol[0],
                                      trace);
          }
          enh_block_data[i].data_level = max_level - level_index;
        }
      } else {
        t = enh_block_data[i].data_level - (max_level-level_index);
        k = ( t>3 ) ? 3 : t;
        if ( fgs_global->fgs_frequency_weighting_enable &&
             fgs_global->fgs_shift_matrix[0]>=2 ) {
          if ( fgs_global->fgs_shift_matrix[0]>=4 ) {
            Nbits += dec_one_block_bp(stream,i,t,table_bpc4fw_h[k],
                                      table_symbol4fw_h[k],trace);
          } else {
            Nbits += dec_one_block_bp(stream,i,t,table_bpc4fw_m[k],
                                      table_symbol4fw_m[k],trace);
          }
        } else {
          Nbits += dec_one_block_bp(stream,i,t,table_bpc[k],table_symbol[k],
                                    trace);
        }
      }
    }
    if ( fgs_global->fgs_vop_end==0 ) {
      enh_block_data[i].decoded_level = max_level - level_index;
	  	
	  /* added by MSRCN for FGS error resilience, 2000/8/20 */
	  if (!fgs_global->fgs_resync_marker_disable && i%6 == 5 
		  && i != fgs_global -> Nblock-1 )
	  {
		misalignement = 8-stream->bitcnt %8;
		if (misalignement == 0) misalignement = 8;
		iValue = BitstreamShowBits(stream, misalignement);
		tmpvar = BitstreamShowBits(stream, NUMBITS_FGS_VP_MARKER + misalignement);
		tmpvar = (tmpvar & 0x007fffff);
		if (tmpvar == 1 && iValue == ifillValue[misalignement - 1])
		{
			BitstreamFlushBits(stream, misalignement);
			/* read FGS VP marker */
			BitstreamReadBits(stream,NUMBITS_FGS_VP_MARKER,"fgs_error_resilience_marker",trace,CODE);
			/* read MB index */
			BitstreamReadBits(stream,fgs_global->fgs_number_bits_mb,"fgs_number_bits_mb",trace,CODE);
		}
	  }
      i += 1;
    }
  }
  printf("decoded level = %d and decoded block = %d\n",level_index,i-1);
  return(Nbits);
}

Int
dec_one_block_bp(Bitstream *stream,Int mb_index,Int level,
		 Int *table_one_bp,Int *symbols,Trace *trace)
{
  Int j, n, table_size, end, run, Nbits, mod_index, os, s;

  table_size=0;
  for ( j=0; j<3; j++ ) table_size += symbols[j];
  table_size += 1;   /* add the escape symbol */

  n = 0;
  end=0;
  Nbits = 0;
  os = 0;

  while ( end==0 ) {
    mod_index = decode_symbol(stream,table_one_bp,table_size,trace);
    if ( fgs_global->fgs_vop_end==1 ) break;
    Nbits += table_one_bp[2*mod_index+1];
    if ( mod_index<symbols[0] ) { /* not end of plane */
      enh_block_data[mb_index].run[level][n] = mod_index;
      enh_block_data[mb_index].end[level][n] = 0;
      os += enh_block_data[mb_index].run[level][n]+1;
      n += 1;
      end = 0;
      
      if ( enh_block_data[mb_index].residue_sign[os-1]==0 ) {
	s=(Int) BitstreamReadBits(stream,1,"s",trace,NUM);
        if ( fgs_global->fgs_vop_end==1 ) break;
	enh_block_data[mb_index].residue_sign[os-1] = (s==0) ? 1 : -1;
        Nbits += 1;
      }
    } else if ( mod_index<symbols[0]+symbols[1] ) { /* end of plane */
      enh_block_data[mb_index].run[level][n] = mod_index-symbols[0];
      enh_block_data[mb_index].end[level][n] = 1;
      os += enh_block_data[mb_index].run[level][n]+1;
      n += 1;
      end = 1;
      if ( enh_block_data[mb_index].residue_sign[os-1]==0 ) {
	s=(Int) BitstreamReadBits(stream,1,"s",trace,NUM);
        if ( fgs_global->fgs_vop_end==1 ) break;
	enh_block_data[mb_index].residue_sign[os-1] = (s==0) ? 1 : -1;
        Nbits += 1;
      }
    } else if ( mod_index<symbols[0]+symbols[1]+symbols[2] ) { /* all zero */
      enh_block_data[mb_index].run[level][0] = 63;
      enh_block_data[mb_index].end[level][0] = 0;
      end = 1;
    } else { /* escape case */
      run=(Int) BitstreamReadBits(stream,6,"run",trace,NUM);
      if ( fgs_global->fgs_vop_end==1 ) break;
      end=(Int) BitstreamReadBits(stream,1,"end",trace,NUM);
      if ( fgs_global->fgs_vop_end==1 ) break;
      enh_block_data[mb_index].run[level][n] = run;
      enh_block_data[mb_index].end[level][n] = end;
      Nbits += 7;
      os += run+1;
      n += 1;
      if ( enh_block_data[mb_index].residue_sign[os-1]==0 ) {
	s=(Int) BitstreamReadBits(stream,1,"s",trace,NUM);
        if ( fgs_global->fgs_vop_end==1 ) break;
	enh_block_data[mb_index].residue_sign[os-1] = (s==0) ? 1 : -1;
        Nbits += 1;
      }
    }
    if ( os>64 ) {
      printf("total length = %d > 64 !!!\n",os);
      exit(1);
    }
  }

  if ( fgs_global->fgs_vop_end==1 ) {
    enh_block_data[mb_index].run[level][0] = 63;
    enh_block_data[mb_index].end[level][0] = 0;
  }

  return(Nbits);
}

Void FGSMBzero(Int mbnum,Int dct_type,Int mode)
{
  Int comp;

  for(comp=0;comp<6;comp++){
    enh_block_data[mbnum*6+comp].dct_type=dct_type;
    enh_block_data[mbnum*6+comp].mode=mode;
  }
}

Void
FGSPutBlock(Int comp, Int mbnum, Int  *DCT_block, Int dct_type, Int mode)
{
  Int i;

  enh_block_data[mbnum*6+comp].dct_type=dct_type;
  enh_block_data[mbnum*6+comp].mode=mode;
  for(i=0;i<64;i++)
    enh_block_data[mbnum*6+comp].sign[zigzag[i]]=DCT_block[i];
}

void bp2block(Int m, Int *diff, Int *sign)
{
  Int i, j, level, x[64];
  Int value;

  level = enh_block_data[m].data_level;
  for ( i=0; i<64; i++ ) diff[i] = 0;
  for ( j=0; j<level; j++ ) {
    one_bp2block(m, x,j);
    for ( i=0; i<64; i++ ) {
      diff[i] = (diff[i]<<1)+x[i];
    }
  }

  for ( i=0; i<64; i++ ) {
    sign[i] = enh_block_data[m].residue_sign[i];
  }
}

void one_bp2block(Int m, Int *out,Int level)
{
  Int i, j, n, run[64], os;

  if ( enh_block_data[m].run[level][0] == 63 &&
       enh_block_data[m].end[level][0] == 0 ) { /* all zero case */
    n = 0;
  } else {
    n = 0;
    while ( enh_block_data[m].end[level][n]==0 && n<64 ) n += 1;
    if ( n>=64 ) {
      printf("something is wrong: n = %d, m = %d, level = %d\n",n,m,level);
      exit(1);
    }
    n += 1;
  }
  for(i=0;i<n;i++){
    run[i] = enh_block_data[m].run[level][i];
  }
  if ( n==0 ) {
    for ( j=0; j<64; j++ ) out[j] = 0;
  } else {
    os = 0;
    for ( i=0; i<n; i++ ) {
      for ( j=0; j<run[i]; j++ ) out[os+j] = 0;
      out[os+j] = 1;
      os += run[i]+1;
    }
    for ( i=os; i<64; i++ ) out[i] = 0;
  }
}

Int decode_symbol(Bitstream *stream, Int *vlc_table,Int size,Trace *trace)
{
  Int i, j=0, k;
  Int n;
  Int min, max;
  Int vlc_code;
       
  min=1;max=17;
	  
  n=BitstreamShowBits(stream,max); 
  fgs_global->fgs_vop_end=0;
  for(i=min;i<=max;i++){
    k=n>>(max-i);
    for(j=0;j<size;j++)
      if(k==vlc_table[2*j] && i==vlc_table[2*j+1]) break;
    if(j!=size) break;
  }
  if(i==max+1){
    fgs_global->fgs_vop_end=1;
    return(0);
  }

  vlc_code = (Int) BitstreamReadBits(stream,i,"vlc_code",trace,CODE);
  if ( vlc_code != vlc_table[2*j] ) {
    printf("Something Wrong!!!\n");
    exit(-1);
  }

  return(j);
}

/* FGS CBP (Optivision) */

Int
get_comb_index(Int max_level_y,Int max_level_u,Int max_level_v,Int remain_level)
{
  Int comb_index;

  if ( max_level_y>=remain_level &&
       max_level_u>=remain_level &&
       max_level_v>=remain_level ) {
    comb_index = 0;
  } else if ( max_level_y>=remain_level &&
              max_level_u>=remain_level &&
              max_level_v<remain_level ) {
    comb_index = 1;
  } else if ( max_level_y>=remain_level &&
              max_level_u<remain_level &&
              max_level_v>=remain_level ) {
    comb_index = 2;
  } else if ( max_level_y>=remain_level &&
              max_level_u<remain_level &&
              max_level_v<remain_level ) {
    comb_index = 3;
  } else if ( max_level_y<remain_level &&
              max_level_u>=remain_level &&
              max_level_v>=remain_level ) {
    comb_index = 4;
  } else if ( max_level_y<remain_level &&
              max_level_u>=remain_level &&
              max_level_v<remain_level ) {
    comb_index = 5;
  } else if ( max_level_y<remain_level &&
              max_level_u<remain_level &&
              max_level_v>=remain_level ) {
    comb_index = 6;
  } else {
    comb_index = 7;
    printf("This cannot happen. Something is wrong!!!\n");
    exit(1);
  }
  return(comb_index);
}

Int
read_cbp_index(Int comb_index,Int level_index,Int block_index,
               Bitstream *stream,Trace *trace)
{
  Int Nbits, b;
  Int cbp_table_3[4] = {7, 11, 13, 14};
  Int cbp_table_2[6] = {3, 5, 6, 9, 10, 12};
  Int cbp_table_1[4] = {8, 4, 2, 1};

  Nbits = 0;
  if ( level_index==0 ) 
  {
    switch(comb_index) {
      case 0: b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
              Nbits += 1;
              if ( b==1 ) {
                use_code[block_index/6] = 1;
                enh_block_data[block_index].cbp = 1;
                enh_block_data[block_index+1].cbp = 1;
                enh_block_data[block_index+2].cbp = 1;
                enh_block_data[block_index+3].cbp = 1;
                enh_block_data[block_index+4].cbp = 1;
                enh_block_data[block_index+5].cbp = 1;
              } else {
                use_code[block_index/6] = 0;
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index].cbp =
                  (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                enh_block_data[block_index+1].cbp =
                  (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                enh_block_data[block_index+2].cbp =
                  (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                enh_block_data[block_index+3].cbp =
                  (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 4;
                if ( b==1 ) {
                  enh_block_data[block_index+4].cbp = 1;
                  enh_block_data[block_index+5].cbp = 1;
                } else {
                  b = (Int) BitstreamReadBits(stream,2,"cbp_code",trace,NUM);
                  Nbits += 2;
                  enh_block_data[block_index+4].cbp = 0x1 & ((b-1)>>1);
                  enh_block_data[block_index+5].cbp = 0x1 & (b-1);
                }
              }
              break;
      case 1: b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
              Nbits += 1;
              if ( b==1 ) {
                use_code[block_index/6] = 1;
                enh_block_data[block_index].cbp = 1;
                enh_block_data[block_index+1].cbp = 1;
                enh_block_data[block_index+2].cbp = 1;
                enh_block_data[block_index+3].cbp = 1;
                enh_block_data[block_index+4].cbp = 1;
              } else {
                use_code[block_index/6] = 0;
                b = (Int) BitstreamReadBits(stream,5,"cbp_code",trace,NUM);
                Nbits += 5;
                enh_block_data[block_index].cbp = 0x1 & ((b-1)>>4);
                enh_block_data[block_index+1].cbp = 0x1 & ((b-1)>>3);
                enh_block_data[block_index+2].cbp = 0x1 & ((b-1)>>2);
                enh_block_data[block_index+3].cbp = 0x1 & ((b-1)>>1);
                enh_block_data[block_index+4].cbp = 0x1 & (b-1);
              }
              break;
      case 2: b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
              Nbits += 1;
              if ( b==1 ) {
                use_code[block_index/6] = 1;
                enh_block_data[block_index].cbp = 1;
                enh_block_data[block_index+1].cbp = 1;
                enh_block_data[block_index+2].cbp = 1;
                enh_block_data[block_index+3].cbp = 1;
                enh_block_data[block_index+5].cbp = 1;
              } else {
                use_code[block_index/6] = 0;
                b = (Int) BitstreamReadBits(stream,5,"cbp_code",trace,NUM);
                Nbits += 5;
                enh_block_data[block_index].cbp = 0x1 & ((b-1)>>4);
                enh_block_data[block_index+1].cbp = 0x1 & ((b-1)>>3);
                enh_block_data[block_index+2].cbp = 0x1 & ((b-1)>>2);
                enh_block_data[block_index+3].cbp = 0x1 & ((b-1)>>1);
                enh_block_data[block_index+5].cbp = 0x1 & (b-1);
              }
              break;
      case 3: b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
              Nbits += 1;
              if ( b==1 ) {
                use_code[block_index/6] = 1;
                enh_block_data[block_index].cbp = 1;
                enh_block_data[block_index+1].cbp = 1;
                enh_block_data[block_index+2].cbp = 1;
                enh_block_data[block_index+3].cbp = 1;
              } else {
                use_code[block_index/6] = 0;
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                if ( b==1 ) {
                  b = (Int) BitstreamReadBits(stream,2,"cbp_code",trace,NUM);
                  Nbits += 2;
                  enh_block_data[block_index].cbp = 0x1&(cbp_table_3[b]>>3);
                  enh_block_data[block_index+1].cbp = 0x1&(cbp_table_3[b]>>2);
                  enh_block_data[block_index+2].cbp = 0x1&(cbp_table_3[b]>>1);
                  enh_block_data[block_index+3].cbp = 0x1&cbp_table_3[b];
                } else {
                  b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                  Nbits += 1;
                  if ( b==1 ) {
                    b = (Int) BitstreamReadBits(stream,3,"cbp_code",trace,NUM);
                    Nbits += 3;
                    enh_block_data[block_index].cbp = 0x1&(cbp_table_2[b]>>3);
                    enh_block_data[block_index+1].cbp = 0x1&(cbp_table_2[b]>>2);
                    enh_block_data[block_index+2].cbp = 0x1&(cbp_table_2[b]>>1);
                    enh_block_data[block_index+3].cbp = 0x1&cbp_table_2[b];
                  } else {
                    b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                    Nbits += 1;
                    if ( b==1 ) {
                      b=(Int)BitstreamReadBits(stream,2,"cbp_code",trace,NUM);
                      Nbits += 2;
                      enh_block_data[block_index].cbp = 0x1&(cbp_table_1[b]>>3);
                      enh_block_data[block_index+1].cbp=0x1&(cbp_table_1[b]>>2);
                      enh_block_data[block_index+2].cbp=0x1&(cbp_table_1[b]>>1);
                      enh_block_data[block_index+3].cbp=0x1&cbp_table_1[b];
                    } else {
                      b=(Int)BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                      Nbits += 1;
                      enh_block_data[block_index].cbp = 0;
                      enh_block_data[block_index+1].cbp = 0;
                      enh_block_data[block_index+2].cbp = 0;
                      enh_block_data[block_index+3].cbp = 0;
                    }
                  }
                }
              }
              break;
      case 4: b = (Int) BitstreamReadBits(stream,2,"cbp_code",trace,NUM);
              Nbits += 2;
              enh_block_data[block_index+4].cbp = 0x1 & (b>>1);
              enh_block_data[block_index+5].cbp = 0x1 & b;
              if ( b==3 ) {
                use_code[block_index/6] = 1;
              } else {
                use_code[block_index/6] = 0;
              }
              break;
      case 5: b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
              Nbits += 1;
              enh_block_data[block_index+4].cbp = b;
              use_code[block_index/6] = b;
              break;
      case 6: b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
              Nbits += 1;
              enh_block_data[block_index+5].cbp = b;
              use_code[block_index/6] = b;
              break;
      default: printf("Invalid combination!\n");
               exit(1);
    }
  } else if ( use_code[block_index/6]==1 ) {
    switch(comb_index) {
      case 0: b = (Int) BitstreamReadBits(stream,3,"cbp_code",trace,NUM);
              Nbits += 3;
              if ( b==0 ) {
                b = (Int) BitstreamReadBits(stream,6,"cbp_code",trace,NUM);
                Nbits += 6;
                enh_block_data[block_index].cbp = 0x1 & (b>>5);
                enh_block_data[block_index+1].cbp = 0x1 & (b>>4);
                enh_block_data[block_index+2].cbp = 0x1 & (b>>3);
                enh_block_data[block_index+3].cbp = 0x1 & (b>>2);
                enh_block_data[block_index+4].cbp = 0x1 & (b>>1);
                enh_block_data[block_index+5].cbp = 0x1 & b;
              } else if ( b==1 ) {
                b = (Int) BitstreamReadBits(stream,2,"cbp_code",trace,NUM);
                Nbits += 2;
                if ( b<2 ) {
                  enh_block_data[block_index].cbp = 0;
                  enh_block_data[block_index+1].cbp = 0;
                  enh_block_data[block_index+2].cbp = 0;
                  enh_block_data[block_index+3].cbp = 0;
                } else {
                  enh_block_data[block_index].cbp = 1;
                  enh_block_data[block_index+1].cbp = 1;
                  enh_block_data[block_index+2].cbp = 1;
                  enh_block_data[block_index+3].cbp = 1;
                }
                enh_block_data[block_index+4].cbp = 0x1 & b;
                enh_block_data[block_index+5].cbp = 1 - (0x1 & b);
              } else if ( b==2 ) {
                enh_block_data[block_index].cbp = 0;
                enh_block_data[block_index+1].cbp = 0;
                enh_block_data[block_index+2].cbp = 0;
                enh_block_data[block_index+3].cbp = 0;
                enh_block_data[block_index+4].cbp = 1;
                enh_block_data[block_index+5].cbp = 1;
              } else if ( b==3 ) {
                enh_block_data[block_index].cbp = 1;
                enh_block_data[block_index+1].cbp = 1;
                enh_block_data[block_index+2].cbp = 1;
                enh_block_data[block_index+3].cbp = 1;
                enh_block_data[block_index+4].cbp = 1;
                enh_block_data[block_index+5].cbp = 1;
              } else {
                b <<= 2;
                b |= (Int) BitstreamReadBits(stream,2,"cbp_code",trace,NUM);
                Nbits += 2;
                enh_block_data[block_index].cbp = 0x1 & (b>>3);
                enh_block_data[block_index+1].cbp = 0x1 & (b>>2);
                enh_block_data[block_index+2].cbp = 0x1 & (b>>1);
                enh_block_data[block_index+3].cbp = 0x1 & b;
                if ( b==16 || b==31 ) {
                  enh_block_data[block_index+4].cbp = 0;
                  enh_block_data[block_index+5].cbp = 0;
                } else {
                  enh_block_data[block_index+4].cbp = 1;
                  enh_block_data[block_index+5].cbp = 1;
                }
              }
              break;
      case 1: b = (Int) BitstreamReadBits(stream,5,"cbp_code",trace,NUM);
              Nbits += 5;
              enh_block_data[block_index].cbp = 0x1 & (b>>4);
              enh_block_data[block_index+1].cbp = 0x1 & (b>>3);
              enh_block_data[block_index+2].cbp = 0x1 & (b>>2);
              enh_block_data[block_index+3].cbp = 0x1 & (b>>1);
              enh_block_data[block_index+4].cbp = 0x1 & b;
              break;
      case 2: b = (Int) BitstreamReadBits(stream,5,"cbp_code",trace,NUM);
              Nbits += 5;
              enh_block_data[block_index].cbp = 0x1 & (b>>4);
              enh_block_data[block_index+1].cbp = 0x1 & (b>>3);
              enh_block_data[block_index+2].cbp = 0x1 & (b>>2);
              enh_block_data[block_index+3].cbp = 0x1 & (b>>1);
              enh_block_data[block_index+5].cbp = 0x1 & b;
              break;
      case 3: b = (Int) BitstreamReadBits(stream,4,"cbp_code",trace,NUM);
              Nbits += 4;
              enh_block_data[block_index].cbp = 0x1 & (b>>3);
              enh_block_data[block_index+1].cbp = 0x1 & (b>>2);
              enh_block_data[block_index+2].cbp = 0x1 & (b>>1);
              enh_block_data[block_index+3].cbp = 0x1 & b;
              break;
      case 4: b = (Int) BitstreamReadBits(stream,2,"cbp_code",trace,NUM);
              Nbits += 2;
              enh_block_data[block_index+4].cbp = 0x1 & (b>>1);
              enh_block_data[block_index+5].cbp = 0x1 & b;
              break;
      case 5: b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
              Nbits += 1;
              enh_block_data[block_index+4].cbp = b;
              break;
      case 6: b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
              Nbits += 1;
              enh_block_data[block_index+5].cbp = b;
              break;
      default: printf("Invalid combination!\n");
               exit(1);
    }
  } else {
    switch(comb_index) {
      case 0: if ( enh_block_data[block_index].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index].cbp = b;
              }
              if ( enh_block_data[block_index+1].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index+1].cbp = b;
              }
              if ( enh_block_data[block_index+2].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index+2].cbp = b;
              }
              if ( enh_block_data[block_index+3].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index+3].cbp = b;
              }
              if ( enh_block_data[block_index+4].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index+4].cbp = b;
              }
              if ( enh_block_data[block_index+5].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index+5].cbp = b;
              }
              break;
      case 1: if ( enh_block_data[block_index].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index].cbp = b;
              }
              if ( enh_block_data[block_index+1].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index+1].cbp = b;
              }
              if ( enh_block_data[block_index+2].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index+2].cbp = b;
              }
              if ( enh_block_data[block_index+3].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index+3].cbp = b;
              }
              if ( enh_block_data[block_index+4].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index+4].cbp = b;
              }
              break;
      case 2: if ( enh_block_data[block_index].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index].cbp = b;
              }
              if ( enh_block_data[block_index+1].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index+1].cbp = b;
              }
              if ( enh_block_data[block_index+2].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index+2].cbp = b;
              }
              if ( enh_block_data[block_index+3].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index+3].cbp = b;
              }
              if ( enh_block_data[block_index+5].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index+5].cbp = b;
              }
              break;
      case 3: if ( enh_block_data[block_index].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index].cbp = b;
              }
              if ( enh_block_data[block_index+1].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index+1].cbp = b;
              }
              if ( enh_block_data[block_index+2].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index+2].cbp = b;
              }
              if ( enh_block_data[block_index+3].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index+3].cbp = b;
              }
              break;
      case 4: if ( enh_block_data[block_index+4].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index+4].cbp = b;
              }
              if ( enh_block_data[block_index+5].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index+5].cbp = b;
              }
              break;
      case 5: if ( enh_block_data[block_index+4].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index+4].cbp = b;
              }
              break;
      case 6: if ( enh_block_data[block_index+5].data_level == -1 ) {
                b = (Int) BitstreamReadBits(stream,1,"cbp_code",trace,NUM);
                Nbits += 1;
                enh_block_data[block_index+5].cbp = b;
              }
              break;
      default: printf("Invalid combination!\n");
               exit(1);
    }
  }
  return(Nbits);
}

/* FGS CBP (Optivision) */
