/******************************************************************************
 *                                                                          
 * This software module was originally developed by 
 *
 *    Weiping Li (Optivision, Inc.) and
 *    Fan Ling (General Instrument)                   	
 *
 * and edited by
 *    Yingwei Chen - SE (Philips Research)
 *    Mihaela van der Schaar - FGST (Philips Research, Mihaela.vanderschaar@philips.com)
 *    Feng Wu - FGS enhancemcent resilience (Microsoft Research China, fengwu@microsoft.com)
 *
 * in the course of development of the MPEG-4 Video (ISO/IEC 14496-2) standard.
 * This software module is an implementation of a part of one or more MPEG-4 
 * Video (ISO/IEC 14496-2) tools as specified by the MPEG-4 Video (ISO/IEC 
 * 14496-2) standard. 
 *
 * ISO/IEC gives users of the MPEG-4 Video (ISO/IEC 14496-2) standard free 
 * license to this software module or modifications thereof for use in hardware
 * or software products claiming conformance to the MPEG-4 Video (ISO/IEC 
 * 14496-2) standard. 
 *
 * Those intending to use this software module in hardware or software products
 * are advised that its use may infringe existing patents. The original 
 * developer of this software module and his/her company, the subsequent 
 * editors and their companies, and ISO/IEC have no liability for use of this 
 * software module or modifications thereof in an implementation. Copyright is 
 * not released for non MPEG-4 Video (ISO/IEC 14496-2) standard conforming 
 * products. 
 *
 * ACTS-MoMuSys partners retain full right to use the code for his/her own 
 * purpose, assign or donate the code to a third party and to inhibit third 
 * parties from using the code for non MPEG-4 Video (ISO/IEC 14496-2) standard
 * conforming products. This copyright notice must be included in all copies or
 * derivative works. 
 *
 * Copyright (c) 1997
 *
 *****************************************************************************/
/***********************************************************HeaderBegin*******
 *                                                                         
 * File:        bp_fgs_enc.c
 *
 * Author:      Weiping Li, Optivision
                Fan Ling, General Instrument
 * Created:     
 *                                                                         
 * Description:
 *
 * Notes:
 *
 * Modified:
 *     August-06-99 Yingwei Chen: Changed FGS enhancement layer residual
 *                                computation and coding.
 *
 ***********************************************************HeaderEnd*********/

/************************    INCLUDE FILES    ********************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h> 
#include "vm_config.h"
#include "momusys.h"
#include "mom_vop.h"
#include "mom_bitstream_i.h"
#include "mom_access.h"
#include "io_generic.h"
#include "putvlc.h"
#include "text_code_mb.h"
#include "vm_vop_code.h"
#include "vm_compos.h"
#include "vm_stats.h"
#include "newpred.h"	/** added for NEWPRED (Oki) 16-AUG-1999 **/
#include "mom_putbits.h"
#include "vm_enc_main.h"
#include "vm_vop_bound.h"
#include "mom_vol.h"
#include "mom_vo.h"
#include "mot_util.h"
#include "zigzag.h"
#include "post_filter.h"
#include "startcode.h"

#include "sait_rc.h"
#include "do_bgc.h"

#include "enhance_vop.h"
#include "bgc.h"

#include "vlcbp.h"
#include "vlcbp4fw.h"
#include "bp_fgs_enh.h"
#include "bp_fgs_enc.h"

static ENH_BLOCK *enh_block_data;
static Int use_code[1620];
FGS_GLOBAL *fgs_global;

/* Philips Research; SE */
Int shift_factor[720*576/256];
Int mb_shift_factor_sent[720*576/256];
/* end SE */
/* Philips Research; FGST */
Int fgs_vop_max_level, fgs_vop_max_level_y, fgs_vop_max_level_u, fgs_vop_max_level_v, fgs_vop_mc_bit_plane_not_used;
/* end FGST */

Void
fgs_enc_init(Vop *curr_vop)
{
  Int nblock = 6*GetVopWidth(curr_vop)*GetVopHeight(curr_vop)/(16*16);
  Int i; /* Philips SE */
  printf("Initialize FGS data array....\n");

  if(enh_block_data==NULL)
    enh_block_data = (ENH_BLOCK *)malloc(nblock*sizeof(ENH_BLOCK));
  fgs_global->Nblock = 0;
  fgs_global->fgs_co_located_not_coded = 0;

  /* Philips Research */
  /* SE */
  if (fgs_global->fgs_selective_enhancement_enable)
  {
    if (fgs_global->fgs_mask_type==0) /* rectangular */
    {
      fill_fgs_rectangular_mask(GetVopWidth(curr_vop),GetVopHeight(curr_vop),
                                fgs_global->fgs_mask_start_x,
                                fgs_global->fgs_mask_start_y,
                                fgs_global->fgs_mask_end_x,
                                fgs_global->fgs_mask_end_y,
                                fgs_global->fgs_rectangular_shift_factor);
    }
    else /* from file */
    {
      get_fgs_mask_from_file(fgs_global->fgs_mask_file,
                             GetVopWidth(curr_vop),GetVopHeight(curr_vop));
    }
    for (i=0;i<nblock/6;i++)
      mb_shift_factor_sent[i]=0; /* to signal that shift factor for mb has not been sent yet */
  }
  /* end SE */
}

/*  Philips Research; FGST */
/* This function was originally part of FGSVopCode */
Int
EncodeFGSVopHeader(Int vo_id, Int vol_id, Int vop_coding_type)
{
  Int fgs_vop_quantizer=1;
  Int fgs_vop_mc_bit_plane_not_used;
  Int i, Nbits;
  Image *fgs_bitstream;
  Int bits, idx, time_increment_resolution;
  Float time, time_modulo, time_inc;

  idx = fgs_global->index;
  time_increment_resolution = fgs_global->time_increment_resolution;
  time = fgs_global->time;

  fgs_bitstream = BitstreamInit();

  BitstreamPutBits(fgs_bitstream,FGS_VOP_START_CODE,FGS_VOP_START_CODE_LENGTH);
  Nbits = FGS_VOP_START_CODE_LENGTH;
  BitstreamPutBits(fgs_bitstream,vop_coding_type,2);
  Nbits += 2;
  
  if (vop_coding_type == 0)  /* if FGS */
  {
     time_modulo = time - idx*1000;
     while(time_modulo >= 1000) {
       BitstreamPutBits(fgs_bitstream,1,1);
       Nbits += 1;
       time_modulo = time_modulo - 1000;
       idx += 1;
     }
     BitstreamPutBits(fgs_bitstream,0,1);
     Nbits += 1;

     time_inc = time - idx*1000;

     bits = ceil(log((double)time_increment_resolution)/log(2.0));
     if (bits<1) bits=1;
     time_inc = time_inc*time_increment_resolution/1000.0;

     /* marker bit */
     BitstreamPutBits(fgs_bitstream,1,1);
     Nbits += 1;

     BitstreamPutBits(fgs_bitstream,(Int)(time_inc+0.001),bits);
     Nbits += bits;

     /* marker bit */
     BitstreamPutBits(fgs_bitstream,1,1);
     Nbits += 1;
  }
  
  find_max_level(&fgs_vop_max_level_y,
                 &fgs_vop_max_level_u,
                 &fgs_vop_max_level_v);
  fgs_vop_max_level = fgs_vop_max_level_y;
  
  if ( fgs_vop_max_level_u > fgs_vop_max_level )
    fgs_vop_max_level = fgs_vop_max_level_u;
  if ( fgs_vop_max_level_v > fgs_vop_max_level )
    fgs_vop_max_level = fgs_vop_max_level_v;

  fgs_vop_mc_bit_plane_not_used = fgs_vop_max_level;

  /*BitstreamPutBits(fgs_bitstream, fgs_vop_quantizer, 6);
  Nbits += 6;
  commented out by M. van der Schaar*/
  if (vop_coding_type == 0) /* if FGS */
  {
     BitstreamPutBits(fgs_bitstream,fgs_vop_max_level_y,5);
     Nbits += 5;
     BitstreamPutBits(fgs_bitstream,fgs_vop_max_level_u,5);
     Nbits += 5;
     BitstreamPutBits(fgs_bitstream,fgs_vop_max_level_v,5);
     Nbits += 5;
     /* mofidied by M. van der Schaar, Nov. 27, 2000 */
     /* marker bit */
     BitstreamPutBits(fgs_bitstream,1,1);
     Nbits += 1;
     BitstreamPutBits(fgs_bitstream,fgs_global->fgs_vop_number_of_vop_bp_coded,5);
     printf("!!number_of_vop_bp_coded=%d\n",fgs_global->fgs_vop_number_of_vop_bp_coded);
     Nbits += 5;
     BitstreamPutBits(fgs_bitstream,fgs_global->fgs_vop_mc_bit_plane_not_used,5);
     Nbits += 5;
     /* end modification M. van der Schaar, Nov. 27, 2000 */
     /* Philips Research; SE */
     BitstreamPutBits(fgs_bitstream,fgs_global->fgs_selective_enhancement_enable,1);
     /* end SE */
     Nbits += 1;     
  }
  BitstreamPut(fgs_bitstream,vo_id,vol_id+10);
  
  if(vop_coding_type == 0) /* if FGS, this is the end of the header */ 
    Nbits += NextStartCode(vo_id,vol_id+10);
  BitstreamFree(fgs_bitstream);
  return(Nbits);
}
/* end FGST */
/***********************************************************
 *
 * -- FGSVopCode -- Coding FGS enhancement VOP
 *
 * Author :
 *	Weiping Li, Optivision
 *
 * Created :
 * 	12-18-98
 *
 * Purpose :
 * 
 * Arguments in : 
 *
 * Arguments in/out :
 *
 * Arguments out :
 *
 * Return values :
 * none
 *
 * Side effects :
 * -
 *
 * Description :
 * -
 *
 * See also :
 *
 * Modified : M. van der Schaar, to incorporate FGST-syntax.
 *
 ***********************************************************/
Int
FGSVopCode(Int vo_id, Int vol_id)
{
  Int fgs_vop_quantizer=1;
  Int fgs_vop_mc_bit_plane_not_used;
  Int i, Nbits=0;
  Int max_level /* added M. van der Schaar*/;
  Image *fgs_bitstream;

  fgs_bitstream = BitstreamInit();

  for ( i=0; i<fgs_global->Nblock; i++ ) enh_block_data[i].cbp=1;    

  /* added by MSRCN for FGS error resilience, 2000/8/20 */
  fgs_global->fgs_number_bits_mb = (int) ceil( log10((double)(fgs_global->Nblock/6)) / log10(2.0));

  /* added M. van der Schaar*/
  if (fgs_global->fgs_vop_number_of_vop_bp_coded > fgs_vop_max_level)
        max_level = fgs_vop_max_level;  
  else
        max_level = fgs_global->fgs_vop_number_of_vop_bp_coded;
  /* end added M. van der Schaar*/
  for ( i=0; i<max_level; i++ ) {
    /* if (i) commented out by M. van der Schaar since now a resync marker is needed prior to each bp */
    { /* beginning of bitplane and not MSB */
      BitstreamPutBits(fgs_bitstream,FGS_VOP_BP_MARKER_HEADER,
                       FGS_VOP_BP_MARKER_HEADER_LENGTH);
      Nbits += FGS_VOP_BP_MARKER_HEADER_LENGTH;
      BitstreamPutBits(fgs_bitstream,i/*-1 M. van der Schaar*/,FGS_VOP_BP_MARKER_ID_LENGTH);
      Nbits += FGS_VOP_BP_MARKER_ID_LENGTH;
    }
    Nbits += output_one_bp(i,fgs_vop_max_level,
                           fgs_vop_max_level_y,
                           fgs_vop_max_level_u,
                           fgs_vop_max_level_v,
                           fgs_bitstream);
/* we assume volID "vol_id+10" is not used by other vols, and 
   vol_id+10< MAX_NUM_VOLS 
*/
    BitstreamPut(fgs_bitstream,vo_id,vol_id+10);
    Nbits += NextStartCode(vo_id,vol_id+10);
    BitstreamFree(fgs_bitstream);
    fgs_bitstream = BitstreamInit();
  }

  BitstreamFree(fgs_bitstream);
  return(Nbits);
}

Void
find_max_level(Int *mx_y, Int *mx_u, Int *mx_v)
{
  Int i;

  *mx_y = *mx_u = *mx_v = 0;
  for ( i=0; i<fgs_global->Nblock; i++ ) {
    if ( (i%6)<4 ) {
      if ( enh_block_data[i].data_level > *mx_y )
        *mx_y = enh_block_data[i].data_level;
    } else if ( (i%6)==4 ) {
      if ( enh_block_data[i].data_level > *mx_u )
        *mx_u = enh_block_data[i].data_level;
    } else {
      if ( enh_block_data[i].data_level > *mx_v )
        *mx_v = enh_block_data[i].data_level;
    }
  }
}

Int
output_one_bp(Int level_index,Int max_level,
              Int max_level_y,Int max_level_u,Int max_level_v,
              Image *fgs_bitstream)
{
  Int i, j, t, k, Nbits, start_encode;
  Int comb_index, cbp_index;
  /* Philips Research; SE */
  Int non_zero_macroblock;
  /* end SE */
  Int iVPCounter;  /* added by MSRCN for FGS error resilience, 2000/8/18*/

  comb_index = get_comb_index(max_level_y,max_level_u,max_level_v,
                              max_level-level_index);

  iVPCounter = 0;  /* added by MSRCN for FGS error resilience, 2000/8/18 */
  Nbits  = 0;
  for ( i=0; i<fgs_global->Nblock; i++ ) {
    if ( (i%6)<4 ) {
      start_encode = (level_index < (max_level-max_level_y)) ? 0 : 1;
    } else if ( (i%6)==4 ) {
      start_encode = (level_index < (max_level-max_level_u)) ? 0 : 1;
    } else {
      start_encode = (level_index < (max_level-max_level_v)) ? 0 : 1;
    }
    if ( (i%6)==0 && (level_index==0 || level_index==1) ) {
      cbp_index = get_cbp_index(comb_index,i,max_level-level_index);
      Nbits += write_cbp_index(comb_index,level_index,i,cbp_index,
                               max_level-level_index,fgs_bitstream);
    }

    /* Philips Research; SE */
    if (fgs_global->fgs_selective_enhancement_enable && (i%6)==0)
    {
      if (level_index<2 &&
          enh_block_data[i].cbp&&enh_block_data[i+1].cbp&&
          enh_block_data[i+2].cbp&&enh_block_data[i+3].cbp&&
          enh_block_data[i+4].cbp&&enh_block_data[i+5].cbp)
        non_zero_macroblock=0;
      else
        non_zero_macroblock=1;

      if (!mb_shift_factor_sent[i/6] && non_zero_macroblock)
      { /* shift factor not sent yet AND mb is non-empty */
        BitstreamPutBits(fgs_bitstream,1,shift_factor[i/6]+1);
        Nbits += shift_factor[i/6]+1;
        mb_shift_factor_sent[i/6]=1;
      }
    }
    /* end SE */

    if ( start_encode==1 ) {
      t = enh_block_data[i].data_level - (max_level-level_index);
      j=(t>=0?1:0);
      if(enh_block_data[i].cbp!=0 && level_index>1){
	BitstreamPutBits(fgs_bitstream,1-j,1);
        enh_block_data[i].cbp=1-j;
        Nbits += 1;
      }
      if(enh_block_data[i].cbp==0){
        k = ( t>3 ) ? 3 : t;
        if ( fgs_global->fgs_frequency_weighting_enable &&
             fgs_global->fgs_shift_matrix[0]>=2 ) {
          if ( fgs_global->fgs_shift_matrix[0]>=4 ) {
            Nbits += enc_one_block_bp(i,t,table_bpc4fw_h[k],
                                      table_symbol4fw_h[k],fgs_bitstream);
          } else {
            Nbits += enc_one_block_bp(i,t,table_bpc4fw_m[k],
                                      table_symbol4fw_m[k],fgs_bitstream);
          }
        } else {
          Nbits += enc_one_block_bp(i,t,table_bpc[k],table_symbol[k],
                                  fgs_bitstream);
        }
      }
    }
	/* added by MSRCN for FGS error resilience, 2000/8/20 */
	if (!fgs_global -> fgs_resync_marker_disable && i%6 == 5 && i != fgs_global -> Nblock-1)
	{
		if (Nbits - iVPCounter > fgs_global->fgs_video_packet_length)
		{
			iVPCounter = Nbits;
			BitstreamFlush(fgs_bitstream);
			/* VP marker */
			BitstreamPutBits(fgs_bitstream, FGS_VP_MARKER, NUMBITS_FGS_VP_MARKER);
			/* MB index */
			BitstreamPutBits(fgs_bitstream, (i+1)/6, fgs_global->fgs_number_bits_mb);
		}
	}
  }
  return(Nbits);
}

Int
enc_one_block_bp(Int mb_index,Int level,Int *table_one_bp,Int *symbols,
                 Image *fgs_bitstream)
{
  Int mod_index, Nbits, i, os, ind, n;

  Nbits = 0;
  n = 0;
  if ( enh_block_data[mb_index].run[level][0]==63 &&
       enh_block_data[mb_index].end[level][0]==0) {
    /* all zero case */
    mod_index = symbols[0]+symbols[1];
    BitstreamPutBits(fgs_bitstream,
                     table_one_bp[2*mod_index],
                     table_one_bp[2*mod_index+1]);
    Nbits += table_one_bp[2*mod_index+1];
  } else {
    while ( enh_block_data[mb_index].end[level][n]==0 ) n += 1;
    n += 1;
  }
  os = 0;
  for(i=0;i<n;i++){
    ind = enh_block_data[mb_index].run[level][i];
    if ( i!=n-1 ) {
      mod_index=(ind<symbols[0]?ind:-1);
    } else {
      mod_index=(ind<symbols[1]?ind+symbols[0]:-1);
    }

    if(mod_index!=-1) {
      BitstreamPutBits(fgs_bitstream,
                       table_one_bp[2*mod_index],
                       table_one_bp[2*mod_index+1]);
      Nbits += table_one_bp[2*mod_index+1];
    } else { /* escape case */
      mod_index=symbols[0]+symbols[1]+symbols[2];
      BitstreamPutBits(fgs_bitstream,
                       table_one_bp[2*mod_index],
                       table_one_bp[2*mod_index+1]);
      BitstreamPutBits(fgs_bitstream,
                       enh_block_data[mb_index].run[level][i],
                       6);
      BitstreamPutBits(fgs_bitstream,
                       (i==n-1?1:0),
                       1);
      Nbits += table_one_bp[2*mod_index+1]+7;
    }
    if ( enh_block_data[mb_index].sign[os+ind]!=2 &&
         enh_block_data[mb_index].sign[os+ind]!=0 ) {
      BitstreamPutBits(fgs_bitstream,
                       (enh_block_data[mb_index].sign[os+ind]==-1?1:0),
                       1);
      Nbits += 1;
      enh_block_data[mb_index].sign[os+ind] = 2;
    }
    os += enh_block_data[mb_index].run[level][i]+1;
  }
  return(Nbits);
}

Void
FGSBlockQuant(Int *coeff,Int *rcoeff)
{
  Int i;

  Int diff[64], sign[64];
  Int Q_enh = 1; /* Lossless enhancement for now */

  if(fgs_global->fgs_co_located_not_coded)
    for ( i=0; i<64; i++ ) {
      diff[i]=ABS(coeff[i]);
      sign[i]=(coeff[i]==0? 0: SIGN(coeff[i]));
    }
  else
    for ( i=0; i<64; i++ ) {
      diff[i] = coeff[i] - rcoeff[i];
      sign[i] = SIGN(diff[i]);
      diff[i] = ABS(diff[i]);
    }

  /* Philips Research; SE */
  if (fgs_global->fgs_selective_enhancement_enable)
    for ( i=0; i<64; i++ ) diff[i] <<= shift_factor[fgs_global->Nblock/6];
  /* end SE */

  /* Frequency Weighting, Optivision */
  if ( fgs_global->fgs_frequency_weighting_enable )
    for ( i=0; i<64; i++ ) diff[i] <<= fgs_global->fgs_shift_matrix[zigzag[i]];
  /* Frequency Weighting, Optivision */
  
  for ( i=0; i<64; i++ ) {
    diff[i] = diff[i]/Q_enh;
    sign[i] = (diff[i]==0) ? 0 : sign[i];
  }
  block2bp(diff,sign);
  
  return;
}

Void
block2bp(Int *diff, Int *sign)
{
  Int i, j, level, x[64];

  level = find_level(diff);
  enh_block_data[fgs_global->Nblock].data_level = level;
  for ( j=0; j<level; j++ ) {
    for ( i=0; i<64; i++ ){
      x[zigzag[i]] = 0x1 & (diff[i]>>(level-1-j));
    }
    one_bp(x,j);
  }
  for ( i=0; i<64; i++ ) {
    enh_block_data[fgs_global->Nblock].sign[zigzag[i]] = sign[i];
  }
  fgs_global->Nblock += 1;
}

Void
one_bp(Int *in, Int level)
{
  Int i;
  Int n;
  Int curr_run;
  Int runs[64];

  curr_run = 0;
  n = 0;
  for ( i=0; i<64; i++ ) {
    if ( in[i]==0 ) {
      curr_run += 1;
    } else {
      runs[n] = curr_run;
      curr_run = 0;
      n++;
    }
  }

  if(n==0) { /* all zero case */
    enh_block_data[fgs_global->Nblock].run[level][0] = 63;
    enh_block_data[fgs_global->Nblock].end[level][0] = 0;
  }
  for(i=0;i<n;i++){
    enh_block_data[fgs_global->Nblock].run[level][i] = runs[i];
    if ( i!=n-1 ) {
      enh_block_data[fgs_global->Nblock].end[level][i] = 0;
    } else {
      enh_block_data[fgs_global->Nblock].end[level][i] = 1;
    }
  }
}

Int
find_level(Int *in)
{
  Int i, mx, bp_level;
  Int Lmax=MAX_BP_LEVEL;

  mx = 0;
  for ( i=0; i<64; i++ ) {
    if ( in[i]>mx ) {
      mx = in[i];
    }
  }
  i = Lmax-1;
  while ( (mx & (1<<i))==0 && i>=0 ) {
    i -= 1;
  }
  bp_level = i+1;
  return(bp_level);
}

/* FGS CBP (Optivision) */

Int
get_comb_index(Int max_level_y,Int max_level_u,Int max_level_v,Int remain_level)
{
  Int comb_index;

  if ( max_level_y>=remain_level &&
       max_level_u>=remain_level &&
       max_level_v>=remain_level ) {
    comb_index = 0;
  } else if ( max_level_y>=remain_level &&
              max_level_u>=remain_level &&
              max_level_v<remain_level ) {
    comb_index = 1;
  } else if ( max_level_y>=remain_level &&
              max_level_u<remain_level &&
              max_level_v>=remain_level ) {
    comb_index = 2;
  } else if ( max_level_y>=remain_level &&
              max_level_u<remain_level &&
              max_level_v<remain_level ) {
    comb_index = 3;
  } else if ( max_level_y<remain_level &&
              max_level_u>=remain_level &&
              max_level_v>=remain_level ) {
    comb_index = 4;
  } else if ( max_level_y<remain_level &&
              max_level_u>=remain_level &&
              max_level_v<remain_level ) {
    comb_index = 5;
  } else if ( max_level_y<remain_level &&
              max_level_u<remain_level &&
              max_level_v>=remain_level ) {
    comb_index = 6;
  } else {
    comb_index = 7;
    printf("This cannot happen. Something is wrong!!!\n");
    exit(1);
  }
  return(comb_index);
}

Int
get_cbp_index(Int comb_index,Int block_index,Int remain_level)
{
  Int cbp, i, t[6], b[6];

  for ( i=0; i<6; i++ ) {
    t[i] = enh_block_data[block_index+i].data_level - remain_level;
    b[i] = (t[i]>=0) ? 0 : 1;
  }
  switch(comb_index) {
    case 0: cbp = (b[0] << 5) + (b[1] << 4) + (b[2] << 3) + (b[3] << 2)
                + (b[4] << 1) + b[5];
            break;
    case 1: cbp = (b[0] << 4) + (b[1] << 3) + (b[2] << 2) + (b[3] << 1) + b[4];
            break;
    case 2: cbp = (b[0] << 4) + (b[1] << 3) + (b[2] << 2) + (b[3] << 1) + b[5];
            break;
    case 3: cbp = (b[0] << 3) + (b[1] << 2) + (b[2] << 1) + b[3];
            break;
    case 4: cbp = (b[4] << 1) + b[5];
            break;
    case 5: cbp = b[4];
            break;
    case 6: cbp = b[5];
            break;
    default: printf("Invalid combination!\n");
             exit(1);
  }
  return(cbp);
}

Int
write_cbp_index(Int comb_index,Int level_index,Int block_index,Int cbp_index,
                Int remain_level,Image *fgs_bitstream)
{
  Int i, Nbits, b;
  Int cbp_table_3[4] = {7, 11, 13, 14};
  Int cbp_table_2[6] = {3, 5, 6, 9, 10, 12};
  Int cbp_table_1[4] = {8, 4, 2, 1};

  Nbits = 0;
  if ( level_index==0 )  {
    switch(comb_index) {
      case 0: if ( cbp_index==63 ) {
                use_code[block_index/6] = 1;
                BitstreamPutBits(fgs_bitstream,1,1);
                Nbits += 1;
              } else {
                use_code[block_index/6] = 0;
                BitstreamPutBits(fgs_bitstream,0,1);
                Nbits += 1;
                if ( (cbp_index & 0x3)==3 ) {
                  BitstreamPutBits(fgs_bitstream,1,1);
                  BitstreamPutBits(fgs_bitstream,cbp_index>>2,4);
                  Nbits += 5;
                } else {
                  BitstreamPutBits(fgs_bitstream,0,1);
                  BitstreamPutBits(fgs_bitstream,cbp_index>>2,4);
                  BitstreamPutBits(fgs_bitstream,(cbp_index & 0x3)+1,2);
                  Nbits += 7;
                }
              }
              enh_block_data[block_index].cbp = 0x1 & (cbp_index >> 5);
              enh_block_data[block_index+1].cbp = 0x1 & (cbp_index >> 4);
              enh_block_data[block_index+2].cbp = 0x1 & (cbp_index >> 3);
              enh_block_data[block_index+3].cbp = 0x1 & (cbp_index >> 2);
              enh_block_data[block_index+4].cbp = 0x1 & (cbp_index >> 1);
              enh_block_data[block_index+5].cbp = 0x1 & cbp_index;
              break;
      case 1: if ( cbp_index==31 ) {
                use_code[block_index/6] = 1;
                BitstreamPutBits(fgs_bitstream,1,1);
                Nbits += 1;
              } else {
                use_code[block_index/6] = 0;
                BitstreamPutBits(fgs_bitstream,0,1);
                BitstreamPutBits(fgs_bitstream,cbp_index+1,5);
                Nbits += 6;
              }
              enh_block_data[block_index].cbp = 0x1 & (cbp_index >> 4);
              enh_block_data[block_index+1].cbp = 0x1 & (cbp_index >> 3);
              enh_block_data[block_index+2].cbp = 0x1 & (cbp_index >> 2);
              enh_block_data[block_index+3].cbp = 0x1 & (cbp_index >> 1);
              enh_block_data[block_index+4].cbp = 0x1 & cbp_index;
              break;
      case 2: if ( cbp_index==31 ) {
                use_code[block_index/6] = 1;
                BitstreamPutBits(fgs_bitstream,1,1);
                Nbits += 1;
              } else {
                use_code[block_index/6] = 0;
                BitstreamPutBits(fgs_bitstream,0,1);
                BitstreamPutBits(fgs_bitstream,cbp_index+1,5);
                Nbits += 6;
              }
              enh_block_data[block_index].cbp = 0x1 & (cbp_index >> 4);
              enh_block_data[block_index+1].cbp = 0x1 & (cbp_index >> 3);
              enh_block_data[block_index+2].cbp = 0x1 & (cbp_index >> 2);
              enh_block_data[block_index+3].cbp = 0x1 & (cbp_index >> 1);
              enh_block_data[block_index+5].cbp = 0x1 & cbp_index;
              break;
      case 3: if ( cbp_index==15 ) {
                use_code[block_index/6] = 1;
                BitstreamPutBits(fgs_bitstream,1,1);
                Nbits += 1;
              } else {
                use_code[block_index/6] = 0;
                BitstreamPutBits(fgs_bitstream,0,1);
                Nbits += 1;
                if ( ones_of(cbp_index)==3 ) {
                  BitstreamPutBits(fgs_bitstream,1,1);
                  Nbits += 1;
                  for ( i=0; i<4; i++ ) {
                    if ( cbp_index==cbp_table_3[i] ) {
                      BitstreamPutBits(fgs_bitstream,i,2);
                      Nbits += 2;
                    }
                  }
                } else {
                  BitstreamPutBits(fgs_bitstream,0,1);
                  Nbits += 1;
                  if ( ones_of(cbp_index)==2 ) {
                    BitstreamPutBits(fgs_bitstream,1,1);
                    Nbits += 1;
                    for ( i=0; i<6; i++ ) {
                      if ( cbp_index==cbp_table_2[i] ) {
                        BitstreamPutBits(fgs_bitstream,i,3);
                        Nbits += 3;
                      }
                    }
                  } else {
                    BitstreamPutBits(fgs_bitstream,0,1);
                    Nbits += 1;
                    if ( ones_of(cbp_index)==1 ) {
                      BitstreamPutBits(fgs_bitstream,1,1);
                      Nbits += 1;
                      for ( i=0; i<4; i++ ) {
                        if ( cbp_index==cbp_table_1[i] ) {
                          BitstreamPutBits(fgs_bitstream,i,2);
                          Nbits += 2;
                        }
                      }
                    } else {
                      BitstreamPutBits(fgs_bitstream,0,1);
                      BitstreamPutBits(fgs_bitstream,1,1);
                      Nbits += 2;
                    }
                  }
                }
              }
              enh_block_data[block_index].cbp = 0x1 & (cbp_index >> 3);
              enh_block_data[block_index+1].cbp = 0x1 & (cbp_index >> 2);
              enh_block_data[block_index+2].cbp = 0x1 & (cbp_index >> 1);
              enh_block_data[block_index+3].cbp = 0x1 & cbp_index;
              break;
      case 4: BitstreamPutBits(fgs_bitstream,cbp_index,2);
              Nbits += 2;
              enh_block_data[block_index+4].cbp = 0x1 & (cbp_index >> 1);
              enh_block_data[block_index+5].cbp = 0x1 & cbp_index;
              if ( cbp_index==3 ) {
                use_code[block_index/6] = 1;
              } else {
                use_code[block_index/6] = 0;
              }
              break;
      case 5: BitstreamPutBits(fgs_bitstream,cbp_index,1);
              Nbits += 1;
              enh_block_data[block_index+4].cbp = cbp_index;
              use_code[block_index/6] = cbp_index;
              break;
      case 6: BitstreamPutBits(fgs_bitstream,cbp_index,1);
              Nbits += 1;
              enh_block_data[block_index+5].cbp = cbp_index;
              use_code[block_index/6] = cbp_index;
              break;
      default: printf("Invalid combination!\n");
               exit(1);
    }
  } else if ( use_code[block_index/6]==1 ) {
    switch(comb_index) {
      case 0: if ( cbp_index==63 ) {
                BitstreamPutBits(fgs_bitstream,3,3);
                Nbits += 3;
              } else if ( cbp_index==3 ) {
                BitstreamPutBits(fgs_bitstream,2,3);
                Nbits += 3;
              } else if ( (cbp_index & 0x3)==3 ) {
                BitstreamPutBits(fgs_bitstream,1,1);
                BitstreamPutBits(fgs_bitstream,cbp_index>>2,4);
                Nbits += 5;
              } else if ( cbp_index==0 ) {
                BitstreamPutBits(fgs_bitstream,16,5);
                Nbits += 5;
              } else if ( cbp_index==1 ) {
                BitstreamPutBits(fgs_bitstream,4,5);
                Nbits += 5;
              } else if ( cbp_index==2 ) {
                BitstreamPutBits(fgs_bitstream,5,5);
                Nbits += 5;
              } else if ( cbp_index==60 ) {
                BitstreamPutBits(fgs_bitstream,31,5);
                Nbits += 5;
              } else if ( cbp_index==61 ) {
                BitstreamPutBits(fgs_bitstream,6,5);
                Nbits += 5;
              } else if ( cbp_index==62 ) {
                BitstreamPutBits(fgs_bitstream,7,5);
                Nbits += 5;
              } else {
                BitstreamPutBits(fgs_bitstream,0,3);
                BitstreamPutBits(fgs_bitstream,cbp_index,6);
                Nbits += 9;
              }
              enh_block_data[block_index].cbp = 0x1 & (cbp_index >> 5);
              enh_block_data[block_index+1].cbp = 0x1 & (cbp_index >> 4);
              enh_block_data[block_index+2].cbp = 0x1 & (cbp_index >> 3);
              enh_block_data[block_index+3].cbp = 0x1 & (cbp_index >> 2);
              enh_block_data[block_index+4].cbp = 0x1 & (cbp_index >> 1);
              enh_block_data[block_index+5].cbp = 0x1 & cbp_index;
              break;
      case 1: BitstreamPutBits(fgs_bitstream,cbp_index,5);
              Nbits += 5;
              enh_block_data[block_index].cbp = 0x1 & (cbp_index >> 4);
              enh_block_data[block_index+1].cbp = 0x1 & (cbp_index >> 3);
              enh_block_data[block_index+2].cbp = 0x1 & (cbp_index >> 2);
              enh_block_data[block_index+3].cbp = 0x1 & (cbp_index >> 1);
              enh_block_data[block_index+4].cbp = 0x1 & cbp_index;
              break;
      case 2: BitstreamPutBits(fgs_bitstream,cbp_index,5);
              Nbits += 5;
              enh_block_data[block_index].cbp = 0x1 & (cbp_index >> 4);
              enh_block_data[block_index+1].cbp = 0x1 & (cbp_index >> 3);
              enh_block_data[block_index+2].cbp = 0x1 & (cbp_index >> 2);
              enh_block_data[block_index+3].cbp = 0x1 & (cbp_index >> 1);
              enh_block_data[block_index+5].cbp = 0x1 & cbp_index;
              break;
      case 3: BitstreamPutBits(fgs_bitstream,cbp_index,4);
              Nbits += 4;
              enh_block_data[block_index].cbp = 0x1 & (cbp_index >> 3);
              enh_block_data[block_index+1].cbp = 0x1 & (cbp_index >> 2);
              enh_block_data[block_index+2].cbp = 0x1 & (cbp_index >> 1);
              enh_block_data[block_index+3].cbp = 0x1 & cbp_index;
              break;
      case 4: BitstreamPutBits(fgs_bitstream,cbp_index,2);
              Nbits += 2;
              enh_block_data[block_index+4].cbp = 0x1 & (cbp_index >> 1);
              enh_block_data[block_index+5].cbp = 0x1 & cbp_index;
              break;
      case 5: BitstreamPutBits(fgs_bitstream,cbp_index,1);
              Nbits += 1;
              enh_block_data[block_index+4].cbp = cbp_index;
              break;
      case 6: BitstreamPutBits(fgs_bitstream,cbp_index,1);
              Nbits += 1;
              enh_block_data[block_index+5].cbp = cbp_index;
              break;
      default: printf("Invalid combination!\n");
               exit(1);
    }
  } else {
    switch(comb_index) {
      case 0: if ( enh_block_data[block_index].cbp==1 ) {
                b = 0x1 & (cbp_index >> 5);
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index].cbp = b;
              }
              if ( enh_block_data[block_index+1].cbp==1 ) {
                b = 0x1 & (cbp_index >> 4);
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index+1].cbp = b;
              }
              if ( enh_block_data[block_index+2].cbp==1 ) {
                b = 0x1 & (cbp_index >> 3);
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index+2].cbp = b;
              }
              if ( enh_block_data[block_index+3].cbp==1 ) {
                b = 0x1 & (cbp_index >> 2);
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index+3].cbp = b;
              }
              if ( enh_block_data[block_index+4].cbp==1 ) {
                b = 0x1 & (cbp_index >> 1);
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index+4].cbp = b;
              }
              if ( enh_block_data[block_index+5].cbp==1 ) {
                b = 0x1 & cbp_index;
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index+5].cbp = b;
              }
              break;
      case 1: if ( enh_block_data[block_index].cbp==1 ) {
                b = 0x1 & (cbp_index >> 4);
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index].cbp = b;
              }
              if ( enh_block_data[block_index+1].cbp==1 ) {
                b = 0x1 & (cbp_index >> 3);
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index+1].cbp = b;
              }
              if ( enh_block_data[block_index+2].cbp==1 ) {
                b = 0x1 & (cbp_index >> 2);
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index+2].cbp = b;
              }
              if ( enh_block_data[block_index+3].cbp==1 ) {
                b = 0x1 & (cbp_index >> 1);
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index+3].cbp = b;
              }
              if ( enh_block_data[block_index+4].cbp==1 ) {
                b = 0x1 & cbp_index;
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index+4].cbp = b;
              }
              break;
      case 2: if ( enh_block_data[block_index].cbp==1 ) {
                b = 0x1 & (cbp_index >> 4);
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index].cbp = b;
              }
              if ( enh_block_data[block_index+1].cbp==1 ) {
                b = 0x1 & (cbp_index >> 3);
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index+1].cbp = b;
              }
              if ( enh_block_data[block_index+2].cbp==1 ) {
                b = 0x1 & (cbp_index >> 2);
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index+2].cbp = b;
              }
              if ( enh_block_data[block_index+3].cbp==1 ) {
                b = 0x1 & (cbp_index >> 1);
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index+3].cbp = b;
              }
              if ( enh_block_data[block_index+5].cbp==1 ) {
                b = 0x1 & cbp_index;
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index+5].cbp = b;
              }
              break;
      case 3: if ( enh_block_data[block_index].cbp==1 ) {
                b = 0x1 & (cbp_index >> 3);
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index].cbp = b;
              }
              if ( enh_block_data[block_index+1].cbp==1 ) {
                b = 0x1 & (cbp_index >> 2);
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index+1].cbp = b;
              }
              if ( enh_block_data[block_index+2].cbp==1 ) {
                b = 0x1 & (cbp_index >> 1);
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index+2].cbp = b;
              }
              if ( enh_block_data[block_index+3].cbp==1 ) {
                b = 0x1 & cbp_index;
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index+3].cbp = b;
              }
              break;
      case 4: if ( enh_block_data[block_index+4].cbp==1 ) {
                b = 0x1 & (cbp_index >> 1);
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index+4].cbp = b;
              }
              if ( enh_block_data[block_index+5].cbp==1 ) {
                b = 0x1 & cbp_index;
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index+5].cbp = b;
              }
              break;
      case 5: if ( enh_block_data[block_index+4].cbp==1 ) {
                b = 0x1 & cbp_index;
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index+4].cbp = b;
              }
              break;
      case 6: if ( enh_block_data[block_index+5].cbp==1 ) {
                b = 0x1 & cbp_index;
                BitstreamPutBits(fgs_bitstream,b,1);
                Nbits += 1;
                enh_block_data[block_index+5].cbp = b;
              }
              break;
      default: printf("Invalid combination!\n");
               exit(1);
    }
  }
  return(Nbits);
}

Int
ones_of(Int number)
{
  Int i, ones;

  ones = 0;
  for ( i=0; i<31; i++ ) {
    ones += 0x1 & (number>>i);
  }
  return(ones);
}

/* FGS CBP (Optivision) */

/* Philips Research; SE */
Void fill_fgs_rectangular_mask(Int width,Int height,Int start_x,Int start_y,
                               Int end_x,Int end_y,Int sf)
{
  Int mb, mb_x, mb_y;
  Int start_mb_x,start_mb_y,end_mb_x,end_mb_y;
  Int mbs_per_slice, slices_per_vop;

  mbs_per_slice=width/16;
  slices_per_vop=height/16;
  for (mb=0; mb<mbs_per_slice*slices_per_vop; mb++)
    shift_factor[mb]=0;

  start_mb_x=start_x/16;
  start_mb_y=start_y/16;
  end_mb_x=(end_x+15)/16;
  end_mb_y=(end_y+15)/16;
  if (end_mb_x>=mbs_per_slice)
    end_mb_x=mbs_per_slice;
  if (end_mb_y>=slices_per_vop)
    end_mb_y=slices_per_vop;
  for (mb_x=start_mb_x;mb_x<=end_mb_x;mb_x++)
    for (mb_y=start_mb_y;mb_y<=end_mb_y;mb_y++)
      shift_factor[mb_x+mb_y*mbs_per_slice]=sf;
}

Void
get_fgs_mask_from_file(Char *mask_file,Int width,Int height)
{
  FILE *fp_mask;
  Int mb;

  if (!(fp_mask=fopen(mask_file,"r")))
  {
    fprintf(stderr,"error opening %s for reading\n",mask_file);
    exit(1);
  }

  for (mb=0;mb<width*height/256;mb++)
    fscanf(fp_mask,"%d ",&shift_factor[mb]);

  fclose(fp_mask);
}
/* end SE */
